(* file: cr.sml *)
(* description: confluence *)
(* author: Takahito Aoto *)

signature CR = 
sig 
    type atom = Atom.key
    type var = Var.key
    type term = Term.key
    type fresh_env = FreshEnv.key
    type perm = Perm.key
    type subst = Subst.subst
    type rule = Rule.rule
    type rules = Rule.rules
    type bcpk = fresh_env * term * term * term

    val runDebug: bool ref

    val prBCPK: bcpk -> string
    val prBCP: bcpk -> string
    val areEquivBCPK: bcpk * bcpk -> bool

    val bcpkRule: bool -> rule -> rule -> bcpk list
    val bcpkRuleOutIn: bool -> rule -> rule -> (bcpk list * bcpk list)
    val notProperlyOverlapping: rules -> bool
    val isOrthogonal: rules -> bool

    val pbcpeaks: rules -> bcpk list 
    val bcpeaks: rules -> bcpk list 
    val pbcpeaksOutIn:  rules -> bcpk list * bcpk list 
    val bcpeaksOutIn:  rules -> bcpk list * bcpk list 

    datatype result = YES | NO | MAYBE

    val checkConfluenceByOrthogonal: rules -> result
    val checkConfluenceByStronglyClosed: rules -> result
    val checkConfluenceByKnuthBendix: rules -> result
    val checkConfluenceByParallelClosed: rules -> result

    val testMethod: rules -> int -> string
    val prover: rules -> string
    val test: unit -> unit

end


structure Cr: CR =
struct 

local 
    structure L = List
    structure AL = AssocList
    structure LP = ListPair
    structure LU = ListUtil
    structure T = Term
    structure PU = PrintUtil
    structure SU = StringUtil
    structure TU = TestUtil
in

type atom = Atom.key
type var = Var.key
type term = Term.key
type fresh_env = FreshEnv.key
type perm = Perm.key
type subst = Subst.subst
type rule = Rule.rule
type rules = Rule.rules
type bcpk = fresh_env * term * term * term

val runDebug = ref false
fun dprint mesg = if (!runDebug) then PU.print mesg else ()
fun dprintln mesg = if (!runDebug) then PU.println mesg else ()

fun prBCPK (env,x,y,z) = ((FreshEnv.toString env) ^ " |- " ^
			"< " ^ (Term.toString x) ^ " <= " ^
			(Term.toString y) ^ " => " ^ (Term.toString z) ^ " >")

fun prBCP (env,x,y,z) = ((FreshEnv.toString env) ^ " |- " ^
			"<  " ^ (Term.toString x) ^ ",   " ^ (Term.toString z) ^ "  >")


(* equal by ()^\pi and renaming of var?  for env, set equal is OK *)
fun areEquivBCPK ((env1,u1,v1,w1), (env2,u2,v2,w2)) =
    let val atms1 = LU.toSet ((FreshEnv.atoms env1) @ (Term.atoms u1) @ (Term.atoms v1) @ (Term.atoms w1))
	val atms2 = LU.toSet ((FreshEnv.atoms env2) @ (Term.atoms u2) @ (Term.atoms v2) @ (Term.atoms w2))
	val vrs1 = LU.toSet ((FreshEnv.vars env1) @ (Term.vars u1) @ (Term.vars v1) @ (Term.vars w1))
	val vrs2 = LU.toSet ((FreshEnv.vars env2) @ (Term.vars u2) @ (Term.vars v2) @ (Term.vars w2))
    in if length atms1 <> length atms2
	  orelse length vrs1 <> length vrs2
       then false
       else let val natms = Atom.newAtoms [] (length atms1)
		val aren1 = LP.zip (atms1,natms)
		val aren2 = LP.zip (atms2,natms)
		val nvrs = Var.newVars [] (length vrs1)
		val vren1 = LP.zip (vrs1,nvrs)
		val vren2 = LP.zip (vrs2,nvrs)
		val env1a = FreshEnv.rename vren1 (FreshEnv.apply aren1 env1)
		val env2a = FreshEnv.rename vren2 (FreshEnv.apply aren2 env2)
		val sub1 = LP.zip (vrs1,L.map (fn x => T.Susp ([],x)) nvrs)
		val sub2 = LP.zip (vrs2,L.map (fn x => T.Susp ([],x)) nvrs)
		val u1a = Subst.apply sub1 (Term.apply aren1 u1)
		val v1a = Subst.apply sub1 (Term.apply aren1 v1)
		val w1a = Subst.apply sub1 (Term.apply aren1 w1)
		val u2a = Subst.apply sub2 (Term.apply aren2 u2)
		val v2a = Subst.apply sub2 (Term.apply aren2 v2)
		val w2a = Subst.apply sub2 (Term.apply aren2 w2)
	    in (u1a,v1a,w1a) = (u2a,v2a,w2a) andalso LU.isEqset env1a env2a
	    end
    end

val _ = TU.test "areEquivBCPK"
		[ 
		  (areEquivBCPK ((FreshEnv.fromString "b#Y,b#Y,a#Y",
				 Term.fromString "[b][]Y",
				 Term.fromString "(f <[b][]Y,[a][]Y>)",
				 Term.fromString "[b][a][]Y"),
				 (FreshEnv.fromString "c#Y,c#Y,a#Y",
				 Term.fromString "[c][]Y",
				 Term.fromString "(f <[c][]Y,[a][]Y>)",
				 Term.fromString "[c][a][]Y"))
		   = true)
		  ]


(* Computing (proper) basic critical peaks of rule1 on rule2.
   We assume variables in two rules are disjoint.
   Do not put "properOnly = true" if two rules are not identical. *)
local
    fun cpkTerm cntxt properOnly t = let val (xs,ys) = cpkTermOutIn cntxt properOnly t in xs@ys end
    and cpkTermOutIn cntxt properOnly (T.Susp _) = ([],[]) (* variable position do not overlaps see SCSS Def.14 *)
      | cpkTermOutIn (cntxt as (env1,l1,env2,atoms)) properOnly t = 
	let val atRoot = if properOnly then []
			 else L.map (fn (sigma,env,pimap) => (fn x=>x,sigma,env,pimap))
				    (Equivariant.unifyForBCP (env1,l1) (env2,t,atoms))
	    val atArgs = case t of
			     (T.Abst (a,s)) => 
			     L.map (fn (C,sigma,env,pimap) => 
				       (fn x => T.Abst(Perm.apply pimap a,C x),sigma,env,pimap)) (cpkTerm cntxt false s)
			   | (T.Fun (f,s)) => L.map (fn (C,sigma,env,pimap) => (fn x => T.Fun (f,C x),sigma,env,pimap)) (cpkTerm cntxt false s)
			   | (T.Tuple ts) => L.map (fn (C,sigma,env,pimap) => (fn x => T.Tuple (C x),sigma,env,pimap)) (cpkTermList cntxt [] ts)
			   | _ => []
	in (atRoot,atArgs)
	end
    and cpkTermList cntxt _ [] = []
      | cpkTermList cntxt us (t::ts) = 
	let val atHead = L.map (fn (C,sigma,env,pimap) => (fn x => (L.map (Term.apply pimap) us)
								   @ ((C x)::(L.map (Term.apply pimap) ts)),
							   sigma,env,pimap)) 
			       (cpkTerm cntxt false t)
	    val atTail = cpkTermList cntxt (us@[t]) ts
	in atHead @ atTail
	end
in
fun bcpkRule properOnly (env1,l1,r1) (env2,l2,r2) =
    let  val atoms = LU.union (Term.atoms l2, Term.atoms r2)
    in  LU.toSet' areEquivBCPK
		  (L.map (fn (C,sigma,env,pimap) => (LU.toSet env,
				       Subst.apply sigma (C r1),
				       Subst.apply sigma (Term.apply pimap l2),
				       Subst.apply sigma (Term.apply pimap r2)))
			 (cpkTerm (env1,l1,env2,atoms) properOnly l2))
    end
(* outer BCPs and inner BCPs *)
fun bcpkRuleOutIn properOnly (env1,l1,r1) (env2,l2,r2) =
    let  val atoms = LU.union (Term.atoms l2, Term.atoms r2)
	 val (ansout,ansin) = (cpkTermOutIn (env1,l1,env2,atoms) properOnly l2)
	 fun construct (C,sigma,env,pimap) = (LU.toSet env, 
					      Subst.apply sigma (C r1),
					      Subst.apply sigma (Term.apply pimap l2),
					      Subst.apply sigma (Term.apply pimap r2))
    in (L.map construct ansout, L.map construct ansin)
    end
end

val _ = TU.test "bcpkRule"
		[  (* example 14 of [SKAT RTA2015] *)
		  (bcpkRule false 
			    (Rule.fromString "a#X |-  (lam [a](app <[]X,a>)) -> []X")
			    (Rule.fromString "|- (app <(lam [a][]Y), []Z>) -> (sub <[a][]Y, []Z>)")
		   = [(FreshEnv.fromString "a#X",
                       Term.fromString "(app <[]X,[]Z>)",
                       Term.fromString "(app <(lam [a](app <[]X,a>)),[]Z>)",
                       Term.fromString "(sub <[a](app <[]X,a>),[]Z>)"),
		      (FreshEnv.fromString "b#X,a#X",
		       Term.fromString "(app <[]X,[]Z>)",
		       Term.fromString "(app <(lam [b](app <[(a b)]X,b>)),[]Z>)",
		       Term.fromString "(sub <[b](app <[(a b)]X,b>),[]Z>)") ])
		, (* examples 15, 18 of [SKAT RTA2015] *)
		  (bcpkRule false
			  (Rule.fromString "a#Y_1 |- (and <(forall [a][]X_1),[]Y_1>) -> (forall [a](and <[]X_1,[]Y_1>))")
			  (Rule.fromString "a#X_2 |- (and <[]X_2,(forall [a][]Y_2)>) -> (forall [a](and <[]X_2,[]Y_2>))")
(*		   = [([],
                       Term.fromString "(forall [a](and <[]X_1,(forall [a][]Y_2)>))",
                       Term.fromString "(and <(forall [a][]X_1),(forall [a][]Y_2)>)",
                       Term.fromString "(forall [a](and <(forall [a][]X_1),[]Y_2>))"),
		      (FreshEnv.fromString "b#X_1,a#Y_2",
                       Term.fromString "(forall [a](and <[]X_1,(forall [b][]Y_2)>))",
                       Term.fromString "(and <(forall [a][]X_1),(forall [b][]Y_2)>)",
                       Term.fromString "(forall [b](and <(forall [a][]X_1),[]Y_2>))")])
*)
		   = [([],  
                       Term.fromString "(forall [a](and <[]X_1,(forall [a][]Y_2)>))",
                       Term.fromString "(and <(forall [a][]X_1),(forall [a][]Y_2)>)",
                       Term.fromString "(forall [a](and <(forall [a][]X_1),[]Y_2>))"),
		      (FreshEnv.fromString "a#Y_2,b#X_1",
                       Term.fromString "(forall [a](and <[]X_1,(forall [b][]Y_2)>))",
                       Term.fromString "(and <(forall [a][]X_1),(forall [b][]Y_2)>)",
                       Term.fromString "(forall [b](and <(forall [a][]X_1),[]Y_2>))")])
		,
		     (bcpkRule false
			  (Rule.fromString "|- (f <[]X,[]X>) -> []X")
			  (Rule.fromString "|- (f <[a][]Y,[b][]Y>) -> [a][b][]Y")
		     (* = [(FreshEnv.fromString "b#Y,c#Y",
			  Term.fromString "[b][]Y",
			  Term.fromString "(f <[b][]Y,[c][]Y>)",
			  Term.fromString "[b][c][]Y")]) *)
		      = [(FreshEnv.fromString "d#Y,c#Y",
			  Term.fromString "[d][]Y",
			  Term.fromString "(f <[d][]Y,[c][]Y>)",
			  Term.fromString "[d][c][]Y")])
		]

val _ = TU.test "bcpkRuleOutIn"
		[ 
		  (bcpkRuleOutIn false 
				 (Rule.fromString "a#X |-  (lam [a](app <[]X,a>)) -> []X")
				 (Rule.fromString "|- (app <(lam [a][]Y), []Z>) -> (sub <[a][]Y, []Z>)")
		   = ([], [(FreshEnv.fromString "a#X",
                       Term.fromString "(app <[]X,[]Z>)",
                       Term.fromString "(app <(lam [a](app <[]X,a>)),[]Z>)",
                       Term.fromString "(sub <[a](app <[]X,a>),[]Z>)"),
		      (FreshEnv.fromString "b#X,a#X",
		       Term.fromString "(app <[]X,[]Z>)",
		       Term.fromString "(app <(lam [b](app <[(a b)]X,b>)),[]Z>)",
		       Term.fromString "(sub <[b](app <[(a b)]X,b>),[]Z>)") ]))
		, 
		  (bcpkRuleOutIn false
			  (Rule.fromString "a#Y_1 |- (and <(forall [a][]X_1),[]Y_1>) -> (forall [a](and <[]X_1,[]Y_1>))")
			  (Rule.fromString "a#X_2 |- (and <[]X_2,(forall [a][]Y_2)>) -> (forall [a](and <[]X_2,[]Y_2>))")
		   = ([([],
                       Term.fromString "(forall [a](and <[]X_1,(forall [a][]Y_2)>))",
                       Term.fromString "(and <(forall [a][]X_1),(forall [a][]Y_2)>)",
                       Term.fromString "(forall [a](and <(forall [a][]X_1),[]Y_2>))"),
		      ((* FreshEnv.fromString "b#X_1,a#Y_2", *)
			FreshEnv.fromString "a#Y_2,b#X_1",
			Term.fromString "(forall [a](and <[]X_1,(forall [b][]Y_2)>))",
			Term.fromString "(and <(forall [a][]X_1),(forall [b][]Y_2)>)",
			Term.fromString "(forall [b](and <(forall [a][]X_1),[]Y_2>))")], []))
		]

(*
		  (bcpkRuleOutIn false
			  (Rule.fromString "|- (forall [a](forall [b][]X)) -> (forall [b](forall [a][]X))")
			  (Rule.fromString "|- (forall [a](forall [b][]Y)) -> (forall [b](forall [a][]Y))")
*)

local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in
fun notProperlyOverlapping rs =
    let val nums = L.tabulate (L.length rs, fn x=>x)
	val _ = dprintln "check proper overlapping"
	fun checkDifferent (i,j) = let val rule1 = L.nth (rs,i)
				       val rule2 = L.nth (rs,j)
				       val (rule1',rule2') = Rule.rename (rule1,rule2)
				       val overlaps = bcpkRule false  rule1' rule2' 
				   in if null overlaps
				      then true
				      else (dprintln "a proper overlap exists"; 
					    dprintln ((Int.toString i) ^ "-" ^ (Int.toString j));
					    dprintln (prBCPK (hd overlaps));
					    false)
				   end
	fun checkIdentical i = let val rule1 = L.nth (rs,i)
				   val rule2 = L.nth (rs,i)
				   val (rule1',rule2') = Rule.rename (rule1,rule2)
				   val overlaps = bcpkRule true rule1' rule2'
			       in if null overlaps
				  then true
				  else (dprintln "a proper overlap exists"; 
					dprintln ((Int.toString i) ^ "-" ^ (Int.toString i));
					dprintln (prBCPK (hd overlaps));
					false)
			       end
	fun check i = L.all (fn j => if i <> j then checkDifferent (i,j)
				     else checkIdentical i) nums
    in if L.all (fn i => check i) nums
       then (PU.hintln "no proper overlaps"; true) 
       else (PU.hintln "found proper overlaps"; false) 
    end
end

fun isOrthogonal rs =
    Nrs.isLeftLinear rs
    andalso notProperlyOverlapping rs


local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
    fun checkDifferent rs (i,j) = let val rule1 = L.nth (rs,i)
				      val rule2 = L.nth (rs,j)
				      val (rule1',rule2') = Rule.rename (rule1,rule2)
				      val bcpk = bcpkRule false rule1' rule2'
				      val _ = if not (null bcpk)
					      then (dprintln ("(" ^ (Int.toString i) ^ " on " ^ (Int.toString j) ^ "): ");
						    dprint (LU.toStringCommaLnCurly prBCPK bcpk))
					      else ()
				  in bcpk end
    fun checkIdentical rs i = let val rule1 = L.nth (rs,i)
				  val rule2 = L.nth (rs,i)
				  val (rule1',rule2') = Rule.rename (rule1,rule2)
				  (* val _ = dprintln ("(" ^ (Int.toString i) ^ ") on (" ^ (Int.toString i) ^ "): ") *)
				  val bcpk = bcpkRule true rule1' rule2'
				  val _ = if not (null bcpk)
					  then (dprintln ("(" ^ (Int.toString i) ^ ") on (" ^ (Int.toString i) ^ "): ");
						dprint (LU.toStringCommaLnCurly prBCPK bcpk))
					  else ()
			      in bcpk
			      end
    fun checkDifferentOutIn rs (i,j) = let val rule1 = L.nth (rs,i)
					   val rule2 = L.nth (rs,j)
					   val (rule1',rule2') = Rule.rename (rule1,rule2)
					   val (bcpkout,bcpkin) = bcpkRuleOutIn false rule1' rule2'
					   val _ = if not (null bcpkout)
						   then (dprintln ("(" ^ (Int.toString i) ^ " on " ^ (Int.toString j) ^ "(out)): ");
							 dprint (LU.toStringCommaLnCurly prBCPK bcpkout))
					      else ()
					   val _ = if not (null bcpkin)
						   then (dprintln ("(" ^ (Int.toString i) ^ " on " ^ (Int.toString j) ^ "(in)): ");
							 dprint (LU.toStringCommaLnCurly prBCPK bcpkin))
					      else ()
				  in (bcpkout,bcpkin) end
in
fun pbcpeaks rs = 
    let val nums = L.tabulate (L.length rs, fn x=>x)
	(* val _ = dprintln "NRS:" *)
	(* val _ = dprint (LU.toStringCommaLnSquare (fn i => "(" ^ (Int.toString i) ^ "): " ^ (Rule.toString (L.nth (rs,i))))  nums) *)
	val _ = dprintln "PBCPs:"
	fun check i = LU.mapAppend (fn j => if i <> j then checkDifferent rs (i,j)  else checkIdentical rs i) nums
    in LU.mapAppend (fn i => check i) nums
    end
fun bcpeaks rs = 
    let val nums = L.tabulate (L.length rs, fn x=>x)
	(* val _ = dprintln "NRS:" *)
	(* val _ = dprint (LU.toStringCommaLnSquare (fn i => "(" ^ (Int.toString i) ^ "): " ^ (Rule.toString (L.nth (rs,i))))  nums) *)
	val _ = dprintln "BCPs:"
	fun check i = LU.mapAppend (fn j => checkDifferent rs (i,j)) nums
    in LU.mapAppend (fn i => check i) nums
    end
fun pbcpeaksOutIn rs = 
    let val nums = L.tabulate (L.length rs, fn x=>x)
	(* val _ = dprintln "NRS:" *)
	(* val _ = dprint (LU.toStringCommaLnSquare (fn i => "(" ^ (Int.toString i) ^ "): " ^ (Rule.toString (L.nth (rs,i))))  nums) *)
	val _ = dprintln "PBCPs:"
	fun check i = L.map (fn j => if i <> j then checkDifferentOutIn rs (i,j)  else ([], checkIdentical rs i)) nums
	val (xss,yss) = LP.unzip (LU.mapAppend (fn i => check i) nums)
    in (LU.mapAppend (fn x=>x) xss, LU.mapAppend (fn x=>x) yss) 
    end
fun bcpeaksOutIn rs = 
    let val nums = L.tabulate (L.length rs, fn x=>x)
	(* val _ = dprintln "NRS:" *)
	(* val _ = dprint (LU.toStringCommaLnSquare (fn i => "(" ^ (Int.toString i) ^ "): " ^ (Rule.toString (L.nth (rs,i))))  nums) *)
	val _ = dprintln "BCPs:"
	fun check i = L.map (fn j => checkDifferentOutIn rs (i,j)) nums
	val (xss,yss) = LP.unzip (LU.mapAppend (fn i => check i) nums)
    in (LU.mapAppend (fn x=>x) xss, LU.mapAppend (fn x=>x) yss) 
    end
end

val _ = TU.test "pbcpeaks"
		[ (* Example 29 of [SKTA, SCSS 2016] *)
		  (pbcpeaks (Nrs.example 28)
		   = [ ([],
			Term.fromString "(not (forall [a](not []Q)))",
			Term.fromString "(not (not (exists [a][]Q)))", 
			Term.fromString "(exists [a][]Q)"),
		       ([],
			Term.fromString "(not (exists [a](not []Q)))",
		       Term.fromString "(not (not (forall [a][]Q)))",
		       Term.fromString "(forall [a][]Q)"),
		       ([],
			Term.fromString "(not []X)",
			Term.fromString "(not (not (not []X)))",
			Term.fromString "(not []X)") ])
		,  (* Example 33 of [SKTA, SCSS 2016] *)
		  (pbcpeaks (Nrs.example 25)
		   = [ ([],
			Term.fromString "(forall [b](forall [b](forall [a][]X)))",
			Term.fromString "(forall [b](forall [a](forall [b][]X)))",
			Term.fromString "(forall [a](forall [b](forall [b][]X)))"),
		       ([],
			Term.fromString "(forall [c](forall [b](forall [a][]X)))", 
			Term.fromString "(forall [c](forall [a](forall [b][]X)))",
			Term.fromString "(forall [a](forall [c](forall [b][]X)))"),
		       ([],
			Term.fromString "(forall [a](forall [b](forall [a][]X)))",
			Term.fromString "(forall [a](forall [b](forall [a][(a b)]X)))",
			Term.fromString "(forall [b](forall [a](forall [a][(a b)]X)))"),
		       ([],
			Term.fromString "(forall [c](forall [b](forall [a][]X)))",
			Term.fromString "(forall [c](forall [b](forall [a][(a b)]X)))",
			Term.fromString "(forall [b](forall [c](forall [a][(a b)]X)))"),
		       (FreshEnv.fromString "c#X",
			Term.fromString "(forall [b](forall [b](forall [a][]X)))",
			Term.fromString "(forall [b](forall [c](forall [b][(a c)]X)))", 
			Term.fromString "(forall [c](forall [b](forall [b][(a c)]X)))"),
		       (FreshEnv.fromString "c#X",
			Term.fromString "(forall [a](forall [b](forall [a][]X)))",
			Term.fromString "(forall [a](forall [c](forall [b][(a c)]X)))",
			Term.fromString "(forall [c](forall [a](forall [b][(a c)]X)))"),
		       (FreshEnv.fromString "d#X",
			Term.fromString "(forall [c](forall [b](forall [a][]X)))",
			Term.fromString "(forall [c](forall [d](forall [b][(a d)]X)))",
			Term.fromString "(forall [d](forall [c](forall [b][(a d)]X)))") 
		     ])
		]

val _ = TU.test "bcpeaksOutIn"
		[
		  (bcpeaksOutIn (Nrs.example 2)
		   = ([ (FreshEnv.fromString "a#X_1",
			Term.fromString "[]X_1",
			Term.fromString "(lam [a](app <[]X_1,a>))",
			Term.fromString "[]X_1"),
		       (FreshEnv.fromString "b#X_1,a#X_1", 
			Term.fromString "[(a b)]X_1",
			Term.fromString "(lam [b](app <[]X_1,b>))",
			Term.fromString "[]X_1")], []))
		]

local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in
fun hasStronglyClosedBCPs rs =
    let exception Join of term
	fun findJoin (env,x,_,y) = 
	    let val xs = x::(Rewrite.reducts rs (env,x))
		val ys = y::(Rewrite.reducts rs (env,y))
		val _ = ListXProd.appX (fn (x1,y1) => if Equiv.aeq env (x1,y1) then raise Join x1 else ())  (xs,ys)
	    in NONE end 
	    handle Join t => SOME t
	fun isClosed (bcpk as (env,x,_,y)) = 
	    Equiv.aeq env (x,y)
	    orelse case findJoin bcpk of
		       NONE => (dprintln ((prBCP bcpk) ^ " is not closed"); false)
		     | SOME t => (dprintln ((prBCP bcpk) ^ " is closed at " ^ (Term.toString t)); true)
    in if L.all isClosed (bcpeaks rs)
       then (PU.hintln "all BCPs are strongly closed"; true) 
       else (PU.hintln "failed to show all BCPs are strongly closed"; false) 
    end
end

local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in
fun hasJoinableBCPs rs = 
    let val _ = PU.hintln "check each BCP has alpha-equivalent normal forms or not"
	fun checknf (bcpk as (env,x,_,y)) = 
	    Equiv.aeq env (x,y)
	    orelse let val _ = dprintln ("BCP: " ^ (prBCP bcpk))
		       val x1 = Rewrite.linf rs (env,x)
		       val _ = dprint ("===> nf(lhs):" ^ (Term.toString x1))
		       val y1 = Rewrite.linf rs (env,y)
		       val _ = dprint (" and nf(rhs):" ^ (Term.toString y1))
		       val ans = Equiv.aeq env (x1,y1)
		       val _ = dprintln (if ans then " are aeq" else " are not aeq")
		   in ans end
    in  if L.all checknf (bcpeaks rs) 
	then (PU.hintln "all BCPs are joinable"; true) 
	else (PU.hintln "some BCP is not joinable"; false) 

    end
end


local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in
fun hasParallelClosedBCPs rs =
    let exception Join of term
	fun findJoin (env,x,_,y) = let val xs = Rewrite.parallelReducts rs (env,x)
				       val ys = Rewrite.parallelReducts rs (env,y)
				       val _ = ListXProd.appX (fn (x1,y1) => if Equiv.aeq env (x1,y1) then raise Join x1 else ())  (xs,ys)
				   in NONE end 
				   handle Join t => SOME t
	fun checkOuter (bcpk as (env,x,_,y)) = 
	    Equiv.aeq env (x,y)
	    orelse
	    case findJoin bcpk of
		NONE => (dprintln ("outBCP: " ^ (prBCP bcpk) ^ " is not closed"); false)
	      | SOME t => (dprintln ("outBCP: " ^ (prBCP bcpk) ^ " is closed at " ^ (Term.toString t)); true)
	fun checkInner (bcpk as (env,x,_,y)) = 
	    Equiv.aeq env (x,y)
	    orelse 
	    let val xs = Rewrite.parallelReducts rs (env,x)
		val ans = L.exists (fn x1 => Equiv.aeq env (x1,y)) xs
		val _ = if ans then () else dprintln ("inBCP: " ^ (prBCP bcpk) ^ " is not closed")
	    in ans end
	val (bcpkout,bcpkin) = bcpeaksOutIn rs
    in  if (L.all checkInner bcpkin) andalso (L.all checkOuter bcpkout) 
	then (PU.hintln "all BCPs are parallel closed"; true)
	else (PU.hintln "failed to show all BCPs are parallel closed"; false) 
    end
end

	
datatype result = YES | NO | MAYBE

fun prResult YES = "YES"
  | prResult NO = "NO"
  | prResult MAYBE = "MAYBE"

fun checkConfluenceByOrthogonal rs =
    if Nrs.isAbstractSkeletonPreserving rs
       andalso isOrthogonal rs
    then YES
    else MAYBE

fun checkConfluenceByStronglyClosed rs =
    if Nrs.isLinear rs
       andalso Nrs.isUniform rs
       andalso hasStronglyClosedBCPs rs
    then YES
    else MAYBE

fun checkConfluenceByKnuthBendix rs =
    if Nrs.isUniform rs
       andalso SnSolver.isTerminatingByExternalSolver rs
    then if hasJoinableBCPs rs then YES else NO
    else MAYBE


fun checkConfluenceByParallelClosed rs =
    if Nrs.isLeftLinear rs
       andalso Nrs.isUniform rs
       andalso hasParallelClosedBCPs rs
    then YES
    else MAYBE

local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
    exception Solved of string
    val totalOfMethods = 4
    fun method 1 = checkConfluenceByOrthogonal
      | method 2 = checkConfluenceByStronglyClosed
      | method 3 = checkConfluenceByKnuthBendix
      | method 4 = checkConfluenceByParallelClosed
      | method _ = (fn x => MAYBE)
    fun message 1 = "Check confluence by orthogonality"
      | message 2 = "Check confluence by strongly closed criterion"
      | message 3 = "Check confluence by Knuth-Bendix criterion"
      | message 4 = "Check confluence by parallel closed criterion"
      | message _ = ""
in 
fun printNRSwithNum rs = 
    let val nums = L.tabulate (L.length rs, fn x=>x)
	val _ = PU.hintln "NRS:"
	val _ = PU.hint (LU.toStringCommaLnSquare (fn i => "(" ^ (Int.toString i) ^ "): " ^ (Rule.toString (L.nth (rs,i))))  nums)
    in () end
fun try rs j = 
    let val _ = PU.hintln (message j)
	val ans = prResult (method j rs)
    in if ans = "YES" orelse ans = "NO" then raise Solved ans else ()
    end
fun testMethod rs num = 
    let val _ = printNRSwithNum rs
    in if Nrs.breakVarCond rs
       then (PU.hintln "result: MAYBE"; "MAYBE")
       else let val _ = try rs num 
	    in (PU.hintln "result: MAYBE"; "MAYBE") end
	    handle Solved ans => (PU.hintln ("result: " ^ ans); ans)
    end
fun prover rs = 
    let val _ = printNRSwithNum rs
    in if Nrs.breakVarCond rs
       then (PU.hintln "result: MAYBE"; "MAYBE")
       else let val methods = L.tabulate (totalOfMethods, fn x=>x+1)
		val _ = L.app (try rs) methods
	    in (PU.hintln "result: MAYBE"; "MAYBE") end
	    handle Solved ans => (PU.hintln ("result: " ^ ans); ans)
    end
end

fun test () =
    let val _ = PU.print "No."
	val _ = PU.print ": Orthogonal"
	val _ = PU.print ": StronglyClosed"
	val _ = PU.print ": KnuthBendix"
	val _ = PU.print ": ParallelClosed"
	val _ = PU.println ""

	fun method 1 = checkConfluenceByOrthogonal
	  | method 2 = checkConfluenceByStronglyClosed
	  | method 3 = checkConfluenceByKnuthBendix
	  | method 4 = checkConfluenceByParallelClosed

	val totalOfExamples = 30
	val totalOfMethods = 4
    in L.app (fn i => (print (Int.toString i);
		       L.app (fn j => PU.print (": " ^ prResult (method j (Nrs.example i))))
			     (L.tabulate (totalOfMethods, fn x=>x+1));
		       PU.println ""))
	     (L.tabulate (totalOfExamples, fn x=>x+1))
    end


end (* of local *)
end (* of struct *)
