(* file: equivariant.sml *)
(* description: equivariant unification and related definisions *)
(* author: Takahito Aoto *)

signature EQUIVARIANT = 
sig 
    type atom = Atom.key
    type var = Var.key
    type fun_key = Fun.key
    type term = Term.key
    type fresh_env = FreshEnv.key
    type perm = Perm.key
    type subst = Subst.key

    val runDebug: bool ref

    datatype atom_expr = Ca of atom   (* atom constants *)
		       | Xa of string  (* atom variables *)

    datatype perm_expr = Id         
		       | Xp of string  (* permutation variables *)
		       | Swap of atomic_expr * atomic_expr
		       | Inv of perm_expr 
		       | Comp of perm_expr * perm_expr
	 and atomic_expr = Simp of atom_expr   (* simple *)
			 | Mod of perm_expr * atomic_expr  (* moderated *)
						 
    datatype term_expr = Atomic of atomic_expr
		       | Susp of perm_expr * var 
		       | Abst of atomic_expr * term_expr
		       | Fun of fun_key * term_expr
		       | Tuple of term_expr list

    val term2expr: Term.term -> term_expr

    val actAtom: perm_expr -> atom_expr -> atomic_expr
    val actAtomic: perm_expr -> atomic_expr -> atomic_expr
    val actPerm: perm_expr -> perm_expr -> perm_expr

    val actDotTerm: perm_expr -> term_expr -> term_expr
    val actHatTerm: perm_expr -> term_expr -> term_expr

    (* identity_problem: atom_identity_problem *)
    datatype identity_problem = I_Eq of (atom_expr * atom_expr)              
			      | I_Neq of (atom_expr * atom_expr)
			      | I_Sub of string * atom_expr                         (* string for atom variable *)
			      | I_Pmap of (string * atom_expr * atom_expr)           (* string for permunation variable *)
			      (* | I_Pnmap of (string * atom_expr * atom_expr)  *)         (* string for permunation variable *)
			      | I_Fresh of (atom_expr * var)           
			      | I_Dfresh of var * perm_expr * perm_expr  (* delayed fresh *)
			      | I_Vmap of var * term_expr   (* solved term variable *)

    val prIdProblem: identity_problem -> string

    val solveIdProblem: identity_problem list -> (identity_problem list) option

    (* atomic_problem: atomic_problem *)
    datatype atomic_problem = A_Eq of atomic_expr * atomic_expr
		     | A_Neq of atomic_expr * atomic_expr
		     | A_Fresh of atomic_expr * var
		     | A_Dfresh of var * perm_expr * perm_expr  (* delayed fresh *)
		     | A_Vmap of var * term_expr   (* solved term variable *)

    val solveAtomicProblem: atomic_problem list -> (identity_problem list) list

    (* unif_problem: equivariant_nominal_unification_problem *)
    datatype unif_problem = U_Eq of term_expr * term_expr
		     | U_Fresh of atomic_expr * term_expr

    val unify: unif_problem list -> (identity_problem list) list

    exception EqvUnifError
    val unifyForBCP: fresh_env * term
		     -> fresh_env * term * atom list
		     -> (subst * fresh_env * (atom * atom) list) list

    val match0: unif_problem list * ((var * term_expr) list) -> (atomic_problem list * ((var * term_expr) list)) list
    val match: unif_problem list -> (identity_problem list) list

    val matchForRewrite: term * atom list -> term -> (subst * fresh_env * (atom * atom) list) list

end

structure Equivariant: EQUIVARIANT =
struct 

local 
    structure L = List
    structure AL = AssocList
    structure LP = ListPair
    structure LU = ListUtil
    structure T = Term
    structure PU = PrintUtil
    structure SU = StringUtil
    structure TU = TestUtil
in

val runDebug = ref false
fun dprint mesg = if (!runDebug) then PU.print mesg else ()
fun dprintln mesg = if (!runDebug) then PU.println mesg else ()

(******* basic data structure *****)

type atom = Atom.key     (* a, b, c .. *)
type var = Var.key       (* X, Y, Z, ... *)
type fun_key = Fun.key   (* f, g, h, ... *)
type term = Term.key     
type fresh_env = FreshEnv.key
type perm = Perm.key     (* pi, pi', ... *)
type subst = Subst.key

datatype atom_expr = Ca of atom        (* a, b,c  ...: atom *)
		   | Xa of string     (* A, B, C, ..: atom variables  *)
(* denoted by alpha, beta, ... *)

datatype perm_expr = Id
		   | Xp of string  (* P, Q, R,.. : permutation variables *)
		   | Swap of atomic_expr * atomic_expr 
		   | Inv of perm_expr 
		   | Comp of perm_expr * perm_expr
     (* denoted by Pi, Pi', Psi, Phi, ... *)
     and atomic_expr = Simp of atom_expr 
		     | Mod of perm_expr * atomic_expr
     (* denoted by  v, w, ... *)

datatype term_expr = Atomic of atomic_expr
		   | Susp of perm_expr * var 
		   | Abst of atomic_expr * term_expr
		   | Fun of fun_key * term_expr
		   | Tuple of term_expr list

(******* checking functions *****)

fun isAtomConst (Ca _) = true
  | isAtomConst (Xa _) = false

fun isAtomVar (Ca _) = false
  | isAtomVar (Xa _) = true

(******* print functions *****)

fun prAtom (Ca a) = Atom.toString a
  | prAtom (Xa str) = str

fun prPerm Id = "Id"
  | prPerm (Xp str) = str
  | prPerm (Swap (v,w)) =  "(" ^ (prAtomic v) ^ " " ^ (prAtomic w) ^ ")"
  | prPerm (Inv Pi) =  (prPerm Pi) ^ "^{-1}"
  | prPerm (Comp (Pi,Psi)) =  "(" ^ (prPerm Pi) ^ " o " ^ (prPerm Psi) ^ ")" 
and prAtomic (Simp alpha) = prAtom alpha
  | prAtomic (Mod (Pi,v)) = (prPerm Pi) ^"*" ^ (prAtomic v)

fun prTerm (Atomic v) =  (prAtomic v)
  | prTerm (Susp (Id,X))  = (Var.toString X)
  | prTerm (Susp (Pi,X))  = "(" ^ (prPerm Pi) ^ ")*" ^ (Var.toString X)
  | prTerm (Abst (v,exp))  = "[" ^ (prAtomic v) ^ "]" ^ (prTerm exp)
  | prTerm (Fun (f,exp))  = "(" ^ (Fun.toString f) ^  " " ^ (prTerm exp) ^ ")"
  | prTerm (Tuple exps) = LU.toStringCommaAngle prTerm exps

(** tranlation from T.term **)
fun atom2expr a = Ca a
fun perm2expr [] = Id
  | perm2expr ((a,b)::xs) = Comp (Swap (Simp (atom2expr a), Simp (atom2expr b)), perm2expr xs)

fun term2expr (T.Atom a) = Atomic (Simp (atom2expr a))
  | term2expr (T.Susp (pi,X))  = Susp (perm2expr pi, X)
  | term2expr (T.Abst (a,t))  = Abst (Simp (atom2expr a), term2expr t)
  | term2expr (T.Fun (f,t))  = Fun (f, term2expr t)
  | term2expr (T.Tuple ts) = Tuple (L.map term2expr ts)

(** tranlation to T.term **)
exception TranslationFromExprToTermFailure
fun atom_expr2atom (Ca a) = a
  | atom_expr2atom (Xa _) = raise TranslationFromExprToTermFailure

fun perm_expr2perm Id = []
  | perm_expr2perm (Swap (Simp alpha, Simp beta)) =  [(atom_expr2atom alpha, atom_expr2atom beta)]
  | perm_expr2perm (Comp (Pi,Psi)) = (perm_expr2perm Pi) @ (perm_expr2perm Psi)
  | perm_expr2perm _ =  raise TranslationFromExprToTermFailure
and atomic_expr2atomic (Simp alpha) = ([], atom_expr2atom alpha)
  | atomic_expr2atomic (Mod (Pi,v)) = let val (xs,y) = atomic_expr2atomic v
					  val zs = perm_expr2perm Pi
				      in (zs@xs, y)
				      end

fun term_expr2term (Atomic (Simp alpha)) =  T.Atom (atom_expr2atom alpha)
  | term_expr2term (Susp (Pi,X))  = T.Susp (perm_expr2perm Pi, X)
  | term_expr2term (Abst (Simp alpha,exp))  = T.Abst (atom_expr2atom alpha, term_expr2term exp)
  | term_expr2term (Fun (f,exp))  = T.Fun (f, term_expr2term exp)
  | term_expr2term (Tuple exps) = T.Tuple (L.map term_expr2term exps)
  | term_expr2term _ =  raise TranslationFromExprToTermFailure

(******* actions *****)

fun actAtom Pi alpha = Mod (Pi, Simp alpha)  (* return atomic *)

fun invPerm (Inv Pi) = Pi
  | invPerm Id = Id
  | invPerm Pi = Inv Pi

fun actPerm Id Psi = Psi
  | actPerm Pi Id = Pi
  | actPerm Pi Psi = Comp (Pi, Psi)
    
fun actAtomic Pi v = Mod (Pi, v)

fun actDotTerm Pi (Atomic v) = Atomic (actAtomic Pi v)
  | actDotTerm Pi (Susp (Psi,X)) = Susp (Comp (Pi,Psi), X) 
  | actDotTerm Pi (Abst (v,exp)) = Abst (actAtomic Pi v, actDotTerm Pi exp)
  | actDotTerm Pi (Fun (f,exp)) = Fun (f, actDotTerm Pi exp)
  | actDotTerm Pi (Tuple exps) = Tuple (L.map (actDotTerm Pi) exps)

fun actHatTerm Pi (Atomic v) = Atomic (actAtomic Pi v)
  | actHatTerm Pi (Susp (Psi,X)) = Susp (actPerm Pi (actPerm Psi (invPerm Pi)), X) 
  | actHatTerm Pi (Abst (v,exp)) = Abst (actAtomic Pi v, actHatTerm Pi exp)
  | actHatTerm Pi (Fun (f,exp)) = Fun (f, actHatTerm Pi exp)
  | actHatTerm Pi (Tuple exps) = Tuple (L.map (actHatTerm Pi) exps)

(******* atom substitutions, we call 'update' *****)

fun updateAtom sigma (Ca a) = (Ca a)
  | updateAtom sigma (Xa str) = (case AL.find str sigma of SOME alpha => alpha | NONE => Xa str)

fun updatePerm sigma Id = Id
  | updatePerm sigma (Xp str) = (Xp str)
  | updatePerm sigma (Swap (v,w)) = Swap (updateAtomic sigma v, updateAtomic sigma w)
  | updatePerm sigma (Inv Pi) = Inv (updatePerm sigma Pi)
  | updatePerm sigma (Comp (Pi,Psi)) = Comp (updatePerm sigma Pi, updatePerm sigma Psi)
and updateAtomic sigma (Simp alpha) = Simp (updateAtom sigma alpha)
  | updateAtomic sigma (Mod (Psi,v)) = Mod (updatePerm sigma Psi, updateAtomic sigma v)

fun updateTerm sigma (Atomic v) = Atomic (updateAtomic sigma v)
  | updateTerm sigma (Susp (Psi,X)) = Susp (updatePerm sigma Psi, X)
  | updateTerm sigma (Abst (v,exp)) = Abst (updateAtomic sigma v, updateTerm sigma exp)
  | updateTerm sigma (Fun (f,exp)) = Fun (f, updateTerm sigma exp)
  | updateTerm sigma (Tuple exps) = Tuple (L.map (updateTerm sigma) exps)

(******* simplification *****)

fun simplifyAtom alpha = alpha

(* rho = [((P,a), b), ...] i.e. P:a |->b *)
fun simplifyPerm0 0 (Comp (Id, Pi)) = Pi
  | simplifyPerm0 1 (Comp (Pi, Id)) = Pi
  | simplifyPerm0 2 (Inv (Comp (Pi, Psi))) = Comp (invPerm Psi, invPerm Pi)
  | simplifyPerm0 3 (Inv (Inv Pi)) = Pi
  | simplifyPerm0 4 (Comp (Inv Pi, Psi)) = if Pi = Psi then Id else Comp (invPerm Pi, Psi)
  | simplifyPerm0 5 (Comp (Pi, Inv Psi)) = if Pi = Psi then Id else Comp (Pi, invPerm Psi)
  | simplifyPerm0 6 (Inv (Swap (v,w))) = Swap (v,w)
  | simplifyPerm0 7 (Inv Id) = Id
  | simplifyPerm0 8 (Pi as (Comp (Swap (v,w), Swap (v',w')))) 
    = if (v = v' andalso w = w') orelse (v = w' andalso w = v')
      then Id else Pi
  | simplifyPerm0 9 (Pi as (Comp (Swap (v,w), Comp (Swap (v',w'), u))))
    = if (v = v' andalso w = w') orelse (v = w' andalso w = v')
      then u else Pi
  | simplifyPerm0 10 (Comp (Comp (Pi, Psi), Phi)) = Comp (Pi, Comp (Psi, Phi))
  | simplifyPerm0 _ Pi = Pi
val numOfSimplifyPermRules = 11
fun simplifyPerm0all Pi = L.foldl (fn (n, Psi) => simplifyPerm0 n Psi) Pi (L.tabulate (numOfSimplifyPermRules + 1, fn x=>x))

val numOfSimplifyAtomicRules = 6
fun simplifyAtomic0 0 (Mod (Comp (Pi,Psi), v)) = Mod (Pi, Mod (Psi, v))
  | simplifyAtomic0 1 (v as (Mod (Swap (Simp alpha, Simp beta), Simp gamma))) =
    if alpha = gamma then Simp beta else if beta = gamma then Simp alpha
    else if L.all isAtomConst [alpha,beta,gamma]
    then Simp gamma else v 
  | simplifyAtomic0 2 (v as (Mod (Swap (Simp alpha, w), Simp beta))) = if alpha = beta then w else v
  | simplifyAtomic0 3 (v as (Mod (Swap (w, Simp alpha), Simp beta))) = if alpha = beta then w else v
  | simplifyAtomic0 4 (Mod (Id, v)) = v
  | simplifyAtomic0 5 (v as (Mod (Swap (v1,w1), Mod (Swap (v2,w2), v3)))) =
    if v1 = v2 andalso w1 = w2 then v3
    else if v1 = w2 andalso w1 = v2 then v3
    else v
  | simplifyAtomic0 _  v = v
and simplifyAtomic0all v = L.foldl (fn (n, w) => simplifyAtomic0 n w) v (L.tabulate (numOfSimplifyAtomicRules + 1, fn x=>x))

fun simplifyPerm1 Id = Id
  | simplifyPerm1 (Xp str) = (Xp str)
  | simplifyPerm1 (Swap (v,w)) = simplifyPerm0all (Swap (simplifyAtomic1 v, simplifyAtomic1 w))
  | simplifyPerm1 (Inv Pi) = simplifyPerm0all (invPerm (simplifyPerm1 Pi))
  | simplifyPerm1 (Comp (Pi,Psi)) = simplifyPerm0all (Comp (simplifyPerm1 Pi, simplifyPerm1 Psi))
and simplifyAtomic1 (Simp alpha) = (Simp alpha)
  | simplifyAtomic1 (Mod (Psi,v)) = simplifyAtomic0all (Mod (simplifyPerm1 Psi, simplifyAtomic1 v))

fun simplifyPerm Pi = let val Psi = simplifyPerm1 Pi
		      in if Pi = Psi then Pi
			 else simplifyPerm Psi
			 end
and simplifyAtomic v = let val w = simplifyAtomic1 v
		       in if v = w then v
			  else simplifyAtomic w
		       end

fun simplifyTerm (Atomic v) = Atomic (simplifyAtomic v)
  | simplifyTerm (Susp (Psi,X)) = Susp (simplifyPerm Psi, X)
  | simplifyTerm (Abst (v,exp)) = Abst (simplifyAtomic v, simplifyTerm exp)
  | simplifyTerm (Fun (f,exp)) = Fun (f, simplifyTerm exp)
  | simplifyTerm (Tuple exps) = Tuple (L.map (simplifyTerm) exps)

fun simplifyPermByAssign1 rho Id  = Id
  | simplifyPermByAssign1 rho (Xp str) = (Xp str)
  | simplifyPermByAssign1 rho (Swap (v,w)) = simplifyPerm0all (Swap (simplifyAtomicByAssign1 rho v, simplifyAtomicByAssign1 rho w))
  | simplifyPermByAssign1 rho (Inv Pi) = simplifyPerm0all (invPerm (simplifyPermByAssign1 rho Pi))
  | simplifyPermByAssign1 rho (Comp (Pi,Psi)) = simplifyPerm0all (Comp (simplifyPermByAssign1 rho Pi, simplifyPermByAssign1 rho  Psi))
and simplifyAtomicByAssign1 rho (Simp alpha) = (Simp alpha)
  | simplifyAtomicByAssign1 rho (Mod (Psi, Simp alpha)) = 
    let val Psi' = simplifyPermByAssign1 rho Psi
    (* val _ = PU.println (prAtomic (Psi,alpha)) *)
    in case Psi' of
    	   (Xp P) => (case AL.find (P,alpha) rho of
    			  SOME beta => (* PU.println (" ====> " ^ (prAtomic (Id,beta))); *) (Simp beta)
    			| NONE => simplifyAtomic0all (Mod (Psi', Simp alpha)))
    	 | Inv (Xp P) => (case AL.find (P,alpha) 
				       (L.map (fn ((P',alpha'),beta') => ((P',beta'),alpha')) rho) of
    			      SOME beta => (Simp beta)
    			    | NONE => simplifyAtomic0all (Mod (Psi', Simp alpha)))
    	 | _ =>  Mod (Psi', Simp alpha)
    end
  | simplifyAtomicByAssign1 rho (Mod (Psi, v)) = Mod (simplifyPermByAssign1 rho Psi, simplifyAtomicByAssign1 rho v)

fun simplifyPermByAssign rho Pi = let val Psi = simplifyPerm (simplifyPermByAssign1 rho Pi)
				  in if Pi = Psi then Pi
				     else simplifyPermByAssign rho Psi
				  end
fun simplifyAtomicByAssign rho v = let val w = simplifyAtomic (simplifyAtomicByAssign1 rho v)
				   in if v = w then w
				      else simplifyAtomicByAssign rho w
				   end

fun simplifyTermByAssign rho (Atomic v) = Atomic (simplifyAtomicByAssign rho v)
  | simplifyTermByAssign rho (Susp (Psi,X)) = Susp (simplifyPermByAssign rho Psi, X)
  | simplifyTermByAssign rho (Abst (v,exp)) = Abst (simplifyAtomicByAssign rho v, simplifyTermByAssign rho exp)
  | simplifyTermByAssign rho (Fun (f,exp)) = Fun (f, simplifyTermByAssign rho exp)
  | simplifyTermByAssign rho (Tuple exps) = Tuple (L.map (simplifyTermByAssign rho) exps)

(* replacement of atomic *)

fun replaceAtomicInPerm (u,alpha) Id  = Id
  | replaceAtomicInPerm (u,alpha) (Xp str) = (Xp str)
  | replaceAtomicInPerm (u,alpha) (Swap (v,w)) = Swap (replaceAtomicInAtomic (u,alpha) v, replaceAtomicInAtomic (u,alpha) w)
  | replaceAtomicInPerm (u,alpha) (Inv Pi) = Inv (replaceAtomicInPerm (u,alpha) Pi)
  | replaceAtomicInPerm (u,alpha) (Comp (Pi,Psi)) = Comp (replaceAtomicInPerm (u,alpha) Pi, replaceAtomicInPerm (u,alpha) Psi)
and replaceAtomicInAtomic (u,alpha) w = 
    if u = w then (Simp alpha) 
    else case w of 
	     (Simp _) => w
	   | (Mod (Pi,v)) => Mod (replaceAtomicInPerm (u,alpha) Pi, replaceAtomicInAtomic (u,alpha) v)

fun replaceAtomicInTerm (v,alpha) (Atomic w) = Atomic (replaceAtomicInAtomic (v,alpha) w)
  | replaceAtomicInTerm (v,alpha) (Susp (Pi,X)) = Susp (replaceAtomicInPerm (v,alpha) Pi, X)
  | replaceAtomicInTerm (v,alpha) (Abst (beta,exp)) = Abst (beta, replaceAtomicInTerm (v,alpha) exp)
  | replaceAtomicInTerm (v,alpha) (Fun (f,exp)) = Fun (f, replaceAtomicInTerm (v,alpha) exp)
  | replaceAtomicInTerm (v,alpha) (Tuple exps) = Tuple (L.map (replaceAtomicInTerm (v,alpha)) exps)

(******* term substitutions *****)

fun applySubstToTerm sigma (Atomic v) = Atomic v 
  | applySubstToTerm sigma (Susp (Pi,X)) = (case AL.find X sigma of SOME exp => actDotTerm Pi exp | NONE => Susp (Pi,X))
            (*** note here that we need to apply action ***)
  | applySubstToTerm sigma (Abst (alpha,exp)) = Abst (alpha, applySubstToTerm sigma exp)
  | applySubstToTerm sigma (Fun (f,exp)) = Fun (f, applySubstToTerm sigma exp)
  | applySubstToTerm sigma (Tuple exps) = Tuple (L.map (applySubstToTerm sigma) exps)

(******* set of atom constants ********)

fun atomConstsInAtom (Ca a) = [a]
  | atomConstsInAtom (Xa x) = []

fun atomConstsInPerm Id = []
  | atomConstsInPerm (Xp str) = []
  | atomConstsInPerm (Swap (v,w)) = LU.union (atomConstsInAtomic v, atomConstsInAtomic w)
  | atomConstsInPerm (Inv Pi) = atomConstsInPerm Pi
  | atomConstsInPerm (Comp (Pi,Psi)) = LU.union (atomConstsInPerm Pi, atomConstsInPerm Psi)
and atomConstsInAtomic (Simp alpha) = atomConstsInAtom alpha
  | atomConstsInAtomic (Mod (Psi,v)) = LU.union (atomConstsInPerm Psi, atomConstsInAtomic v)

fun atomConstsInTerm (Atomic v) = atomConstsInAtomic v
  | atomConstsInTerm (Susp (Psi,X)) = atomConstsInPerm Psi
  | atomConstsInTerm (Abst (v,exp)) = LU.union (atomConstsInAtomic v, atomConstsInTerm exp)
  | atomConstsInTerm (Fun (f,exp)) = atomConstsInTerm exp
  | atomConstsInTerm (Tuple exps) = LU.unionAll (L.map (atomConstsInTerm) exps)

(******* set of atom variables ********)

fun atomVarsInAtom (Ca a) = []
  | atomVarsInAtom (Xa x) = [x]

fun atomVarsInPerm Id = []
  | atomVarsInPerm (Xp str) = []
  | atomVarsInPerm (Swap (v,w)) = LU.union (atomVarsInAtomic v, atomVarsInAtomic w)
  | atomVarsInPerm (Inv Pi) = atomVarsInPerm Pi
  | atomVarsInPerm (Comp (Pi,Psi)) = LU.union (atomVarsInPerm Pi, atomVarsInPerm Psi)
and atomVarsInAtomic (Simp alpha) = atomVarsInAtom alpha
  | atomVarsInAtomic (Mod (Psi,v)) = LU.union (atomVarsInPerm Psi, atomVarsInAtomic v)

fun atomVarsInTerm (Atomic v) = atomVarsInAtomic v
  | atomVarsInTerm (Susp (Psi,X)) = atomVarsInPerm Psi
  | atomVarsInTerm (Abst (v,exp)) = LU.union (atomVarsInAtomic v, atomVarsInTerm exp)
  | atomVarsInTerm (Fun (f,exp)) = atomVarsInTerm exp
  | atomVarsInTerm (Tuple exps) = LU.unionAll (L.map (atomVarsInTerm) exps)

(******** generation of fresh atom vars *****)

fun newAtomVars xs n = SU.newNames ["A","B", "C", "D", "E"] xs n
fun newAtomVar xs = SU.newName ["A","B", "C", "D", "E"] xs

(******* set of permutation variables ********)

fun permVarsInAtom (Ca a) = []
  | permVarsInAtom (Xa x) = []

fun permVarsInPerm Id = []
  | permVarsInPerm (Xp str) = [str]
  | permVarsInPerm (Swap (v,w)) = LU.union (permVarsInAtomic v, permVarsInAtomic w)
  | permVarsInPerm (Inv Pi) = permVarsInPerm Pi
  | permVarsInPerm (Comp (Pi,Psi)) = LU.union (permVarsInPerm Pi, permVarsInPerm Psi)
and permVarsInAtomic (Simp alpha) = permVarsInAtom alpha
  | permVarsInAtomic (Mod (Psi,v)) = LU.union (permVarsInPerm Psi, permVarsInAtomic v)

fun permVarsInTerm (Atomic v) = permVarsInAtomic v
  | permVarsInTerm (Susp (Psi,X)) = permVarsInPerm Psi
  | permVarsInTerm (Abst (v,exp)) = LU.union (permVarsInAtomic v, permVarsInTerm exp)
  | permVarsInTerm (Fun (f,exp)) = permVarsInTerm exp
  | permVarsInTerm (Tuple exps) = LU.unionAll (L.map (permVarsInTerm) exps)

 (******* set of (outermost) moderated atomic expression ********)

fun moderatedAtomicInPerm Id = []
  | moderatedAtomicInPerm (Xp str) = []
  | moderatedAtomicInPerm (Swap (v,w)) = LU.union (moderatedAtomicInAtomic v, moderatedAtomicInAtomic w)
  | moderatedAtomicInPerm (Inv Pi) = moderatedAtomicInPerm Pi
  | moderatedAtomicInPerm (Comp (Pi,Psi)) = LU.union (moderatedAtomicInPerm Pi, moderatedAtomicInPerm Psi)
and moderatedAtomicInAtomic (Simp alpha) = []
  | moderatedAtomicInAtomic (Mod (Psi,v)) = [Mod (Psi,v)]

fun moderatedAtomicInTerm (Atomic v) = moderatedAtomicInAtomic v
  | moderatedAtomicInTerm (Susp (Psi,X)) = moderatedAtomicInPerm Psi
  | moderatedAtomicInTerm (Abst (v,exp)) = LU.union (moderatedAtomicInAtomic v, moderatedAtomicInTerm exp)
  | moderatedAtomicInTerm (Fun (f,exp)) = moderatedAtomicInTerm  exp
  | moderatedAtomicInTerm (Tuple exps) = LU.unionAll (L.map (moderatedAtomicInTerm) exps)

(******* set of term variables ********)

fun varsInTerm (Atomic v) = []
  | varsInTerm (Susp (Psi,X)) = [X]
  | varsInTerm (Abst (v,exp)) = varsInTerm exp
  | varsInTerm (Fun (f,exp)) = varsInTerm exp
  | varsInTerm (Tuple exps) = LU.unionAll (L.map (varsInTerm) exps)

(**** atom identity problem ***)
datatype identity_problem = I_Eq of atom_expr * atom_expr (* solved forms do not have this case *)
			  | I_Neq of atom_expr * atom_expr
		          | I_Sub of string * atom_expr  (* string for atom variable *)
                          | I_Pmap of string * atom_expr * atom_expr (* string for permunation variable *)
                          (* | I_Pnmap of string * atom_expr * atom_expr  *)(* string for permunation variable *)
                          | I_Fresh of atom_expr * var           
                          | I_Dfresh of var * perm_expr * perm_expr   (* delayed fresh *)
                          | I_Vmap of var * term_expr   (* solved term variable *)

(* fun isSolvedIdProblem xs = L.all (fn x=> case x of I_Eq _ => false | _ => true) xs *)

fun prIdProblem (I_Eq (alpha,beta)) = (prAtom alpha) ^ "==" ^ (prAtom beta)
  | prIdProblem (I_Neq (alpha,beta)) = (prAtom alpha) ^ "!=" ^ (prAtom beta)
  | prIdProblem (I_Sub (str,alpha)) = (str ^ "|->" ^ (prAtom alpha))
  | prIdProblem (I_Pmap (str,alpha,beta)) = (str ^ ":" ^ (prAtom alpha) ^ "|->" ^ (prAtom beta))
  (* | prIdProblem (I_Pnmap (str, alpha,beta)) = (str ^ ":" ^ (prAtom alpha) ^ "|-/->" ^ (prAtom beta)) *)
  | prIdProblem (I_Fresh (alpha, X)) = (prAtom alpha ^ "#" ^ (Var.toString X))
  | prIdProblem (I_Dfresh (X,Pi,Psi)) = (prPerm Pi) ^ "." ^ (Var.toString X) 
					  ^ "==" ^ (prPerm Psi) ^ "." ^ (Var.toString X)
  | prIdProblem (I_Vmap (X, s)) = (Var.toString X) ^ "|->" ^ (prTerm s)

fun updateIdProblem sigma (I_Eq (alpha,beta)) = I_Eq (updateAtom sigma alpha, updateAtom sigma beta)
  | updateIdProblem sigma (I_Neq (alpha,beta)) = I_Neq (updateAtom sigma alpha, updateAtom sigma beta)
  | updateIdProblem sigma (I_Sub (str, alpha)) = I_Sub (str, updateAtom sigma alpha)
  | updateIdProblem sigma (I_Pmap (str,alpha,beta)) = I_Pmap (str, updateAtom sigma alpha, updateAtom sigma beta)
  (* | updateIdProblem sigma (I_Pnmap (str,alpha,beta)) = I_Pnmap (str, updateAtom sigma alpha, updateAtom sigma beta) *)
  | updateIdProblem sigma (I_Fresh (alpha,X)) = I_Fresh (updateAtom sigma alpha, X)
  | updateIdProblem sigma (I_Dfresh (X,Pi,Psi)) = I_Dfresh (X, updatePerm sigma Pi, updatePerm sigma Psi)
  | updateIdProblem sigma (I_Vmap (X, s)) = I_Vmap (X, updateTerm sigma s)

fun simplifyIdProblem (I_Eq (alpha,beta)) = I_Eq (simplifyAtom alpha, simplifyAtom beta)
  | simplifyIdProblem (I_Neq (alpha,beta)) = I_Neq (simplifyAtom alpha, simplifyAtom beta)
  | simplifyIdProblem (I_Sub (str, alpha)) = I_Sub (str, simplifyAtom alpha)
  | simplifyIdProblem (I_Pmap (str,alpha,beta)) = I_Pmap (str, simplifyAtom alpha, simplifyAtom beta)
  (* | simplifyIdProblem (I_Pnmap (str,alpha,beta)) = I_Pnmap (str, simplifyAtom alpha, simplifyAtom beta) *)
  | simplifyIdProblem (I_Fresh (alpha,X)) = I_Fresh (simplifyAtom alpha, X)
  | simplifyIdProblem (I_Dfresh (X,Pi,Psi)) = I_Dfresh (X, simplifyPerm Pi, simplifyPerm Psi)
  | simplifyIdProblem (I_Vmap (X, s)) = I_Vmap (X, simplifyTerm s)

fun simplifyIdProblemByAssign rho (I_Vmap (X, s)) = I_Vmap (X, simplifyTermByAssign rho s)
  | simplifyIdProblemByAssign rho (I_Dfresh (X,Pi,Psi)) = I_Dfresh (X, simplifyPermByAssign rho Pi, simplifyPermByAssign rho Psi)
  | simplifyIdProblemByAssign _ x = x 

fun permVarsInIdProblem (I_Eq (alpha,beta)) = LU.union (permVarsInAtom alpha, permVarsInAtom beta)
  | permVarsInIdProblem (I_Neq (alpha,beta)) = LU.union (permVarsInAtom alpha, permVarsInAtom beta)
  | permVarsInIdProblem (I_Sub (str, alpha)) = permVarsInAtom alpha
  | permVarsInIdProblem (I_Pmap (str,alpha,beta)) = LU.add str (LU.union (permVarsInAtom alpha, permVarsInAtom beta))
  (* | permVarsInIdProblem (I_Pnmap (str,alpha,beta)) = LU.add str (LU.union (permVarsInAtom alpha, permVarsInAtom beta)) *)
  | permVarsInIdProblem (I_Fresh (alpha,X)) = permVarsInAtom alpha
  | permVarsInIdProblem (I_Dfresh (X,Pi,Psi)) = LU.union (permVarsInPerm Pi, permVarsInPerm Psi)
  | permVarsInIdProblem (I_Vmap (X, s)) = permVarsInTerm s

fun atomVarsInIdProblem (I_Eq (alpha,beta)) = LU.union (atomVarsInAtom alpha, atomVarsInAtom beta)
  | atomVarsInIdProblem (I_Neq (alpha,beta)) = LU.union (atomVarsInAtom alpha, atomVarsInAtom beta)
  | atomVarsInIdProblem (I_Sub (str, alpha)) = LU.add str (atomVarsInAtom alpha)
  | atomVarsInIdProblem (I_Pmap (str,alpha,beta)) = LU.union (atomVarsInAtom alpha, atomVarsInAtom beta)
  (* | atomVarsInIdProblem (I_Pnmap (str,alpha,beta)) = LU.union (atomVarsInAtom alpha, atomVarsInAtom beta) *)
  | atomVarsInIdProblem (I_Fresh (alpha,X)) = atomVarsInAtom alpha
  | atomVarsInIdProblem (I_Dfresh (X,Pi,Psi)) = LU.union (atomVarsInPerm Pi, atomVarsInPerm Psi)
  | atomVarsInIdProblem (I_Vmap (X, s)) = atomVarsInTerm s

fun unsolvedAtomVarsInIdProblem (I_Eq (alpha,beta)) = LU.union (atomVarsInAtom alpha, atomVarsInAtom beta)
  | unsolvedAtomVarsInIdProblem (I_Neq (alpha,beta)) = LU.union (atomVarsInAtom alpha, atomVarsInAtom beta)
  | unsolvedAtomVarsInIdProblem (I_Sub (str, alpha)) = atomVarsInAtom alpha
  | unsolvedAtomVarsInIdProblem (I_Pmap (str,alpha,beta)) = LU.union (atomVarsInAtom alpha, atomVarsInAtom beta)
  (* | unsolvedAtomVarsInIdProblem (I_Pnmap (str,alpha,beta)) = LU.union (atomVarsInAtom alpha, atomVarsInAtom beta) *)
  | unsolvedAtomVarsInIdProblem (I_Fresh (alpha,X)) = atomVarsInAtom alpha
  | unsolvedAtomVarsInIdProblem (I_Dfresh (X,Pi,Psi)) = LU.union (atomVarsInPerm Pi, atomVarsInPerm Psi)
  | unsolvedAtomVarsInIdProblem (I_Vmap (X, s)) = atomVarsInTerm s

fun atomConstsInIdProblem (I_Eq (alpha,beta)) = LU.union (atomConstsInAtom alpha, atomConstsInAtom beta)
  | atomConstsInIdProblem (I_Neq (alpha,beta)) = LU.union (atomConstsInAtom alpha, atomConstsInAtom beta)
  | atomConstsInIdProblem (I_Sub (str, alpha)) = atomConstsInAtom alpha
  | atomConstsInIdProblem (I_Pmap (str,alpha,beta)) = LU.union (atomConstsInAtom alpha, atomConstsInAtom beta)
  (* | atomConstsInIdProblem (I_Pnmap (str,alpha,beta)) = LU.union (atomConstsInAtom alpha, atomConstsInAtom beta) *)
  | atomConstsInIdProblem (I_Fresh (alpha,X)) = atomConstsInAtom alpha
  | atomConstsInIdProblem (I_Dfresh (X,Pi,Psi)) = LU.union (atomConstsInPerm Pi, atomConstsInPerm Psi)
  | atomConstsInIdProblem (I_Vmap (X, s)) = atomConstsInTerm s
 
(*******************************************)
(** Atomic Identity Elimination Procedure **)
(*******************************************)

(* replace I_Eq to I_Subst, and possibly remove some I_Eq *)
fun elimEq ([],ys) = SOME (rev ys)
 |  elimEq (x::xs,ys) = 
    case x of (I_Eq (alpha, beta)) =>
	      (case (alpha,beta) of 
		   (Ca a, Ca b) => if a = b then elimEq (xs,ys) else NONE
		 | (Ca a, Xa x) => replaceWithSub (x, Ca a) (xs,ys)
		 | (Xa x, Ca a) => replaceWithSub (x, Ca a) (xs,ys)
		 | (Xa x, Xa y) => if x = y then elimEq (xs,ys) else replaceWithSub (x, Xa y) (xs,ys))
	    | _ => elimEq (xs,x::ys)
and replaceWithSub (str,alpha) (xs,ys) = 
    let val zs = I_Sub (str, alpha)::
		 (L.map (updateIdProblem (AL.singleton (str,alpha))) (rev ys @ xs))
    in elimEq (zs,[]) end
	
(* possibly replace I_Pmap with I_Eq *)
fun possiblyReplacePmap (str,alpha,beta) (x as (I_Pmap (str2,alpha2,beta2))) = 
    if str = str2 
    then if alpha = alpha2 then I_Eq (beta, beta2)
	 else if beta = beta2 then I_Eq (alpha,alpha2)
	 else x
    else x
  | possiblyReplacePmap (str,alpha,beta) x = x  (* otherwise *)

(** possibly replace I_Pmap with I_Eq **)
fun pmapElim ([],ys) = ys
  | pmapElim (x::xs, ys) = 
    case x of I_Pmap (str,alpha,beta) => pmapElim (L.map (possiblyReplacePmap (str,alpha,beta)) xs, x::ys)
	    | _ => pmapElim (xs, x::ys)
			     
(** check if I_Eq is not contained in the problem xs **)
fun containsNoEq xs = L.all (fn x => case x of I_Eq _ => false | _ => true) xs

(** Atomic Equality Elimination Procedure **)
fun procedureAeqElim xs = 
    case elimEq (xs,[]) of
	NONE => NONE
      | SOME ys => let val zs = pmapElim (ys,[])
		   in if containsNoEq zs
		      then SOME zs
		      else procedureAeqElim zs
		   end

(*********************************************)
(** Atomic Disequality Saturation Procedure **)
(*********************************************)

(* possibly remove some I_Neq *)
fun elimNeq ([],ys) = SOME (rev ys)
 |  elimNeq (x::xs,ys) = 
    case x of (I_Neq (alpha, beta)) =>
	      (case (alpha,beta) of 
		   (Ca a, Ca b) => if a = b then NONE else elimNeq (xs,ys)
		 | _ => elimNeq (xs,x::ys))
	    | _ => elimNeq (xs,x::ys)

(* possibly replace I_Pnmap with I_Neq *)
(* fun possiblyReplacePnmap (str,alpha,beta) (x as (I_Pnmap (str2,alpha2,beta2))) =  *)
(*     if str = str2  *)
(*     then if alpha = alpha2 then I_Neq (beta, beta2) *)
(* 	 else if beta = beta2 then I_Neq (alpha,alpha2) *)
(* 	 else x *)
(*     else x *)
(*   | possiblyReplacePnmap (str,alpha,beta) x = x  (* otherwise *) *)

(** possibly replace I_Pmap with I_Eq **)
(* fun elimPnmap xs = let val pmaps = L.mapPartial (fn x => case x of I_Pmap k => SOME k | _ => NONE) xs *)
(* 		   in L.foldl (fn (k,ys) => L.map (possiblyReplacePnmap k) ys) *)
(* 			      xs *)
(* 			      pmaps *)
(* 		   end *)

fun pmapImages (str,alpha) xs = 
    L.mapPartial (fn x => case x of 
			      I_Pmap (str2,alpha2,beta2) => if str = str2 andalso alpha = alpha2
							    then SOME beta2 else NONE
			    | _ => NONE) xs
		    
fun getAllPossibleDisequalities neqs xs = 
    let val pvars = LU.mapUnion permVarsInIdProblem xs
    in LU.toSet (LU.mapAppend (fn (alpha,beta) => 
				  LU.mapAppend (fn str => 
						   let val aouts = pmapImages (str,alpha) xs
						       val bouts = pmapImages (str,beta) xs
						   in ListXProd.mapX (fn xy => xy) (aouts,bouts)
						   end)
					       pvars)
			      neqs)
    end

(** Atomic Disequality Saturation Procedure **)
fun procedureDiseqSat xs = 
    let fun procedureDiseqSatRest NONE = NONE
	  | procedureDiseqSatRest (SOME ys) =
	    let val neqs = L.mapPartial (fn x=> case x of I_Neq ab => SOME ab | _ => NONE) ys
		val outs = getAllPossibleDisequalities neqs xs
		val new = LU.differenceByAll (outs,neqs)
		val check = elimNeq (L.map (fn ab => I_Neq ab) new, [])
	    in case check of 
		   NONE => NONE
		 | SOME new2 => if null new2 then SOME ys
				else procedureDiseqSat (ys @ new2)
	    end
(*    in procedureDiseqSatRest (elimNeq (elimPnmap xs, [])) *)
    in procedureDiseqSatRest (elimNeq (xs, []))
    end

(********************************************)
(** Atom Identity Problem Solving Procedure **)
(********************************************)

(** remove all I_Eq, remove possible I_Pmap,I_Pnmap, generate all possible I_Neq, and saturate **)
local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in
fun solveIdProblem xs = 
    let
	val _ = dprintln ("IdentitySolve: input: " ^ (LU.toStringCommaSquare prIdProblem xs))
	val ans = procedureAeqElim xs
	val _ = case ans of
     		    SOME ys => dprintln ("IdentitySolve (eq elim): " ^ (LU.toStringCommaSquare prIdProblem ys))
		  | NONE => dprintln ("IdentitySolve (eq elim): NONE")

	val ans2 = case ans of
		       NONE => NONE
		     | SOME ys => procedureDiseqSat ys

	(* eliminate redundant neq *)
	val ans3 = case ans2 of 
		       NONE => NONE 
		     | SOME ys => SOME (LU.toSet' (fn (x,y)=> case (x,y) of
								  (I_Neq (x1,y1),I_Neq (x2,y2)) => (x1 = x2 andalso y1 = y2)
												   orelse (x1 = y2 andalso x2 = y1)
								| _ => x = y)
						  ys)

	val _ = case ans3 of
     		    SOME ys => dprintln ("IdentitySolve (neq sat): output: " ^ (LU.toStringCommaSquare prIdProblem ys))
		  | NONE => dprintln ("IdentitySolve (neq sat): output: NONE")
    in ans3
    end
end

val _ = TU.test "solveIdProblem"
		[ (solveIdProblem [I_Pmap ("P", Ca (Atom.fromString "a"), Ca (Atom.fromString "b"))]
		   = SOME ([ I_Pmap ("P", Ca (Atom.fromString "a"), Ca (Atom.fromString "b"))])),
		  (solveIdProblem [ I_Eq (Ca (Atom.fromString "a"), Ca (Atom.fromString "b")) ] 
		   = NONE),
		  (solveIdProblem [ I_Eq (Ca (Atom.fromString "a"), Ca (Atom.fromString "a")) ] 
		   = SOME []),
		  (solveIdProblem [I_Pmap ("P", Xa "A", Ca (Atom.fromString "a")), I_Pmap ("P", Xa "A", Ca (Atom.fromString "b"))]
		   = NONE),
		  (solveIdProblem [I_Pmap ("P", Xa "A", Ca (Atom.fromString "a")), I_Pmap ("P", Xa "B", Ca (Atom.fromString "b")),
				   I_Eq (Xa "A", Ca (Atom.fromString "a"))]
		   = (SOME [I_Pmap ("P",Xa "B",Ca (Atom.fromString "b")),  I_Pmap ("P",Ca (Atom.fromString "a"),Ca (Atom.fromString "a")), 
			    I_Sub ("A",Ca (Atom.fromString "a"))])),
		  (solveIdProblem [ I_Pmap ("P", Xa "A",Ca (Atom.fromString "a")), I_Pmap ("P",Xa "B", Ca (Atom.fromString "b")), 
				    I_Eq (Xa "A",Xa "B") ]
		   = NONE),
		  (solveIdProblem [ I_Pmap ("P", Xa "A",Ca (Atom.fromString "a")), I_Pmap ("P",Xa "B", Xa "B"), I_Eq (Xa "A",Xa "B") ]
		   = SOME [I_Sub ("A",Ca (Atom.fromString "a")),I_Pmap ("P",Ca (Atom.fromString "a"),Ca (Atom.fromString "a")), 
			   I_Sub ("B",Ca (Atom.fromString "a"))]),
		  (solveIdProblem [ I_Pmap ("P", Xa "A",Ca (Atom.fromString "a")), I_Pmap ("P",Xa "B", Xa "B"), I_Neq (Xa "A",Xa "B") ]
		   = SOME [I_Neq (Xa "A",Xa "B"),I_Pmap ("P",Xa "B",Xa "B"), I_Pmap ("P",Xa "A",Ca (Atom.fromString "a")),
			   I_Neq (Ca (Atom.fromString "a"),Xa "B")]),
		  (* (solveIdProblem [ I_Pnmap ("P", Xa "A",Ca (Atom.fromString "a")), I_Pmap ("P",Xa "B", Xa "B"), I_Neq (Xa "A",Xa "B") ] *)
		  (*  = SOME [I_Neq (Xa "A",Xa "B"),I_Pmap ("P",Xa "B",Xa "B"), I_Pnmap ("P",Xa "A",Ca (Atom.fromString "a"))]), *)
		  (* (solveIdProblem [ I_Pnmap ("P", Xa "A",Ca (Atom.fromString "a")), I_Pnmap ("P",Xa "B", Xa "B"), I_Eq (Xa "A",Xa "B") ] *)
		  (*  = SOME  [I_Pnmap ("P",Xa "B",Xa "B"),I_Pnmap ("P",Xa "B",Ca (Atom.fromString "a")), I_Sub ("A",Xa "B")]), *)
		  (* (solveIdProblem [ I_Pmap ("P", Xa "A",Ca (Atom.fromString "a")), I_Pnmap ("P",Xa "B", Xa "B"),  *)
		  (* 		    I_Eq (Xa "A",Xa "B"),  I_Pmap ("P",Xa "B", Ca (Atom.fromString "a"))] *)
		  (*  = SOME  [I_Sub ("A",Xa "B"),I_Pmap ("P",Xa "B",Ca (Atom.fromString "a")), I_Neq (Ca (Atom.fromString "a"),Xa "B")]), *)
		  (solveIdProblem [ I_Pmap ("P", Ca (Atom.fromString "b"), Ca (Atom.fromString "b")),
				    I_Eq (Ca (Atom.fromString "a"), Xa "A"),  I_Eq (Xa "A", Ca (Atom.fromString "a"))]
		   = SOME [I_Pmap ("P",Ca (Atom.fromString "b"),Ca (Atom.fromString "b")),I_Sub ("A",Ca (Atom.fromString "a"))])
		  ]


(****++*************************************************)
(** Simplified Atomic (Dis)Equality Solving Procedure **)
(*******************************************************)

(* simplified atomic problem *)
datatype atomic_problem0 = A0_Geq of atomic_expr * atom_expr
			 | A0_Gneq of atomic_expr * atom_expr
			 | A0_Eq of atom_expr * atom_expr
			 | A0_Neq of atom_expr * atom_expr
			 | A0_Pmap of string * atom_expr * atom_expr
(*			 | A0_Pnmap of string * atom_expr * atom_expr *)
			 | A0_Fresh of atom_expr * var 
			 | A0_Dfresh of var * perm_expr * perm_expr  (* delayed fresh *)
			 | A0_Vmap of var * term_expr   (* solved term variable *)

fun prAtomicProblem0 (A0_Geq (v, alpha)) = (prAtomic v) ^ "==" ^ (prAtom alpha)
  | prAtomicProblem0 (A0_Gneq (v, alpha)) = (prAtomic v) ^ "!=" ^ (prAtom alpha)
  | prAtomicProblem0 (A0_Eq (alpha, beta)) = (prAtom alpha) ^ "==" ^ (prAtom beta)
  | prAtomicProblem0 (A0_Neq (alpha, beta)) = (prAtom alpha) ^ "!=" ^ (prAtom beta)
  | prAtomicProblem0 (A0_Pmap (str,alpha,beta)) = (str ^ ":" ^ (prAtom alpha) ^ "|->" ^ (prAtom beta))
(*  | prAtomicProblem0 (A0_Pnmap (str, alpha,beta)) = (str ^ ":" ^ (prAtom alpha) ^ "|-/->" ^ (prAtom beta)) *)
  | prAtomicProblem0 (A0_Fresh (alpha, X)) = (prAtom alpha ^ "#" ^ (Var.toString X))
  | prAtomicProblem0 (A0_Dfresh (X,Pi,Psi)) = (prPerm Pi) ^ "." ^ (Var.toString X) 
					  ^ "==" ^ (prPerm Psi) ^ "." ^ (Var.toString X)
  | prAtomicProblem0 (A0_Vmap (X, s)) = (Var.toString X) ^ "|->" ^ (prTerm s)

fun atomVarsInAtomicProblem0 (A0_Geq (v, alpha)) = LU.union (atomVarsInAtomic v, atomVarsInAtom alpha)
  | atomVarsInAtomicProblem0 (A0_Gneq (v, alpha)) = LU.union (atomVarsInAtomic v, atomVarsInAtom alpha)
  | atomVarsInAtomicProblem0 (A0_Eq (alpha, beta)) = LU.union (atomVarsInAtom alpha, atomVarsInAtom beta)
  | atomVarsInAtomicProblem0 (A0_Neq (alpha, beta)) = LU.union (atomVarsInAtom alpha, atomVarsInAtom beta)
  | atomVarsInAtomicProblem0 (A0_Pmap (str, alpha, beta)) = LU.union (atomVarsInAtom alpha, atomVarsInAtom beta)
(*  | atomVarsInAtomicProblem0 (A0_Pnmap (str,alpha,beta)) = LU.union (atomVarsInAtom alpha, atomVarsInAtom beta) *)
  | atomVarsInAtomicProblem0 (A0_Fresh (alpha,X)) = atomVarsInAtom alpha
  | atomVarsInAtomicProblem0 (A0_Dfresh (X,Pi,Psi)) = LU.union (atomVarsInPerm Pi, atomVarsInPerm Psi)
  | atomVarsInAtomicProblem0 (A0_Vmap (X, s)) = atomVarsInTerm s

(*
fun atomConstsInAtomicProblem0 (A0_Geq (v, alpha)) = LU.union (atomConstsInAtomic v, atomConstsInAtom alpha)
  | atomConstsInAtomicProblem0 (A0_Gneq (v, alpha)) = LU.union (atomConstsInAtomic v, atomConstsInAtom alpha)
  | atomConstsInAtomicProblem0 (A0_Eq (alpha, beta)) = LU.union (atomConstsInAtom alpha, atomConstsInAtom beta)
  | atomConstsInAtomicProblem0 (A0_Neq (alpha, beta)) = LU.union (atomConstsInAtom alpha, atomConstsInAtom beta)
  | atomConstsInAtomicProblem0 (A0_Pmap (str, alpha, beta)) = LU.union (atomConstsInAtom alpha, atomConstsInAtom beta)
  | atomConstsInAtomicProblem0 (A0_Pnmap (str,alpha,beta)) = LU.union (atomConstsInAtom alpha, atomConstsInAtom beta)
  | atomConstsInAtomicProblem0 (A0_Fresh (alpha,X)) = atomConstsInAtom alpha
  | atomConstsInAtomicProblem0 (A0_Dfresh (X,Pi,Psi)) = LU.union (atomConstsInPerm Pi, atomConstsInPerm Psi)
  | atomConstsInAtomicProblem0 (A0_Vmap (X, s)) = atomConstsInTerm s
*)

exception ConvertErrorAtomicProblem0toIdProblem

fun convertAtomicProblem0toIdProblem (A0_Geq (v, alpha)) = raise ConvertErrorAtomicProblem0toIdProblem
  | convertAtomicProblem0toIdProblem (A0_Gneq (v, alpha)) = raise ConvertErrorAtomicProblem0toIdProblem
  | convertAtomicProblem0toIdProblem (A0_Eq (alpha, beta)) = I_Eq (alpha, beta)
  | convertAtomicProblem0toIdProblem (A0_Neq (alpha, beta)) = I_Neq (alpha, beta)
  | convertAtomicProblem0toIdProblem (A0_Pmap (str, alpha, beta)) = I_Pmap (str, alpha, beta)
(*   | convertAtomicProblem0toIdProblem (A0_Pnmap (str,alpha,beta)) = I_Pnmap (str,alpha,beta) *)
  | convertAtomicProblem0toIdProblem (A0_Fresh (alpha, X)) = I_Fresh (alpha, X)
  | convertAtomicProblem0toIdProblem (A0_Dfresh (X,Pi,Psi)) =  I_Dfresh (X,Pi,Psi)
  | convertAtomicProblem0toIdProblem (A0_Vmap (X, s)) = I_Vmap (X,s)

fun replaceAtomicInAtomicProblem0 (u,gamma) (A0_Geq (v, alpha)) = A0_Geq (replaceAtomicInAtomic (u,gamma) v, alpha)
  | replaceAtomicInAtomicProblem0 (u,gamma) (A0_Gneq (v, alpha)) = A0_Gneq (replaceAtomicInAtomic (u,gamma) v, alpha)
  | replaceAtomicInAtomicProblem0 (u,gamma) (A0_Eq (alpha, beta)) = A0_Eq (alpha, beta)
  | replaceAtomicInAtomicProblem0 (u,gamma) (A0_Neq (alpha, beta)) = A0_Neq (alpha, beta)
  | replaceAtomicInAtomicProblem0 (u,gamma) (A0_Pmap (str, alpha, beta)) = A0_Pmap (str, alpha, beta)
(*   | replaceAtomicInAtomicProblem0 (u,gamma) (A0_Pnmap (str,alpha,beta)) = A0_Pnmap (str,alpha,beta) *)
  | replaceAtomicInAtomicProblem0 (u,gamma) (A0_Fresh (alpha, X)) = A0_Fresh (alpha, X)
  | replaceAtomicInAtomicProblem0 (u,gamma) (A0_Dfresh (X,Pi,Psi)) =  A0_Dfresh (X,replaceAtomicInPerm (u,gamma) Pi,replaceAtomicInPerm (u,gamma) Psi)
  | replaceAtomicInAtomicProblem0 (u,gamma) (A0_Vmap (X, s)) = A0_Vmap (X,replaceAtomicInTerm (u,gamma) s)
									  
(* decompose by depth first *)
fun decomposeLHS0 avars (A0_Geq (Simp alpha,beta)) = 
    if L.all isAtomConst [alpha,beta] 
    then if alpha = beta  then [[]] else []
    else [[A0_Eq (alpha, beta)]]
  | decomposeLHS0 avars (A0_Geq (Mod (Psi, Simp alpha),beta)) = 
    (case Psi of 
	 Id => if L.all isAtomConst [alpha,beta] 
	       then if alpha = beta  then [[]] else []
	       else [[A0_Eq (alpha, beta)]]
       | Xp str => [[A0_Pmap (str,alpha,beta)]]
       | Swap (v,w) =>  L.mapPartial (fn x=>x) 
				     [ SOME [A0_Geq (v,alpha), A0_Geq (w,beta)],
				       SOME [A0_Geq (w,alpha), A0_Geq (v,beta)],
				       (if L.all isAtomConst [alpha,beta] 
					then if alpha = beta  then SOME [A0_Gneq (v,alpha), A0_Gneq(w,alpha)] else NONE
					else SOME [A0_Gneq (v,alpha), A0_Gneq(w,alpha), A0_Eq (alpha,beta)])
				     ]
       | Inv Pi => [[A0_Geq (Mod (Pi,Simp beta),alpha)]]
       | Comp (Pi1,Pi2) => let val gamma = Xa (newAtomVar avars)
			   in [[A0_Geq (Mod (Pi2, Simp alpha),gamma), A0_Geq (Mod (Pi1,Simp gamma), beta)]]
			   end)
  | decomposeLHS0 avars (A0_Geq (Mod (Psi, v),beta)) = 
    let val gamma = Xa (newAtomVar avars)
    in [[A0_Geq (v, gamma), A0_Geq (Mod (Psi,Simp gamma), beta)]]
    end
  | decomposeLHS0 avars (A0_Gneq (Simp alpha,beta)) = 
    if L.all isAtomConst [alpha,beta] 
    then if alpha = beta  then [] else [[]]
    else [[A0_Neq (alpha, beta)]]
  | decomposeLHS0 avars (A0_Gneq (Mod (Psi,Simp alpha),beta)) = 
    (case Psi of 
	 Id => if L.all isAtomConst [alpha,beta] 
	       then if alpha = beta  then [] else [[]]
	       else [[A0_Neq (alpha, beta)]]
(*       | Xp str => [[A0_Pnmap (str,alpha,beta)]] *)
       | Xp str => let val gamma = Xa (newAtomVar avars)  (* revised 2016/01/02 !!! not reflected in slides *)
		   in [[A0_Pmap (str,alpha,gamma), A0_Neq (beta, gamma)]]
		   end
       | Swap (v,w) => [[A0_Geq (v,alpha), A0_Gneq (w,beta)],
			[A0_Geq (w,alpha), A0_Gneq (v,beta)],
			[A0_Gneq (v,alpha), A0_Gneq(w,alpha), A0_Neq (alpha,beta)]]
       | Inv Pi => [[A0_Gneq (Mod (Pi, Simp beta),alpha)]]
       | Comp (Pi1,Pi2) => let val gamma = Xa (newAtomVar avars)
			   in [[A0_Geq (Mod (Pi2, Simp alpha),gamma),A0_Gneq (Mod (Pi1,Simp gamma), beta)]]
			   end)
  | decomposeLHS0 avars (A0_Gneq (Mod (Psi, v),beta)) = 
    let val gamma = Xa (newAtomVar avars)
    in [[A0_Geq (v, gamma), A0_Gneq (Mod (Psi,Simp gamma), beta)]]
    end
  | decomposeLHS0 avars x = [[x]]

fun decomposeLHS1 avars ([],ys) = (true, [ys])
  | decomposeLHS1 avars (x::xs,ys) = 
    (case x of A0_Geq (v,alpha) 
	       => let val zss = decomposeLHS0 avars x
		      val xs2 = L.map (replaceAtomicInAtomicProblem0 (v,alpha)) xs
		      val ys2 = L.map (replaceAtomicInAtomicProblem0 (v,alpha)) ys
		  in (false, L.map (fn zs => (zs @ xs2 @ ys2)) zss) (* ......change order ... *)
		  end 
	     | A0_Gneq _ => let val zss = decomposeLHS0 avars x
			    in (false, L.map (fn zs => (zs @ xs @ ys)) zss) (*......change order ....... *)
			    end 
	     | _ => decomposeLHS1 avars (xs,x::ys))

fun decomposeLHS [] =  []
  | decomposeLHS (xs::xss) = 
    let val avars = LU.mapAppend atomVarsInAtomicProblem0 xs
    in case decomposeLHS1 avars (xs,[]) of
	   (true,yss) => yss @ (decomposeLHS xss)
	 | (false,yss) => decomposeLHS (yss @ xss)
    end


(********************************************)
(** Atomic (Dis)Equality Solving Procedure **)
(********************************************)

datatype atomic_problem = A_Eq of atomic_expr * atomic_expr
			| A_Neq of atomic_expr * atomic_expr
			| A_Fresh of atomic_expr * var
			| A_Dfresh of var * perm_expr * perm_expr  (* delayed fresh *)
			| A_Vmap of var * term_expr   (* solved term variable *)

fun prAtomicProblem (A_Eq (v,w)) = (prAtomic v) ^ "==" ^ (prAtomic w)
  | prAtomicProblem (A_Neq (v,w)) = (prAtomic v) ^ "!=" ^ (prAtomic w)
  | prAtomicProblem (A_Fresh (v, X)) = (prAtomic v ^ "#" ^ (Var.toString X))
  | prAtomicProblem (A_Dfresh (X,Pi,Psi)) = (prPerm Pi) ^ "." ^ (Var.toString X) 
					    ^ "==" ^ (prPerm Psi) ^ "." ^ (Var.toString X)
  | prAtomicProblem (A_Vmap (X, s)) = (Var.toString X) ^ "|->" ^ (prTerm s)

fun simplifyAtomicProblem (A_Eq (v, w)) = A_Eq (simplifyAtomic v, simplifyAtomic w)
  | simplifyAtomicProblem (A_Neq (v, w)) = A_Neq (simplifyAtomic v, simplifyAtomic w)
  | simplifyAtomicProblem (A_Fresh (v, X)) = A_Fresh (simplifyAtomic v, X)
  | simplifyAtomicProblem (A_Dfresh (X, Pi, Psi)) = A_Dfresh (X, simplifyPerm Pi, simplifyPerm Psi)
  | simplifyAtomicProblem (A_Vmap (X, s)) = A_Vmap (X, simplifyTerm s)

fun atomVarsInAtomicProblem (A_Eq (v, w)) = LU.union (atomVarsInAtomic v, atomVarsInAtomic w)
  | atomVarsInAtomicProblem (A_Neq (v, w)) = LU.union (atomVarsInAtomic v, atomVarsInAtomic w)
  | atomVarsInAtomicProblem (A_Fresh (v, X)) = atomVarsInAtomic v
  | atomVarsInAtomicProblem (A_Dfresh (X, Pi, Psi)) = LU.union (atomVarsInPerm Pi, atomVarsInPerm Psi)
  | atomVarsInAtomicProblem (A_Vmap (X, s)) = atomVarsInTerm s

fun atomConstsInAtomicProblem (A_Eq (v, w)) = LU.union (atomConstsInAtomic v, atomConstsInAtomic w)
  | atomConstsInAtomicProblem (A_Neq (v, w)) = LU.union (atomConstsInAtomic v, atomConstsInAtomic w)
  | atomConstsInAtomicProblem (A_Fresh (v, X)) = atomConstsInAtomic v
  | atomConstsInAtomicProblem (A_Dfresh (X, Pi, Psi)) = LU.union (atomConstsInPerm Pi, atomConstsInPerm Psi)
  | atomConstsInAtomicProblem (A_Vmap (X, s)) = atomConstsInTerm s

(* convert atomic_problem to atomic_problem0 by fresh atom variables, also replace LHS of v#X with atom var *)
fun replaceRHS xs = 
    let val avars = LU.toSet (LU.mapAppend atomVarsInAtomicProblem xs)
	val num = L.length xs
	val freshes = newAtomVars avars num
    (* val _ = PU.println ("New vars: " ^  (LU.toStringCommaCurly (fn str=>str) freshes)) *)
    in LU.mapAppend (fn x => x)
		    (LP.map (fn (x,y) => case x of A_Eq (v,w) => ( case (v,w) of
								       (_, Simp beta) => [A0_Geq (v, beta)]
								     | (Simp alpha, _ ) => [A0_Geq (w, alpha)]
								     | _  => [A0_Geq (v, Xa y), A0_Geq (w, Xa y)])
						 | A_Neq (v,w) => ( case (v,w) of
									(_, Simp beta) => [A0_Gneq (v, beta)]
								      | (Simp alpha, _ ) => [A0_Gneq (w, alpha)]
								      | _  => [A0_Geq (v, Xa y), A0_Gneq (w, Xa y)])
						 | A_Fresh (v,X) => ( case v of
									  (Simp alpha) => [A0_Fresh (alpha, X)]
									| _ => [A0_Fresh (Xa y, X), A0_Geq (v, Xa y)])
						 | A_Dfresh (X,Pi,Psi) => [A0_Dfresh (X,Pi,Psi)] 
						 | A_Vmap (X,s) => [A0_Vmap (X,s)])       
			    (xs,freshes))  (*** the remainder of longer one (freshed) will be omitted ***)
    end

(* replace moderated atomic expr in subst by atomic var, to that they are also decomposed *)
(* slide not reflected !!! *)
fun replaceModeratedAtomicInSubstitution xs =
    let val avars = LU.mapAppend atomVarsInAtomicProblem0 xs
	val substituted = L.mapPartial (fn x => case x of A0_Vmap (X, s) => SOME s 
						(*	| A0_Dfresh (X,Pi,Psi) => SOME (Tuple [Susp (Pi,X), Susp(Psi,X)])
                                                 ---- this is redundant, but may be helpful for some case ...   *)
							| _ => NONE) xs
	val targets = LU.toSet (LU.mapAppend moderatedAtomicInTerm substituted)
	val num = L.length targets
	val freshes = newAtomVars avars num
	val addition = LP.map (fn (v,str) => A0_Geq (v, Xa str)) (targets,freshes)
	val replaced = LP.foldl (fn (v,str,ys) => L.map (replaceAtomicInAtomicProblem0 (v,Xa str)) ys)
				xs
				(targets,freshes)
    in addition @ replaced
    end

fun relevantAtoms [] (done, avars) = avars
  | relevantAtoms (x::yet) (done, avars) =
    let fun addornot xs = let val ys = LU.differenceByAll (xs,avars)
    			  in if null ys
    			     then relevantAtoms yet (x::done, avars)
    			     else relevantAtoms (L.revAppend (done,x::yet)) ([], avars @ ys)
    			  end
    in case x of
    	   I_Eq (alpha,beta) => relevantAtoms yet (x::done, avars)
    	 | I_Neq (alpha,beta) => relevantAtoms yet (x::done, avars)
    	 | I_Sub (str, alpha) => if LU.member str avars
				 then addornot (atomVarsInAtom alpha)
				 else relevantAtoms yet (x::done, avars)
    	 | I_Pmap (str,alpha,beta) => relevantAtoms yet (x::done, avars)
(*    	 | I_Pnmap (str,alpha,beta) => relevantAtoms yet (x::done, avars) *)
    	 | I_Fresh (alpha,X) => addornot (atomVarsInAtom alpha)
    	 | I_Dfresh (X,Pi,Psi) => addornot (LU.union (atomVarsInPerm Pi, atomVarsInPerm Psi))
    	 | I_Vmap (X, s) => addornot (atomVarsInTerm s)
    end 

(* avars: atomic variables appearing in original problem, obtained by relavantAtoms function *)
(* We check whether constraint is redundant for final answer *)
(* to remove assigments and constraints for fresh variables inserted *)
fun checkDomainOfIdProblem avars problem =
    let fun check2var (Ca _, Ca _) = true
	  | check2var (Xa x, Ca _) = LU.member x avars
	  | check2var (Ca _, Xa y) = LU.member y avars
	  | check2var  (Xa x, Xa y) = LU.member x avars andalso LU.member y avars
	fun restdom (I_Eq (alpha,beta)) = check2var (alpha,beta)
    	  | restdom (I_Neq (alpha,beta)) = check2var (alpha,beta)
    	  | restdom (I_Sub (str, alpha)) = LU.member str avars
    	  | restdom (I_Pmap (str,alpha,beta)) = check2var (alpha,beta)
(*    	  | restdom (I_Pnmap (str,alpha,beta)) = check2var (alpha,beta) *)
    	  | restdom (I_Fresh (alpha,X)) = true
    	  | restdom (I_Dfresh (X,Pi,Psi)) = true
    	  | restdom (I_Vmap (X, s)) = true
    in restdom problem end

(* input: atomic_problem *)
local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in
fun solveAtomicProblem xs = 
    let val avars = LU.toSet (LU.mapAppend atomVarsInAtomicProblem xs)
	val _ = dprint "AtomicSolve: input:\n"
	val _ = dprintln (LU.toStringCommaSquare prAtomicProblem xs)
	val _ = dprintln ("Avars: "  ^ (LU.toStringCommaCurly (fn x=>x) avars))
	val atoms = LU.toSet (LU.mapAppend atomConstsInAtomicProblem xs)
	val _ = dprintln ("Atoms: "  ^ (LU.toStringCommaCurly Atom.toString atoms))
			 
	val ys = replaceRHS xs
	val _ = dprint "AtomicSolve: RHS replaced:\n"
	val _ = dprintln (LU.toStringCommaSquare prAtomicProblem0 ys)

	(* !!!!!!!!! slide not reflected !!!!!!!!!!! *)
	val ys1 = replaceModeratedAtomicInSubstitution ys
	val _ = dprint "AtomicSolve: ModeratedAtomicInSubstition replaced:\n"
	val _ = dprintln (LU.toStringCommaSquare prAtomicProblem0 ys1)

	val zss = decomposeLHS [ys1]
	val _ = dprint "AtomicSolve: LHS decomposed:\n"
	val _ = L.app (dprintln o (LU.toStringCommaSquare prAtomicProblem0)) zss

	val ans = L.mapPartial (solveIdProblem o (L.map convertAtomicProblem0toIdProblem)) zss
	(* val _ = dprint "Ans:\n" *)
	(* val _ = L.app (dprintln o (LU.toStringCommaSquare prIdProblem)) ans *)

        (* 途中で導入したAtom変数についての代入の除去 *)
        (* 変数の変数代入 A:=B を，Aがgiven avar，Bがtmp avarのとき， B:=A に直す *)
	fun possibleSwithVarAssig ([], ans) = rev ans
	  | possibleSwithVarAssig ((x as I_Sub (str1, Xa str2))::xs, ans) = 
	    if LU.member str1 avars andalso not (LU.member str2 avars)
	    then let val sub = (str2, Xa str1)
		     val ans2 = L.map (updateIdProblem (AL.singleton sub)) ans
		     val xs2 = L.map (updateIdProblem (AL.singleton sub)) xs
		 in possibleSwithVarAssig (revAppend (ans2, (I_Sub sub)::xs2), [])
		 end
	    else possibleSwithVarAssig (xs, x::ans)
	  | possibleSwithVarAssig (x::xs, ans) = possibleSwithVarAssig (xs, x::ans)
	val ans1 = L.map (L.filter (fn x=> case x of I_Sub (str,_) => LU.member str avars | _ => true))
			 (L.map (fn xs => possibleSwithVarAssig (xs,[])) ans)
	val _ = dprintln "AtomicSolve: del assignments of tmp avars"
	val _ = L.app (dprintln o (LU.toStringCommaSquare prIdProblem)) ans1

        (* 表現の単純化 (1) *)
	val ans2 = L.map (L.map simplifyIdProblem) ans1
	(* val _ = dprintln "AtomicSolve: simplification" *)
	(* val _ = L.app (dprintln o (LU.toStringCommaSquare prIdProblem)) ans2 *)

        (* 表現の単純化 (2) *)
	fun mkAssign xs = let val ans = L.mapPartial (fn x => case x of 
								  I_Pmap (str, alpha, beta) => SOME ((str,alpha),beta)
								| _ => NONE) xs
			      (* val _ = dprintln (" rho = " ^
				 		(LU.toStringCommaCurly (fn ((x,y),z) => x ^":" ^ (prAtom y) ^ "|->" ^ (prAtom z)) ans)) *)
			  in ans  end 
	val ans3 = L.map (fn xs => L.map (simplifyIdProblemByAssign (mkAssign xs)) xs) ans2
	val _ = dprintln "AtomicSolve: simplification: output"
	val _ = L.app (dprintln o (LU.toStringCommaSquare prIdProblem)) ans3

    in  ans3  end
end

val _ = TU.test "solveAtomicProblem"
		[ let val a0 = Ca (Atom.fromString "a")
		      val b0 = Ca (Atom.fromString "b")
		      val Pi0 = Swap (Mod (Xp "P", Simp b0), (Simp a0))
		      val v1 = (Simp a0)
		      val v2 = (Mod (Pi0, Simp b0))
		  in solveAtomicProblem [ A_Eq (v1, v2)] = [[I_Pmap ("P",b0,b0)]]
		  end,
		  let val a0 = Ca (Atom.fromString "a")
		      val b0 = Ca (Atom.fromString "b")
		      val c0 = Ca (Atom.fromString "c")
		      val Pi0 = Swap (Mod (Xp "P", Simp b0), (Simp a0))
		      val v1 = (Simp a0)
		      val v2 = (Mod (Pi0, Simp b0))
		  in solveAtomicProblem [ A_Neq (v1, v2)] 
		     (* = [[I_Pnmap ("P",b0,b0)]] *)
		     (* = [[I_Pmap ("P",b0,a0)],[I_Pmap ("P",b0,c0)]] *)
		     = [[I_Neq (b0,Xa "A"),I_Pmap ("P",b0,Xa "A")]]
		  end,
		  let val a0 = Ca (Atom.fromString "a")
		      val b0 = Ca (Atom.fromString "b")
		      val A0 = Xa "A"
		      val Pi0 = Swap (Mod (Xp "P", Simp b0), (Simp a0))
		      val v1 = (Simp a0)
		      val v2 = (Mod (Pi0, Simp b0))
		      val Pi1 = Inv (Swap (Mod (Xp "P", Simp b0), (Simp a0)))
		      val v3 = Mod (Pi1, Simp a0)
		  in solveAtomicProblem [ A_Neq (v1, v2), A_Eq (v3, Simp A0) ]
		     (* =  [[I_Pmap ("P",b0, Ca (Atom.fromString "c"))],[I_Pmap ("P",b0,a0)]] *)
		     = [[I_Pmap ("P",b0,A0),I_Neq (b0,A0)],[I_Sub ("A",a0),I_Pmap ("P",b0,a0)]]
		  end,
		  let val a0 = Ca (Atom.fromString "a")
		      val b0 = Ca (Atom.fromString "b")
		      val A0 = Xa "A"
		      val Pi0 = Swap (Mod (Xp "P", Simp b0), (Simp a0))
		      val v1 = (Simp a0)
		      val v2 = (Mod (Pi0, Simp b0))
		      val Pi1 = Inv (Swap (Mod (Xp "P", Simp b0), (Simp a0)))
		      val v3 = Mod (Pi1, Simp a0)
		      val Y0 = Var.fromString "Y"
		  in solveAtomicProblem [ A_Neq (v1, v2), A_Fresh (v3, Y0) ]
		    (* =  [ [I_Pmap ("P",b0,Ca (Atom.fromString "c")), I_Fresh (Ca (Atom.fromString "c"),Y0)],
			  [I_Pmap ("P",b0,a0), I_Fresh (a0,Y0)] ] *)
		     (* =  [ [I_Fresh (Ca (Atom.fromString "c"),Y0), I_Pmap ("P",b0,Ca (Atom.fromString "c"))],
			  [I_Fresh (a0,Y0), I_Pmap ("P",b0,a0)] ] *)
		     = [[I_Pmap ("P",b0,A0),I_Fresh (A0,Y0), I_Neq (b0,A0)],
                        [I_Fresh (a0,Y0),  I_Pmap ("P",b0,a0)]]
		  end
		]


(*****************************************)
(***    Unification for Expressions    ***)
(*****************************************)

datatype unif_problem = U_Eq of term_expr * term_expr
		  | U_Fresh of atomic_expr * term_expr

fun prUnifProblem (U_Eq (s,t)) = (prTerm s) ^ "==" ^ (prTerm t)
  | prUnifProblem (U_Fresh (v, s)) = ((prAtomic v) ^ "#" ^ (prTerm s))

fun applySubstToUnifProblem sigma (U_Eq (s,t)) = U_Eq (applySubstToTerm sigma s, applySubstToTerm sigma t)
  | applySubstToUnifProblem sigma (U_Fresh (v,t)) = U_Fresh (v, applySubstToTerm sigma t)

exception UnifError
local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in
fun unify0 []  = [[]]
  | unify0 (U_Fresh (v,t)::xs) = 
    (dprintln ("Decomp: U_Fresh " ^ (prUnifProblem (U_Fresh (v,t))));
     case t of  (** same as resolve0 ***)
	 Atomic w => (dprintln (" => A_Neq " ^ (prAtomicProblem (A_Neq (v,w))));
		      L.map (fn ys=> A_Neq (v,w)::ys) (unify0 xs))                    (* bug fix 2015/12/30, Eq => Neq *)
       | Susp (Pi,X) => let val y = actAtomic (invPerm Pi) v 
			    val _ = dprintln (" => A_Fresh " ^ (prAtomicProblem (A_Fresh (y,X))))
			in L.map (fn ys=> A_Fresh (y,X)::ys) (unify0 xs)
			end
       | Abst (w,exp) => (dprintln (" => A_Eq " ^ (prAtomicProblem (A_Eq (v,w))));
			  L.map (fn ys=> A_Eq (v,w)::ys) (unify0 xs))
			 @ (dprintln (" => A_Neq " ^ (prAtomicProblem (A_Neq (v,w)))
				      ^ " => U_Fresh " ^ (prUnifProblem (U_Fresh (v,exp))));
			    L.map (fn ys=> A_Neq (v,w)::ys) (unify0 (U_Fresh (v,exp)::xs)))
       | Fun (f,exp) => (dprintln (" => U_Fresh " ^ (prUnifProblem (U_Fresh (v,exp))));
			 unify0 (U_Fresh (v,exp)::xs))
       | Tuple exps => (dprintln (" => U_Fresh ...");
			unify0 ((L.map (fn s=> U_Fresh (v,s)) exps) @ xs)))
  | unify0 (U_Eq (s,t)::xs) = 
    (dprintln ("Decomp: U_Eq " ^ (prUnifProblem (U_Eq (s,t))));
     case (s,t) of 
	 (Atomic v, Atomic w)  => (dprintln (" => A_Eq " ^ (prAtomicProblem (A_Eq (v,w))));
				   L.map (fn ys=> A_Eq (v,w)::ys) (unify0 xs))
       | (Susp (Pi,X),_) => if not (LU.member X (varsInTerm t))
			    then (dprintln (" => A_Vmap " ^ (prAtomicProblem (A_Vmap (X, actDotTerm (invPerm Pi) t))));
				  L.map (fn ys=> (A_Vmap (X, actDotTerm (invPerm Pi) t))::ys) 
					(unify0 (L.map (applySubstToUnifProblem [(X, actDotTerm (invPerm Pi) t)]) xs)))
			    else (case t of 
				      Susp (Psi,Y) => if X = Y
						      then (dprintln (" => A_Dfresh " ^ (prAtomicProblem (A_Dfresh (X,Pi,Psi))));
							    L.map (fn ys=> A_Dfresh (X,Pi,Psi)::ys) (unify0 xs))
						      else []
				    | _  =>  [])
       | (_, Susp (Psi,Y)) => (dprintln (" => U_Eq " ^ (prUnifProblem (U_Eq (t,s))));
			       unify0 (U_Eq (t,s)::xs))
       | (Abst (v,s2), Abst (w,t2)) =>  (dprintln (" => A_Eq " ^ (prAtomicProblem (A_Eq (v,w))));
					 L.map (fn ys=> A_Eq (v,w)::ys) (unify0 (U_Eq (s2,t2)::xs)))
					(* @ (dprintln (" => A_Neq " ^ (prAtomicProblem (A_Neq (v,w))) *)
					(* 	     ^ " U_Eq " ^ (prUnifProblem (U_Eq (actDotTerm (Swap (v,w)) s2, t2))) *)
					(* 	     ^ " U_Fresh " ^ (prUnifProblem (U_Fresh (w,s2)))); *)
					(*    L.map (fn ys=> A_Neq (v,w)::ys) *)
					(* 	 (unify0 ((U_Eq (actDotTerm (Swap (v,w)) s2, t2)) :: (U_Fresh (w,s2)) ::xs))) *)
                                        (* change of def of aeq reflected 2015/12/29 *)
					@ (dprintln (" => A_Neq " ^ (prAtomicProblem (A_Neq (v,w)))
						     ^ " U_Eq " ^ (prUnifProblem (U_Eq (s2, actDotTerm (Swap (v,w)) t2)))
						     ^ " U_Fresh " ^ (prUnifProblem (U_Fresh (v,t2))));
					   L.map (fn ys=> A_Neq (v,w)::ys)
						 (unify0 ((U_Eq (s2, actDotTerm (Swap (v,w)) t2)) :: (U_Fresh (v,t2)) ::xs)))
       | (Fun (f,s2), Fun (g,t2)) => if f = g
				     then (dprintln (" => U_Eq " ^ (prUnifProblem (U_Eq (s2,t2))));
					   unify0  (U_Eq (s2,t2) ::xs))
				     else (dprintln (" => bot ");
					   [])
       | (Tuple ss, Tuple ts) => (dprintln (" => U_Eq ..."); 
				  unify0 ((LP.map (fn (x,y)=> U_Eq (x,y)) (ss,ts)) @  xs))
       | _ => (dprintln "=> bot"; []))
end

local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in
fun unify xs = 
    let val _ = dprint "Unify input: \n"
	val _ = dprintln (LU.toStringCommaSquare prUnifProblem xs)
        val yss = unify0 xs
        val _ = dprint "Unify0 output: \n"
        val _ = L.app (dprintln o (LU.toStringCommaSquare prAtomicProblem)) yss 
	val ans =  LU.mapAppend solveAtomicProblem (L.map (L.map simplifyAtomicProblem) yss)
        val _ = dprint "Unify output: \n"
        val _ = L.app (dprintln o (LU.toStringCommaSquare prIdProblem)) ans
    in ans
end

val _ = TU.test "unify"
		[ (* [a==(P*b a)*b] *)
		  let val a0 = Ca (Atom.fromString "a")
		      val b0 = Ca (Atom.fromString "b")
		      val Pi0 = Swap (Mod (Xp "P", Simp b0), (Simp a0))
		      val t1 = Atomic (Simp a0)
		      val t2 = Atomic (Mod (Pi0, Simp b0))
		  in unify [U_Eq (t1, t2)] = [[I_Pmap ("P",b0,b0)]]
		  end
		,
		(* [(forall [b]X)==(forall [a](forall [b]Y))] *)
		  let val a0 = Ca (Atom.fromString "a")
		      val b0 = Ca (Atom.fromString "b")
		      val X0 = Var.fromString "X"
		      val Y0 = Var.fromString "Y"
		      val f0 = Fun.fromString "forall"
		      val t3a = Abst (Simp b0, Susp (Id, X0))
		      val t3 = Fun (f0, t3a)
		      val t4c = Abst (Simp b0, Susp (Id, Y0))
		      val t4b = Fun (f0, t4c)
		      val t4a = Abst (Simp a0, t4b)
		      val t4 = Fun (f0, t4a)
		  in unify [ U_Eq (t3, t4)] 
		     = [[I_Vmap (X0, Fun (f0, Abst (Simp a0, Susp  (Swap (Simp b0, Simp a0), Y0))))]]
		  end
		,
		  (* Problem: [(forall [P*b]((P o P^{-1}))*X)==(forall [a](forall [b]Y))] *)
                  (* Solution: [ [P:b|->a,X|->(forall [b]Y)], 
                                 [P:b|->b,X|->(forall [a]((b a))*Y)],
                                 [c#Y,X|->(forall [b]((c a))*Y),P:b|->c] ] *)
		  let val a0 = Ca (Atom.fromString "a")
		      val b0 = Ca (Atom.fromString "b")
		      val c0 = Ca (Atom.fromString "c")
		      val X0 = Var.fromString "X"
		      val Y0 = Var.fromString "Y"
		      val f0 = Fun.fromString "forall"
		      val t3a = Abst (Simp b0, Susp (Id, X0))
		      val t3 = Fun (f0, t3a)
		      val t4c = Abst (Simp b0, Susp (Id, Y0))
		      val t4b = Fun (f0, t4c)
		      val t4a = Abst (Simp a0, t4b)
		      val t4 = Fun (f0, t4a)
		      val P0 = Xp "P"
		      val C0 = Xa "C"
		      val t5 = actHatTerm P0 t3
		  in unify [ U_Eq (t5, t4)]
		   (*  = [[I_Pmap ("P",b0,a0), I_Vmap (X0,Fun(f0,Abst(Simp b0,Susp (Id,Y0))))],
			[I_Pmap ("P",b0,b0), I_Vmap (X0,Fun(f0,Abst(Simp a0,Susp(Swap (Simp b0,Simp a0),Y0))))],
			[I_Fresh(c0,Y0),I_Vmap(X0,Fun(f0,Abst(Simp b0,Susp(Swap (Simp c0),Simp a0),Y0)))),I_Pmap ("P",b0,c0)]] *)
		   (*  =   [[I_Pmap("P",b0,a0),I_Vmap(X0,Fun(f0,Abst(Simp b0,Susp(Id,Y0))))],
			  [I_Pmap("P",b0,b0),I_Vmap(X0,Fun(f0,Abst(Simp a0,Susp(Swap(Simp b0,Simp a0),Y0))))],
			  [I_Pmap("P",b0,c0),I_Vmap(X0,Fun(f0,Abst(Simp b0,Susp(Swap(Simp c0,Simp a0),Y0)))),I_Fresh(c0,Y0)]] *)
		     =   [[I_Pmap("P",b0,a0),I_Vmap(X0,Fun(f0,Abst(Simp b0,Susp(Id,Y0))))],
			  [I_Pmap("P",b0,b0),I_Vmap(X0,Fun(f0,Abst(Simp a0,Susp(Swap(Simp b0,Simp a0),Y0))))],
			  (* [I_Pmap("P",b0,c0),I_Vmap(X0,Fun(f0,Abst(Simp b0,Susp(Swap(Simp c0,Simp a0),Y0)))),I_Fresh(c0,Y0)]]*)
			  [I_Fresh(C0,Y0),I_Neq(C0,a0),
			   I_Vmap(X0,Fun(f0,Abst(Mod(Swap (Simp C0,Simp a0),Simp b0),Susp(Swap(Simp C0,Simp a0),Y0)))),
			   I_Pmap("P",b0,C0),I_Neq(b0,C0)]]
		  end
		,
                  (* Problem: [(sub <[P*a](lam [P*b]X,Y>)==(sub <[a](lam [b](app <b,a>)),b>)] *)
                 (* Solutions:  [X|->(app <a,b>),Y|->b,P:a|->b,P:b|->a]
                                [X|->(app <a,c>),Y|->b,P:a|->c,P:b|->a]
                                [X|->(app <b,a>),Y|->b,P:a|->a,P:b|->b]
                                [X|->(app <b,c>),Y|->b,P:a|->c,P:b|->b]
                                [X|->(app <c,a>),Y|->b,P:a|->a,P:b|->c]
                                [X|->(app <c,b>),Y|->b,P:a|->b,P:b|->c]  *)
		  let val P0 = Xp "P"
		      val u6 = term2expr (T.fromString "(sub <[a](lam [b][]X), []Y>)")
		      val t6 = actHatTerm P0 u6
		      val t7 = term2expr (T.fromString "(sub <[a](lam [b](app <b,a>)),b>)")
		      val a0 = Ca (Atom.fromString "a")
		      val b0 = Ca (Atom.fromString "b")
		      val c0 = Ca (Atom.fromString "c") (* obtained as a fresh atom *)
		      val X0 = Var.fromString "X"
		      val Y0 = Var.fromString "Y"
		      val f0 = Fun.fromString "app"
		  in unify [ U_Eq (t6, t7)] =
(*  [[I_Pmap("P",b0,b0),I_Vmap(X0,Fun(f0,Tuple [Atomic (Simp (b0)),Atomic (Simp (a0))])),I_Vmap (Y0,Atomic (Simp (b0))),I_Pmap ("P",a0,a0)],
   [I_Vmap(X0,Fun(f0,Tuple [Atomic (Simp (c0)),Atomic (Simp (a0))])),I_Pmap ("P",a0,a0),I_Pmap ("P",b0,c0),I_Vmap (Y0,Atomic (Simp (b0)))],
   [I_Vmap(Y0,Atomic(Simp (b0))),I_Pmap ("P",a0,b0),I_Pmap ("P",b0,a0),I_Vmap(X0,Fun(f0,Tuple [Atomic (Simp (a0)),Atomic (Simp (b0))]))],
   [I_Vmap(X0,Fun(f0,Tuple [Atomic (Simp (b0)),Atomic (Simp (c0))])),I_Pmap ("P",b0,b0),I_Pmap ("P",a0,c0),I_Vmap (Y0,Atomic (Simp (b0)))],
   [I_Vmap(X0,Fun(f0,Tuple [Atomic (Simp (c0)),Atomic (Simp (b0))])),I_Pmap ("P",b0,c0),I_Pmap ("P",a0,b0),I_Vmap (Y0,Atomic (Simp (b0)))],
   [I_Vmap (Y0,Atomic (Simp (b0))),I_Pmap ("P",a0,c0),I_Pmap ("P",b0,a0),I_Vmap(X0,Fun(f0,Tuple [Atomic (Simp (a0)),Atomic (Simp (c0))]))]]
*)
(*
  [[I_Pmap ("P",b0,b0),I_Vmap(X0,Fun(f0, Tuple [Atomic (Simp b0),Atomic (Simp a0)])), I_Vmap (Y0,Atomic (Simp b0)),I_Pmap ("P",a0,a0)], 
   [I_Vmap (Y0,Atomic (Simp b0)), I_Pmap("P",b0,c0), I_Pmap ("P",a0,a0), I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp c0),Atomic (Simp a0)]))], 
   [I_Vmap (Y0,Atomic(Simp b0)), I_Pmap ("P",a0,b0), I_Pmap ("P",b0,a0), I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp a0),Atomic (Simp b0)]))], 
   [I_Vmap(X0, Fun (f0, Tuple [Atomic (Simp b0),Atomic (Simp c0)])),I_Pmap ("P",b0,b0), I_Pmap ("P",a0,c0), I_Vmap (Y0,Atomic(Simp b0))], 
   [I_Vmap (Y0,Atomic (Simp b0)), I_Pmap("P",a0,b0), I_Pmap ("P",b0,c0), I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp c0),Atomic (Simp b0)]))], 
   [I_Vmap (X0, Fun(f0, Tuple [Atomic (Simp a0),Atomic (Simp c0)])), I_Pmap("P",b0,a0), I_Pmap ("P",a0,c0), I_Vmap (Y0,Atomic (Simp b0))]]
*)
  [[I_Pmap ("P",b0,b0), I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp b0),Atomic (Simp a0)])), I_Vmap (Y0,Atomic (Simp b0)), I_Pmap("P",a0,a0)],
   [I_Neq (Xa "C",b0), I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp (Xa "C")),Atomic (Simp a0)])), I_Pmap ("P",a0,a0),
    I_Pmap ("P",b0,Xa "C"),I_Vmap (Y0,Atomic (Simp b0)), I_Neq (Xa "C",a0)],
   [I_Vmap (Y0,Atomic (Simp b0)), I_Pmap ("P",a0,b0), I_Pmap ("P",b0,a0), I_Vmap (X0, Fun (f0, Tuple [Atomic(Simp a0),Atomic(Simp b0)]))],
   [I_Vmap (Y0,Atomic (Simp b0)), I_Pmap ("P",a0,Xa "A"),I_Neq (Xa "A",a0),I_Pmap ("P",b0,b0),
    I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp b0),Atomic (Simp (Xa "A"))])),I_Neq (Xa "A",b0)],
   [I_Neq (a0,Xa "B0"), I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp (Xa "B0")),Atomic (Simp b0)])),
    I_Pmap ("P",b0,Xa "B0"),I_Neq (b0,Xa "B0"), I_Pmap ("P",a0,b0), I_Vmap (Y0,Atomic (Simp b0))],
   [I_Neq (b0,Xa "B0"),I_Vmap (X0, Fun (f0,Tuple [Atomic (Simp (Xa "B0")),Atomic (Simp (Xa "A"))])),
    I_Pmap ("P",b0,Xa "B0"), I_Neq (Xa "A",b0),I_Neq (Xa "A",Xa "B0"), I_Pmap ("P",a0,Xa "A"), I_Vmap (Y0,Atomic (Simp b0)), I_Neq (Xa "A",a0)]]
		  end
		,
		 let val a0 = Ca (Atom.fromString "a")
		      val b0 = Ca (Atom.fromString "b")
		      val X0 = Var.fromString "X"
		      val Pi0 = Swap (Mod (Xp "P", Simp b0), (Simp a0))
		      val Pi1 = Xp "P"
		      val v1 = (Susp (Pi0, X0))
		      val v2 = (Susp (Pi1, X0))
		  in unify [ U_Eq (v1, v2)] =
		     [[I_Dfresh (X0, Swap (Mod (Xp "P",Simp b0),Simp a0),Xp "P")]]
		 end
		]


(**********************************************)
(** Instantiation of Solved Identity Problem **)
(**********************************************)

(* one step instantitions if possible *)
local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in
fun instantiate1 atoms xs =
    let val vars = LU.unionAll (L.map unsolvedAtomVarsInIdProblem xs)
	(* val _ = dprintln ("Inst: Vars: " ^ (LU.toStringCommaCurly (fn x=>x) vars)) *)
    in if null vars
       then NONE
       else let val var = hd vars
   		val _ = dprintln "Inst input:"
		val _ = dprintln (LU.toStringCommaSquare prIdProblem xs)
   		(* val _ = dprintln ("Inst: Atoms:" ^ (LU.toStringCommaCurly Atom.toString  atoms)) *)
		val allAtoms = LU.toSet (atoms @ (LU.mapUnion atomConstsInIdProblem xs))
   		val new = Atom.newAtom allAtoms
                (* instantiate atom var with all existing atoms + one fresh atom *)
   		val updatedProblems = L.map (fn a =>
   						I_Sub (var, Ca a)::
   						(L.map (updateIdProblem (AL.singleton (var,Ca a))) xs))
   					    (allAtoms@[new])
   		(*    val procedure0: problem0 list -> (problem0 list) option *)
   		val _ = dprintln ("Instantiting " ^ var ^ ":")
		val _ = L.app (fn xs => dprintln (LU.toStringCommaSquare prIdProblem xs)) updatedProblems
   		val updatedProblems' = L.mapPartial solveIdProblem updatedProblems
   		val _ = dprintln "Inst output (solved instantitiations):"
		val _ = L.app (fn xs => dprintln (LU.toStringCommaSquare prIdProblem xs)) updatedProblems'
   	    in SOME (updatedProblems')
   	    end
    end
end

(* repeat instantitions untill no variable remains *)
fun instantiate1full atoms xs =
    case  instantiate1 atoms xs of
   	NONE => [xs]
      | SOME yss => LU.mapAppend (instantiate1full atoms) yss

exception EqvUnifError

(* !!! addition of atoms is required, in order to fully cover the critical pairs (yet not reflected in slide) !!! **)
local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in
fun unifyForBCP (env1,l1) (env2,l2,atoms) =
    let  val P0 = Xp "P"
	 val l1_expr = term2expr l1
	 val l2_expr = actHatTerm P0 (term2expr l2)
	 val fresh1 = L.map (fn (a,X) => U_Fresh (Simp (Ca a), Susp (Id, X)))  (FreshEnv.toList env1)
	 val fresh2 = L.map (fn (a,X) => U_Fresh (Simp (Ca a), Susp (Inv P0, X)))  (FreshEnv.toList env2)	
	 val allAtoms = LU.toSet ((FreshEnv.atoms env1) @ (T.atoms l1)@ (FreshEnv.atoms env2) @ (T.atoms l2) @ atoms)

	 (* for any atom "c" not in l2, we add  "Pi(c) = C" where C is a fresh atom variable *)
	 val neededAtoms = LU.differenceByAll (atoms, T.atoms l2)
	 val freshes = newAtomVars [] (L.length neededAtoms)
	 val eqsForAtoms = LP.map (fn (a,name) => U_Eq (Atomic (actAtom P0 (atom2expr a)), Atomic (Simp (Xa name)))) 
				  (neededAtoms,freshes)

	 val problem = [ U_Eq (l1_expr, l2_expr) ] @ fresh1 @ fresh2 @ eqsForAtoms

	 val _ = dprintln "unifyForBCP..."
	 val _ = dprintln ("input: " ^ (LU.toStringCommaCurly prUnifProblem problem))

	 val solutions = unify problem

        (* 使っていない Atom変数の式の削除 ...正しいか怪しいので pending *)
	val ans4 = solutions
	(* val ans4 = L.map (fn xs => let val avars2 = relevantAtoms xs ([], avars) *)
	(*    			   in L.filter (checkDomainOfIdProblem avars2) xs *)
	(*    			   end) *)
	(*    		 ans3 *)
	(* val _ = dprintln "AtomicSolve: delete redundant " *)
	(* val _ = L.app (dprintln o (LU.toStringCommaSquare prIdProblem)) ans4 *)

        (* Atom変数の具体化 と Atom変数代入の除去 *)
	val ans5 = L.map (fn xs => L.map simplifyIdProblem (L.filter (fn x=> case x of I_Sub _=> false | _ => true)  xs))
			 (LU.mapAppend (instantiate1full allAtoms) ans4)
	val ans5b = LU.toSet' (fn (xs,ys)=> LU.isEqset xs ys) ans5
	val _ = dprintln "unifyForBCP: instantiation:"
	val _ = L.app (dprintln o (LU.toStringCommaSquare prIdProblem)) ans5b

     (* delayed fresh (I_Dfresh of var * perm_expr * perm_expr)
        を freshness constraint (I_Fresh of (atom_expr * var)) へ変換 *)
     (* (a b)*X == P*X のような場合には失敗するが， 危険対の場合にはそのようなのはないはず *) 
	fun transDelayedFresh (I_Dfresh (X, Pi, Psi)) =
	    (let val pi = perm_expr2perm Pi
		val psi = perm_expr2perm Psi
		val atoms = Perm.disagreements (pi,psi)
	    in  L.map (fn a => I_Fresh (atom2expr a, X)) atoms
	    end
	    handle TranslationFromExprToTermFailure => (print "Error: not an equivariant unification for BCP ???\n";
							raise EqvUnifError))
	  | transDelayedFresh x = [x]
	val ans6 = L.map (LU.mapAppend transDelayedFresh) ans5b
	val _ = dprintln "AtomicSolve: translate delayed fresh cond."
	val _ = L.app (dprintln o (LU.toStringCommaSquare prIdProblem)) ans6

	 fun mkAns xs  =
	     let val _ = dprintln (LU.toStringCommaSquare prIdProblem xs)
		 val sigma = L.mapPartial (fn x => case x of I_Vmap (X,s) => SOME (X,term_expr2term s) 
							   | _ => NONE) xs
		 val env = L.mapPartial (fn x => case x of I_Fresh (alpha,X) => SOME (atom_expr2atom alpha, X) 
							 | _ => NONE) xs
		 val pimap = L.mapPartial (fn x => case x of I_Pmap (_,alpha,beta) => SOME (atom_expr2atom alpha, atom_expr2atom beta)
							   | _ => NONE) xs
	     (* in (sigma,env,Perm.fromMapping pimap) *)
	     in (sigma,env,pimap)
	     end
(*    in L.map mkAns solutions *)
    in L.map mkAns ans6
    end
end
end (* of local *)

val _ = TU.test "unifyForBCP"
		[ 
		  (** Example 15 [SKAT 2015]  **)
		  (* input: {(app <(lam [a]Y),Z>)==(app <(lam [P*a]((P o P^{-1}))*X),((P o P^{-1}))*W>)} *)
		  (* solutions: [P:a|->a,Y|->X,Z|->W], [P:a|->b,Z|->W,a#X,Y|->((a b))*X] *)
		  (unifyForBCP (FreshEnv.rdTermInEnv "|- (app <(lam [a][]Y), []Z>)") 
			      (FreshEnv.empty, Term.fromString "(app <(lam [a][]X), []W>)", [Atom.fromString "a"])
		  =  [([(Var.fromString "Y", Term.fromString "[]X"), (Var.fromString "Z", Term.fromString "[]W")],
		       [],
		       [(Atom.fromString "a",Atom.fromString "a")]),
		      ([(Var.fromString "Z", Term.fromString "[]W"), (Var.fromString "Y", Term.fromString "[(a b)]X")],
		       (FreshEnv.fromString "a#X"),
		       [(Atom.fromString "a",Atom.fromString "b")])])
		]


(*****************************************)
(***      Matching for Expressions     ***)
(*****************************************)

fun applySubstToMatchProblem sigma (U_Eq (s,t)) = U_Eq (applySubstToTerm sigma s, t)
  | applySubstToMatchProblem sigma (U_Fresh (v,t)) = U_Fresh (v,t)

(* for any freshness constraint a#t, t is (acted) term from term (i.e. not from pattern)  *)
exception MatchError
local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in
fun match0 ([], sigma) = [([],sigma)]
  | match0 (U_Fresh (v,t)::xs, sigma) = 
    (dprintln ("Decomp: U_Fresh " ^ (prUnifProblem (U_Fresh (v,t))));
     case t of  (** same as resolve0 ***)
	 Atomic w => (dprintln (" => A_Neq " ^ (prAtomicProblem (A_Neq (v,w)))); 
		      L.map (fn (ys,rho) => (A_Neq (v,w)::ys,rho)) (match0 (xs,sigma))) (* bug fix 2015/12/30 Eq => Neq *)
       | Susp (Pi,X) => let val y = actAtomic (invPerm Pi) v 
			    val _ = dprintln (" => A_Fresh " ^ (prAtomicProblem (A_Fresh (y,X))))
			in L.map (fn (ys,rho)=> (A_Fresh (y,X)::ys,rho)) (match0 (xs,sigma))
			end
       | Abst (w,exp) => (dprintln (" => A_Eq " ^ (prAtomicProblem (A_Eq (v,w))));
			  L.map (fn (ys,rho)=> (A_Eq (v,w)::ys,rho)) (match0 (xs,sigma)))
			 @ (dprintln (" => A_Neq " ^ (prAtomicProblem (A_Neq (v,w)))
				      ^ " => U_Fresh " ^ (prUnifProblem (U_Fresh (v,exp))));
			    L.map (fn (ys,rho)=> (A_Neq (v,w)::ys,rho)) (match0 (U_Fresh (v,exp)::xs,sigma)))
       | Fun (f,exp) => (dprintln (" => U_Fresh " ^ (prUnifProblem (U_Fresh (v,exp))));
			 match0 (U_Fresh (v,exp)::xs,sigma))
       | Tuple exps => (dprintln (" => U_Fresh ...");
			match0 ((L.map (fn s=> U_Fresh (v,s)) exps) @ xs, sigma)))
  | match0 (U_Eq (s,t)::xs, sigma) = 
    (dprintln ("Decomp: U_Eq " ^ (prUnifProblem (U_Eq (s,t))));
     case (s,t) of 
	 (Atomic v, Atomic w)  => (dprintln (" => A_Eq " ^ (prAtomicProblem (A_Eq (v,w))));
				   L.map (fn (ys,rho)=> (A_Eq (v,w)::ys,rho)) (match0 (xs, sigma)))
       | (Susp (Pi,X),_) => (case AL.add (X, actDotTerm (invPerm Pi) t) sigma of
				 NONE => []
			       | SOME theta => (* match0 (xs,theta)) *)  (*** !!! bug??? why not update lhs of xs ?? 2015/12/29 **)
					match0 (L.map (applySubstToMatchProblem [(X, actDotTerm (invPerm Pi) t)]) xs, theta))
       | (Abst (v,s2), Abst (w,t2)) =>  (dprintln (" => A_Eq " ^ (prAtomicProblem (A_Eq (v,w))));
					 L.map (fn (ys,rho)=> (A_Eq (v,w)::ys,rho)) (match0 (U_Eq (s2,t2)::xs, sigma)))
					(* @ (dprintln (" => A_Neq " ^ (prAtomicProblem (A_Neq (v,w))) *)
					(* 	     ^ " U_Eq " ^ (prUnifProblem (U_Eq (actDotTerm (Swap (v,w)) s2, t2))) *)
					(* 	     ^ " U_Fresh " ^ (prUnifProblem (U_Fresh (w,s2)))); *)
					(*    L.map (fn (ys,rho)=> (A_Neq (v,w)::ys,rho)) *)
					(* 	 (match0 ((U_Eq (actDotTerm (Swap (v,w)) s2, t2)) :: (U_Fresh (w,s2)) ::xs, sigma))) *)
                                        (* change of def of aeq reflected 2015/12/29 *)
					@ (dprintln (" => A_Neq " ^ (prAtomicProblem (A_Neq (v,w)))
						     ^ " U_Eq " ^ (prUnifProblem (U_Eq (s2, actDotTerm (Swap (v,w)) t2)))
						     ^ " U_Fresh " ^ (prUnifProblem (U_Fresh (v,t2))));
					   L.map (fn (ys,rho)=> (A_Neq (v,w)::ys,rho))
						 (match0 ((U_Eq (s2, actDotTerm (Swap (v,w)) t2)) :: (U_Fresh (v,t2)) ::xs, sigma)))
       | (Fun (f,s2), Fun (g,t2)) => if f = g
				     then (dprintln (" => U_Eq " ^ (prUnifProblem (U_Eq (s2,t2))));
					   match0  (U_Eq (s2,t2) ::xs,sigma))
				     else (dprintln (" => bot "); [])
       | (Tuple ss, Tuple ts) => (dprintln (" => U_Eq ..."); 
				  match0 ((LP.map (fn (x,y)=> U_Eq (x,y)) (ss,ts)) @  xs, sigma))
       | _ => (dprintln "=> bot"; []))
end


local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in
fun match xs = 
    let val _ = dprint "Match input: \n"
	val _ = dprintln (LU.toStringCommaSquare prUnifProblem xs)
	val yss = match0 (xs,[])
	val _ = dprint "Match0: output \n"
	val _ = L.app (fn (xs,sigma) => 
	 		  (dprintln (LU.toStringCommaSquare prAtomicProblem xs);
	 		   dprintln (LU.toStringCommaCurly (fn (X,t)=> (Var.toString X) ^ " := " ^ (prTerm t)) sigma)))
	 	      yss
	fun mkSub sigma = L.map (fn (X,s) => A_Vmap (X, s)) sigma
    in  LU.mapAppend solveAtomicProblem 
		     (L.map (fn (xs,sigma) => L.map simplifyAtomicProblem (xs @ mkSub sigma)) yss)
    end
end (* of local *)

val _ = TU.test "match"
		[ (* Problem: [(sub <[P*a](lam [P*b]X,Y>)==?(sub <[a](lam [b](app <b,a>)),b>)] *)
                  (* Solutions:[ [X|->(app <a,b>),Y|->b,P:a|->b,P:b|->a],
                                 [X|->(app <a,c>),Y|->b,P:a|->c,P:b|->a],
                                 [X|->(app <b,a>),Y|->b,P:b|->b,P:a|->a],
                                 [X|->(app <b,c>),Y|->b,P:b|->b,P:a|->c],
                                 [X|->(app <c,a>),Y|->b,P:a|->a,P:b|->c],
                                 [X|->(app <c,b>),Y|->b,P:b|->c,P:a|->b] ] *)
		  let val P0 = Xp "P"
		      val u6 = term2expr (T.fromString "(sub <[a](lam [b][]X), []Y>)")
		      val t6 = actHatTerm P0 u6
		      val t7 = term2expr (T.fromString "(sub <[a](lam [b](app <b,a>)),b>)")
		      val a0 = Ca (Atom.fromString "a")
		      val b0 = Ca (Atom.fromString "b")
		      val c0 = Ca (Atom.fromString "c")  (* obtained as a fresh var *)
		      val X0 = Var.fromString "X"
		      val Y0 = Var.fromString "Y"
		      val f0 = Fun.fromString "app"
		  in match [ U_Eq (t6, t7)]
(*
		     =  [[I_Pmap ("P",b0,b0),I_Vmap (Y0,Atomic (Simp (b0))), I_Vmap(X0,Fun(f0,Tuple [Atomic (Simp (b0)),Atomic (Simp (a0))])),
                          I_Pmap ("P",a0,a0)],
			 [I_Pmap ("P",a0,a0), I_Pmap ("P",b0,c0), I_Vmap (X0, Fun(f0, Tuple [Atomic (Simp (c0)),Atomic (Simp (a0))])), 
			  I_Vmap (Y0,Atomic (Simp (b0)))],
			 [I_Vmap (Y0,Atomic (Simp (b0))),I_Vmap(X0,Fun (f0,Tuple [Atomic (Simp (a0)),Atomic (Simp (b0))])),I_Pmap ("P",a0,b0),
			  I_Pmap ("P",b0,a0)],
			 [I_Pmap ("P",b0,b0),I_Pmap ("P",a0,c0),I_Vmap(X0,Fun(f0,Tuple [Atomic (Simp (b0)),Atomic (Simp (c0))])),
			  I_Vmap (Y0,Atomic (Simp (b0)))],
			 [I_Pmap ("P",b0,c0),I_Pmap ("P",a0,b0),I_Vmap(X0,Fun(f0,Tuple [Atomic (Simp (c0)),Atomic (Simp (b0))])),
			  I_Vmap (Y0,Atomic (Simp (b0)))],
			 [I_Vmap (Y0,Atomic (Simp (b0))),I_Vmap(X0,Fun(f0,Tuple [Atomic (Simp (a0)),Atomic (Simp (c0))])),I_Pmap ("P",a0,c0),
			  I_Pmap ("P",b0,a0)]]
*)
(*
=  [[I_Pmap ("P",b0,b0), I_Vmap (Y0,Atomic (Simp b0)), I_Vmap (X0, Fun(f0, Tuple [Atomic (Simp b0),Atomic (Simp a0)])), I_Pmap ("P",a0,a0)], 
   [I_Vmap (Y0,Atomic (Simp b0)), I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp c0),Atomic (Simp a0)])), I_Pmap ("P",b0,c0), I_Pmap ("P",a0,a0)], 
   [I_Vmap (Y0,Atomic (Simp b0)), I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp a0),Atomic (Simp b0)])), I_Pmap ("P",a0,b0), I_Pmap ("P",b0,a0)], 
   [I_Pmap ("P",b0,b0), I_Pmap("P",a0,c0), I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp b0),Atomic (Simp c0)])), I_Vmap (Y0,Atomic (Simp b0))], 
   [I_Vmap (Y0,Atomic(Simp b0)), I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp c0),Atomic(Simp b0)])), I_Pmap ("P",a0,b0), I_Pmap ("P",b0,c0)], 
   [I_Pmap("P",b0,a0), I_Pmap ("P",a0,c0), I_Vmap (X0, Fun (f0, Tuple[Atomic (Simp a0),Atomic (Simp c0)])), I_Vmap (Y0,Atomic (Simp b0))]]
*)
=  [[I_Pmap ("P",b0,b0), I_Vmap (Y0,Atomic (Simp b0)), I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp b0),Atomic (Simp a0)])), I_Pmap ("P",a0,a0)],
   [I_Neq (Xa "C",b0),I_Pmap ("P",a0,a0), I_Pmap ("P",b0,Xa "C"),I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp (Xa "C")), Atomic (Simp a0)])), 
    I_Vmap (Y0,Atomic (Simp b0)), I_Neq (Xa "C",a0)],
   [I_Vmap (Y0,Atomic (Simp b0)), I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp a0),Atomic (Simp b0)])), I_Pmap ("P",a0,b0),I_Pmap ("P",b0,a0)],
   [I_Vmap (Y0,Atomic (Simp b0)), I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp b0),Atomic (Simp (Xa "A"))])), I_Pmap ("P",a0,Xa "A"),
    I_Neq (Xa "A",a0), I_Pmap ("P",b0,b0),I_Neq (Xa "A",b0)],
   [I_Neq (a0,Xa "B0"),I_Pmap ("P",b0,Xa "B0"), I_Neq (b0,Xa "B0"),I_Pmap ("P",a0,b0), 
    I_Vmap (X0, Fun (f0, Tuple [Atomic (Simp (Xa "B0")),Atomic (Simp b0)])), I_Vmap (Y0,Atomic (Simp b0))],
   [I_Neq (b0,Xa "B0"),I_Pmap ("P",b0,Xa "B0"), I_Neq (Xa "A",b0),I_Neq (Xa "A",Xa "B0"), I_Pmap ("P",a0,Xa "A"), 
    I_Vmap (X0, Fun (f0,Tuple [Atomic (Simp (Xa "B0")),Atomic (Simp (Xa "A"))])), 
    I_Vmap (Y0,Atomic (Simp b0)), I_Neq (Xa "A",a0)]]
		  end
		]

(* any atoms in rules needs to be considered as non-fresh *)
local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in 
fun matchForRewrite (pattern,atoms) term =
    let  val P0 = Xp "P"
	 val pattern_expr = actHatTerm P0 (term2expr pattern)
	 val term_expr = term2expr term

	 (* for any atom "c" not in the pattern, we add  "Pi(c) = C" where C is a fresh atom variable *)
	 val neededAtoms = LU.differenceByAll (atoms, T.atoms pattern)
	 val freshes = newAtomVars [] (L.length neededAtoms)
	 val eqsForAtoms = LP.map (fn (a,name) => U_Eq (Atomic (actAtom P0 (atom2expr a)), Atomic (Simp (Xa name)))) 
				  (neededAtoms,freshes)
     

	 val problem = [ U_Eq (pattern_expr, term_expr) ] @ eqsForAtoms
	 val _ = dprintln "matchForRewrite..."
	 val _ = dprintln ("input: " ^ (LU.toStringCommaCurly prUnifProblem problem))

   	 val atomsUsed = LU.toSet ((Term.atoms term) @ (Term.atoms pattern) @ atoms)

	 val solutions = match problem

        (* 使っていない Atom変数の式の削除 ...正しいか怪しいので pending *)
	val ans4 = solutions
	(* val ans4 = L.map (fn xs => let val avars2 = relevantAtoms xs ([], avars) *)
	(*    			   in L.filter (checkDomainOfIdProblem avars2) xs *)
	(*    			   end) *)
	(*    		 ans3 *)
	(* val _ = dprintln "AtomicSolve: delete redundant " *)
	val _ = L.app (dprintln o (LU.toStringCommaSquare prIdProblem)) ans4

        (* Atom変数の具体化 と Atom変数代入の除去 *)
	val ans5 = L.map (fn xs => L.map simplifyIdProblem (L.filter (fn x=> case x of I_Sub _=> false | _ => true)  xs))
			 (LU.mapAppend (instantiate1full atomsUsed) ans4)
	val ans5b = LU.toSet' (fn (xs,ys)=> LU.isEqset xs ys) ans5
	val _ = dprintln "matchForRewrite: instantiation:"
	val _ = L.app (dprintln o (LU.toStringCommaSquare prIdProblem)) ans5b

     (* delayed fresh (I_Dfresh of var * perm_expr * perm_expr)
        を freshness constraint (I_Fresh of (atom_expr * var)) へ変換 *)
     (* (a b)*X == P*X のような場合には失敗するが， 書き換えの場合にはそのようなのはないはず *) 
	fun transDelayedFresh (I_Dfresh (X, Pi, Psi)) =
	    (let val pi = perm_expr2perm Pi
		val psi = perm_expr2perm Psi
		val atms = Perm.disagreements (pi,psi)
	    in  L.map (fn a => I_Fresh (atom2expr a, X)) atms
	    end
	    handle TranslationFromExprToTermFailure => (print "Error: not an equivariant unification for rewriting ???\n";
							raise EqvUnifError))
	  | transDelayedFresh x = [x]
	val ans6 = L.map (LU.mapAppend transDelayedFresh) ans5b
	val _ = dprintln "AtomicSolve: translate delayed fresh cond."
	val _ = L.app (dprintln o (LU.toStringCommaSquare prIdProblem)) ans6

	 fun mkAns xs  =
	     let val _ = dprintln (LU.toStringCommaSquare prIdProblem xs)
		 val sigma = L.mapPartial (fn x => case x of I_Vmap (X,s) => SOME (X,term_expr2term s) 
							   | _ => NONE) xs
		 val env = L.mapPartial (fn x => case x of I_Fresh (alpha,X) => SOME (atom_expr2atom alpha, X) 
							 | _ => NONE) xs
		 val pimap = L.mapPartial (fn x => case x of I_Pmap (_,alpha,beta) => SOME (atom_expr2atom alpha, atom_expr2atom beta)
							   | _ => NONE) xs
	    (*  in (sigma,env,Perm.fromMapping pimap) *)
	     in (sigma,env,pimap)
	     end
(*    in L.map mkAns solutions *)
    in L.map mkAns ans6
    end
end (* of local *)


val _ = TU.test "matchForRewrite"
		[
                  (matchForRewrite (Term.fromString "(app <(lam [a][]X), []Y>)", [Atom.fromString "a"])
                                 (Term.fromString "(app <(lam [a](lam [b](app <b,a>))), b>)")
                   = [ (Subst.fromString "{Y := b, X := (lam [b](app <b,a>))}", [], [(Atom.fromString "a", Atom.fromString "a")]),
		       (Subst.fromString "{Y := b, X := (lam [a](app <a,b>))}", [], [(Atom.fromString "a", Atom.fromString "b")]),
		       (Subst.fromString "{X := (lam [b](app <b,c>)), Y := b}", [], [(Atom.fromString "a", Atom.fromString "c")]) ])
		,
                   (* Example 17 of [SKAT2015] *)
		  (matchForRewrite (Term.fromString "(lam [a](app <[]X,a>))", [Atom.fromString "a"])
                                   (Term.fromString "(lam [a](app <a,a>))")
		   = [ (Subst.fromString "{X := a}", [], [(Atom.fromString "a", Atom.fromString "a")]),
		       (Subst.fromString "{X := b}", [], [(Atom.fromString "a", Atom.fromString "b")]) ] )
		,
                  (* Example 19 of [SKAT2015] *)
		  (matchForRewrite (Term.fromString "(ucetaexp []X)", [Atom.fromString "a"])
                                   (Term.fromString "(ucetaexp a)")
		   = [ (Subst.fromString "{X := a}", [], [(Atom.fromString "a", Atom.fromString "a")]),
		       (Subst.fromString "{X := a}", [], [(Atom.fromString "a", Atom.fromString "b")]) ] )
		]


end (* of local *)
end (* of struct *)

				       
