(* file: fresh_env.sml *)
(* description: freshness environment and freshness constraint *)
(* author: Takahito Aoto *)

signature FRESH_ENV = 
sig 
    type atom = Atom.key
    type var = Var.key
    type perm = Perm.key
    type term = Term.key
    type fresh_cnstr = atom * var
    type fresh_env = fresh_cnstr list
    type key = fresh_env

    val toString: fresh_env -> string
    val toStringCurly: fresh_env -> string
    val fromString: string -> fresh_env
    val toList: fresh_env -> (atom * var) list

    val prFreshCnstr: fresh_cnstr -> string
    val rdFreshCnstr: string -> fresh_cnstr 
    val prTermInEnv: fresh_env * term -> string
    val rdTermInEnv: string -> fresh_env * term

    val atoms: fresh_env -> atom list
    val vars: fresh_env -> var list
    val empty: fresh_env
    val single: atom * var -> fresh_env
    val add: atom * var -> fresh_env -> fresh_env
    val incVarIndex: int -> fresh_env -> fresh_env
    val subsume: fresh_env -> fresh_env -> bool
    val act: perm -> fresh_env -> fresh_env
    val apply: (atom * atom) list -> fresh_env -> fresh_env
    val rename: (var * var) list -> fresh_env -> fresh_env

    val isFresh: fresh_env -> atom -> term -> bool
    val resolve: (atom * term) list -> fresh_env option

    structure IO : 
    sig 
      val freshseq: ParseNominal.token list -> fresh_env * ParseNominal.token list
      val fresh: ParseNominal.token list -> fresh_cnstr * ParseNominal.token list
      val env: ParseNominal.token list -> fresh_env * ParseNominal.token list
      val envterm: ParseNominal.token list -> (fresh_env * term) * ParseNominal.token list
    end

end

structure FreshEnv : FRESH_ENV =
struct 

local 
    structure A = Atom
    structure AL = AssocList
    structure L = List
    structure LU = ListUtil
    structure T = Term
    structure TU = TestUtil
    structure V = Var
in
type atom = Atom.key
type var = Var.key
type perm = Perm.key
type term = Term.key
type fresh_cnstr = atom * var
type fresh_env = fresh_cnstr list
type key = fresh_env

structure IO = 
struct
    open ParseNominal
    infix 6 $--; infix 5 --;  infix 3 >>; infix 0 ||;
    open Term.IO
    fun freshseq [] = ([],[])
      | freshseq toks =
        ( fresh -- "," $-- freshseq >> (fn (x,ys) => x::ys) || fresh >> (fn x => [x]) ) toks
    and fresh toks  =
        (id -- "#" $-- id  >> (fn (a,X) => (Atom.fromString a, Var.fromString X)) ) toks
    fun env toks =
        ( freshseq -- $ "|-" >> #1 || $ "|-" >> (fn _ => []) ) toks
    fun envterm toks =
        ( env -- term >> (fn xy => xy) ) toks
    fun fromString str = reader freshseq str
    fun rdFreshCnstr str = reader fresh str
    fun rdTermInEnv str = reader envterm str
end (* of struct *)

val fromString = IO.fromString
val rdFreshCnstr = IO.rdFreshCnstr
val rdTermInEnv = IO.rdTermInEnv

val _ = TU.test "fromString"
		[ (fromString "a#X" = [(A.fromString "a", V.fromString "X")]),
		  (fromString "a#X, b#Y" = [(A.fromString "a", V.fromString "X"), (A.fromString "b", V.fromString "Y")]),
		  (fromString "" = []) ]

fun prFreshCnstr (a,X) = (Atom.toString a) ^ "#" ^ (Var.toString X)
fun toString xs = LU.toStringComma prFreshCnstr xs
fun toStringCurly xs = LU.toStringCommaCurly prFreshCnstr xs
fun prTermInEnv (env,t) = (toString env) ^ " |- " ^ (Term.toString t)

val _ = TU.test "toString"
		[ (toString [(A.fromString "a", V.fromString "X"), (A.fromString "b", V.fromString "Y")] = "a#X,b#Y") ]

fun toList env = env
fun atoms env = L.map (fn (a,_) => a) env
fun vars env = L.map (fn (_,X) => X) env
val empty = []
fun single (a,X) = [(a,X)]
fun add (a,X) env = (a,X)::env

fun incVarIndex n env = L.map (fn (a,X) => (a,Var.incIndex n X)) env

val _ = TU.test "incVarIndex"
		[ (incVarIndex 2 (fromString "a#X,b#Y") = fromString "a#X_2,b#Y_2") ]

(* env1 subsumes env2 *)
fun subsume env1 env2 = L.all (fn x => LU.member x env1) env2

fun act pi env = L.map (fn (a,X) => (Perm.act pi a, X)) env

val _ = TU.test "act"
		[ (act (Perm.fromString "[(a b)]") (fromString "a#X,b#Y") = fromString "b#X,a#Y") ]

fun apply pimap env = L.map (fn (a,X) => (Perm.apply pimap a, X)) env

val _ = TU.test "apply"
		[ (apply [(Atom.fromString "a", Atom.fromString "b")] (fromString "a#X,b#Y") = fromString "b#X,b#Y") ]

fun rename vmap env = L.map (fn (a,X) => (a, case AL.find X vmap of SOME Y => Y | NONE=>X)) env

val _ = TU.test "rename"
		[ (rename [(Var.fromString "X", Var.fromString "Y")] (fromString "a#X,b#Y") = fromString "a#Y,b#Y") ]

(* env |- a # t *)
fun isFresh env a (T.Atom b) = (a <> b)
  | isFresh env a (T.Susp (pi,X)) = LU.member (Perm.act (Perm.inverse pi) a, X)  env
  | isFresh env a (T.Abst (b,t)) = (a = b) orelse (isFresh env a t)
  | isFresh env a (T.Fun (f,t)) = isFresh env a t
  | isFresh env a (T.Tuple ts) = L.all (isFresh env a) ts

val _ = TU.test "isFresh"
		[ (isFresh (fromString "a#X") (A.fromString "a") (T.fromString "[]X") = true),
		  (isFresh (fromString "b#X") (A.fromString "a") (T.fromString "[]X") = false),
		  (isFresh (fromString "b#X") (A.fromString "a") (T.fromString "[(a b)]X") = true),
		  (isFresh (fromString "") (A.fromString "a") (T.fromString "a") = false),
		  (isFresh (fromString "") (A.fromString "a") (T.fromString "b") = true),
		  (isFresh (fromString "") (A.fromString "a") (T.fromString "[a][]X") = true),
		  (isFresh (fromString "") (A.fromString "a") (T.fromString "<a, [a][]X>") = false),
		  (isFresh (fromString "") (A.fromString "a") (T.fromString "(f <a, [a][]X>)") = false)  ]

(* resolve freshness constratins *)
(* val resolve: (atom * term) list -> fresh_env option *)
fun resolve constraints = 
    let fun main [] ans = SOME (L.rev ans)
	  | main ((a,t)::ps) ans = 
	    case t of 	  
		T.Atom b => if (a <> b) then main ps ans else NONE
	      | T.Susp (pi,X) => main ps ((Perm.act (Perm.inverse pi) a, X)::ans)
	      | T.Abst (b,t) => if (a = b) then main ps ans else  (main ((a,t)::ps) ans)
	      | T.Fun (f,t)=> main ((a,t)::ps) ans
	      | T.Tuple ts=> main ((L.map (fn t=> (a,t)) ts) @ ps) ans
    in main constraints []
    end

val _ = TU.test "resolve"
		[ (resolve [(Atom.fromString "a", Term.fromString "(f <[]Y, [a][]X>)")] = SOME (fromString  "a#Y")),
		  (resolve [(Atom.fromString "a", Term.fromString "(f <[]Y, [b][]X>)")] = SOME (fromString  "a#Y,a#X")),
		  (resolve [(Atom.fromString "a", Term.fromString "(f <[]Y, [b][(a b)]X>)")] = SOME (fromString  "a#Y,b#X")) ]


end (* of local *)
end (* of struct *)

