(* file: main.sml *)
(* description: confluence prover for nominal rewriting systems *)
(* author: Takahito Aoto *)

structure Main = 
struct

local
  structure LU = ListUtil
  structure PU = PrintUtil
in

fun print_help () = 
    let val _ = print "nrbox v0.01\n"
	val _ = print "Usage: sml @SMLload=nrbox.x86-linux [option...] filename\n"
	val _ = print "Options:  -h           print this message\n"
	val _ = print "          -v           print logs\n"
	val _ = print "          -vv          print detailed logs\n"
	val _ = print "          -s           run silently\n"
	val _ = print "          -crit=num    test a criterion [num=1,..,4]\n"
    in () end

fun utime () =  Time.toMilliseconds (#usr(Timer.checkCPUTimer(Timer.totalCPUTimer())))
fun time f args =  let val starttime = utime ();
		       val result = f args;
		       val time = utime() - starttime
		   in (result, time) end

fun prepareDelayedOutput () = 
    let val inTmpFile = let val tmpName = ((#file (OS.Path.splitDirFile (OS.FileSys.tmpName ()))) ^ ".out")
			in OS.Path.joinDirFile {dir=".",file=tmpName}
			end
	val _ = PU.outs := TextIO.openOut inTmpFile
    in inTmpFile end

fun printDelayedOutput file = 
    let val _ = TextIO.closeOut (!PU.outs)
	val ins = TextIO.openIn file
	fun readline istream = case TextIO.inputLine istream of
				   SOME(line) => (print line;  readline istream)
                                 | NONE => ()
        val _ = readline ins
        val _ = TextIO.closeIn ins; 
	val _ = OS.FileSys.remove file
    in () end

fun main (name, args) =
    let open TextIO
	(* val _ = print (LU.toStringComma (fn x=>x) args) *)
	val (args0,args1) = LU.takeUntilWithRest (fn s => s = "--") args
	val help = List.exists (fn s => s = "-h") args0
	val silent = List.exists (fn s => s = "-s") args0
	val moreVerbose = (not silent) andalso List.exists (fn s => s = "-vv") args0
	val verbose = (not silent) andalso (List.exists (fn s => s = "-v") args0 orelse moreVerbose)
	val crit = List.exists (String.isPrefix  "-crit=") args0
	val filenames  = (List.filter (not o (String.isPrefix "-")) args0) @ args1
	fun readAll ins str =
	    if TextIO.endOfStream ins then str
	    else readAll ins (str ^ (TextIO.inputN (ins,1)))
    in if help orelse length filenames > 1 
       then (print_help (); OS.Process.success)
       else let val instream = if null filenames then stdIn else openIn (hd filenames)
		val _ = if silent 
			then (PU.runDebug := false; PU.runSilent := true)
			else ()
		val _ = if verbose
			then (Cr.runDebug := true;
			      Rewrite.runDebug := true;
			      PU.runSilent := false;
			      PU.runDebug := true;
			      Nrs.runDebug := true;
			      SnSolver.runDebug := true)
			else ()
		val _ = if moreVerbose 
			then (Equivariant.runDebug := true)
			else ()

		val outfile = if (not moreVerbose)
			      then prepareDelayedOutput ()
			      else "" 

		val rules = IOTpdb.rdFile (readAll instream "")

		val (result,time) = if not crit
				    then time Cr.prover rules
				    else let val arg0a = hd (List.filter (String.isPrefix  "-crit=") args0)
					     val len = String.size "-crit="
					     val expr  = String.substring (arg0a, len, String.size arg0a - len)
					 in case Int.fromString expr of
						SOME num => time (Cr.testMethod rules) num
					      | NONE => time (Cr.testMethod rules) 0
					 end
		val _ = print (result ^ "\n")  (* always show this *)
		val _ = PU.hintln ("time: " ^ (IntInf.toString time) ^ " msec.")

		val _ = if (not moreVerbose)
			then printDelayedOutput outfile
			else ()


	    in OS.Process.success
	    end 
	    handle Io => (print ("Error: failed to read the input");
			  if null filenames
			  then print "\n"
			  else print (" file: " ^ (hd filenames) ^ "\n");
			  OS.Process.success)

    end

end (* of local *)
end (* of struct *)
