(* file: nrs.sml *)
(* description: examples of nominal rewriting systems *)
(* author: Takahito Aoto *)

signature NRS = 
sig 
    type rules = Rule.rules
    val example: int -> rules
    val toString: rules -> string
    val breakVarCond: rules -> bool
    val isLeftLinear: rules -> bool
    val isLinear: rules -> bool
    val isUniform: rules -> bool
    val isAbstractSkeletonPreserving: rules -> bool
    val runDebug: bool ref
end

structure Nrs: NRS =
struct

local 
    structure L = List
    structure LU = ListUtil
    structure PU = PrintUtil
in

type rules = Rule.rules
val runDebug = ref false
fun dprint mesg = if (!runDebug) then PU.print mesg else ()
fun dprintln mesg = if (!runDebug) then PU.println mesg else ()

(* numbers 1-19 reflect Table 1 of  [RTA15] *)
fun example 1 = L.map Rule.fromString
                      (* alpha-red: alpha reduction rule of the lambda calculus *)
		      [ "b#X |- (lam [a][]X) -> (lam [b][(a b)]X)"]
  | example 2 (* 18 *) = L.map Rule.fromString
		     (* eta-red: eta reduction rule of the lambda calculus *)
		     ["a#X |- (lam [a](app <[]X,a>)) -> []X"]
  |  example 3 (* 9 *) = L.map Rule.fromString
                     (* eta-exp: eta expansion rule of the lambda calculus *)
			 ["a#X |- []X -> (lam [a](app <[]X,a>))"]
  |  example 4 (* 17 *) = L.map Rule.fromString
                      (* subst-std: explicit substitution calculus for lambda terms, 
                         Example 43 of [FG07, doi:10.1016/j.ic.2006.12.002]*)
			  [" |- (sub <[a](app <[]X,[]Y>),[]Z>) -> (app <(sub <[a][]X,[]Z>),(sub <[a][]Y,[]Z>)>)",
			   " |- (sub <[a]a,[]X>) -> []X",
			   "a#X |- (sub <[a][]X,[]Y>) -> []X",
			   "b#Y |- (sub <[a](lam [b][]X),[]Y>) -> (lam [b](sub <[a][]X,[]Y>))"]
  |  example 5 (* 8 *) = L.map Rule.fromString
                     (* beta-fg07-43: system of beta reduction for the lambda calculus via explicit substitution *)
                     (* Example 43 of [FG07, doi:10.1016/j.ic.2006.12.002] *)
			 [" |- (app <(lam [a][]X),[]Y>) -> (sub <[a][]X,[]Y>)",
			  " |- (sub <[a](app <[]X,[]Y>),[]Z>) -> (app <(sub <[a][]X,[]Z>),(sub <[a][]Y,[]Z>)>)",
			  " |- (sub <[a]a,[]X>) -> []X",
			  "a#X |- (sub <[a][]X,[]Y>) -> []X",
			  "b#Y |- (sub <[a](lam [b][]X),[]Y>) -> (lam [b](sub <[a][]X,[]Y>))"]
  |  example 6 (* 11 *) = L.map Rule.fromString
                    (* ml-fg07-43: fragment of ML, Example 43 of [FG07, doi:10.1016/j.ic.2006.12.002] *)
			 [" |- (app <(lam [a][]X),[]Y>) -> (sub <[a][]X,[]Y>)",
			  " |- (sub <[a](app <[]X,[]Y>),[]Z>) -> (app <(sub <[a][]X,[]Z>),(sub <[a][]Y,[]Z>)>)",
			  " |- (sub <[a]a,[]X>) -> []X",
			  "a#X |- (sub <[a][]X,[]Y>) -> []X",
			  "b#Y |- (sub <[a](lam [b][]X),[]Y>) -> (lam [b](sub <[a][]X,[]Y>))",
			  " |- (let <[]X',[a][]X>) -> (sub <[a][]X,[]X'>)",
			  " |- (letrec [f]<[a][]X',[]X>) -> (sub <[f][]X,(lam [a](letrec [f]<[a][]X',[]X'>))>)",
			  "a#Y |- (sub <[b](let <[]X',[a][]X>),[]Y>) -> (let <(sub <[b][]X',[]Y>),[a](sub <[b][]X,[]Y>)>)",
			  "f#Y, a#Y |- (sub <[b](letrec [f]<[a][]X',[]X>),[]Y>) -> (letrec [f]<[a](sub <[b][]X',[]Y>),(sub <[b][]X,[]Y>)>)"]
  |  example 7 (* 4 *) = L.map Rule.fromString
                       (* pnf-fg07-44: rules to obtain PNF of first-order formulas
                         Example 44 of [FG07] *)
			  ["a#P |- (and <[]P,(forall [a][]Q)>) -> (forall [a](and <[]P,[]Q>))",
			   "a#P |- (and <(forall [a][]Q),[]P>) -> (forall [a](and <[]Q,[]P>))",
			   "a#P |- (or <[]P,(forall [a][]Q)>) -> (forall [a](or <[]P,[]Q>))",
			   "a#P |- (or <(forall [a][]Q),[]P>) -> (forall [a](or <[]Q,[]P>))",
			   "a#P |- (and <[]P,(exists [a][]Q)>) -> (exists [a](and <[]P,[]Q>))",
			   "a#P |- (and <(exists [a][]Q),[]P>) -> (exists [a](and <[]Q,[]P>))",
			   "a#P |- (or <[]P,(exists [a][]Q)>) -> (exists [a](or <[]P,[]Q>))",
			   "a#P |- (or <(exists [a][]Q),[]P>) -> (exists [a](or <[]Q,[]P>))",
			   " |- (not (exists [a][]Q)) -> (forall [a](not []Q))",
			   " |- (not (forall [a][]Q)) -> (exists [a](not []Q))"]
  |  example 8 (* 16 *) = L.map Rule.fromString
                       (* pnf-fg07-44-addition: rules to obtain PNF of first-order formulas with additional rules, 
                                       Example 44 [FG07, doi:10.1016/j.ic.2006.12.002] *)
			  ["a#P |- (and <[]P,(forall [a][]Q)>) -> (forall [a](and <[]P,[]Q>))",
			   "a#P |- (and <(forall [a][]Q),[]P>) -> (forall [a](and <[]Q,[]P>))",
			   "a#P |- (or <[]P,(forall [a][]Q)>) -> (forall [a](or <[]P,[]Q>))",
			   "a#P |- (or <(forall [a][]Q),[]P>) -> (forall [a](or <[]Q,[]P>))",
			   "a#P |- (and <[]P,(exists [a][]Q)>) -> (exists [a](and <[]P,[]Q>))",
			   "a#P |- (and <(exists [a][]Q),[]P>) -> (exists [a](and <[]Q,[]P>))",
			   "a#P |- (or <[]P,(exists [a][]Q)>) -> (exists [a](or <[]P,[]Q>))",
			   "a#P |- (or <(exists [a][]Q),[]P>) -> (exists [a](or <[]Q,[]P>))",
			   " |- (not (exists [a][]Q)) -> (forall [a](not []Q))",
			   " |- (not (forall [a][]Q)) -> (exists [a](not []Q))",
			   "a#X |- (forall [a][]X) -> []X",
			   " |- (and <(forall [a][]X),(forall [a][]Y)>) -> (forall [a](and <[]X,[]Y>))",
			   "a#X |- (exists [a][]X) -> []X",
			   " |- (or <(exists [a][]X),(exists [a][]Y)>) -> (exists [a](or <[]X,[]Y>))"]
    |  example 9 (* 15 *)= L.map Rule.fromString
                    (* non-uniform-fg56: non-uniform system appearing in the proof of Lemma 56 of [FG07, doi:10.1016/j.ic.2006.12.002],
                       which has a trivial non-joinable critical pair *)
			  [" |- (f b) -> a", (* check !! **)
			   "a#X |- []X -> [a][]X"]
    |  example 10  (* 14 *) = L.map Rule.fromString
                    (* non-uniform-fg56-2: the second rule appearing in the proof of Lemma 56 of [FG07, doi:10.1016/j.ic.2006.12.002] *)
			  ["a#X |- []X -> [a][]X"]
    |  example 11  = L.map Rule.fromString
			   [ "a#X |- (lam [a](app <[]X,a>)) -> []X", 
			     "|- (app <(bot <>),[]Y>) -> (bot <>)" ]
    |  example 12  = L.map Rule.fromString
			   [ "a#X |- (lam [a](app <[]X,a>)) -> []X", 
			     "|- (app <(bot <>),[]Y>) -> (bot <>)",
			     "|- (lam [a](bot <>)) -> (bot <>)" ]
    |  example 13  = L.map Rule.fromString
			   [ "a#F |- (sigma <(0 <>), [a](app <[]F,a>)>) -> (0 <>)", 
			     "a#F |- (sigma <(s []N), [a](app <[]F,a>)>) -> (+ <(app <[]F, (s []N)>), (sigma <[]N, [a](app <[]F,a>)>)>)",
			     "|- (id []X) -> []X",
			     "|- (app <(id <>), []X>) -> (id []X)" ]
    |  example 14  = L.map Rule.fromString
			   [ "a#F |- (sigma <(0 <>), [a](app <[]F,a>)>) -> (0 <>)", 
			     "a#F |- (sigma <(s []N), [a](app <[]F,a>)>) -> (+ <(app <[]F, (s []N)>), (sigma <[]N, [a](app <[]F,a>)>)>)",
			     "|- (id []X) -> []X",
			     "|- (app <(id <>), []X>) -> (id []X)",
			     "|- (app <<>, []X>) -> []X",
			     "|- (sigma <(0 <>), [a]a>) -> (0 <>)",
			     "|- (sigma <(s []N), [a]a>) -> (+ <(s []N), (sigma <[]N, [a]a>)>)" ]
    |  example 15 (* 13 *) = L.map Rule.fromString
                    (* non-asp: a rule satisfying variable conditions but is not abstract skeleton preserving *)
			  [" |- (f []X) -> [a][]X"]
    |  example 16 (* 12 *) = L.map Rule.fromString
		   (* non-asp-fix: an abstract skeleton preserving rule *)
			  ["a#X |- (f []X) -> [a][]X"]
    | example 17 (* 1 *) = L.map Rule.fromString
                      (* subst-orthogonal: explicit substitution calculus for lambda terms, orthogonal version  *)
			  [" |- (sub <[a] (app <[]X,[]Y>),[]Z>) -> (app <(sub <[a][]X,[]Z>), (sub <[a][]Y,[]Z>)>)",
			   " |- (sub <[a]a,[]X>) -> []X",
			   " |- (sub <[a]b,[]X>) -> b",
			   "b#Y |- (sub <[a](lam [b][]X),[]Y>) -> (lam [b](sub <[a][]X,[]Y>))"]
   |  example 18 (* 6 *) = L.map Rule.fromString
                      (* beta-gsubst: system of beta reduction for the lambda calculus via orthogonal explicit substitution rules *)
			 [" |- (app <(lam [a][]X),[]Y>) -> (sub <[a][]X,[]Y>)",
			  " |- (sub <[a](app <[]X,[]Y>),[]Z>) -> (app <(sub <[a][]X,[]Z>), (sub <[a][]Y,[]Z>)>)",
			  " |- (sub <[a]a,[]X>) -> []X",
			  " |- (sub <[a]b,[]Y>) -> b",
			  "b#Y |- (sub <[a](lam [b][]X),[]Y>) -> (lam [b](sub <[a][]X,[]Y>))"]
    |  example 19 (* 7 *) = L.map Rule.fromString
                      (* beta-eta: system of beta eta reduction for the lambda calculus via orthogonal explicit substitution rules *)
		      [" |- (app <(lam [a][]X),[]Y>) -> (sub <[a][]X,[]Y>)",
		       " |- (sub <[a](app <[]X,[]Y>),[]Z>) -> (app <(sub <[a][]X,[]Z>),(sub <[a][]Y,[]Z>)>)",
		       " |- (sub <[a]a,[]X>) -> []X",
		       " |- (sub <[a]b,[]Y>) -> b",
		       "b#Y |- (sub <[a](lam [b][]X),[]Y>) -> (lam [b](sub <[a][]X,[]Y>))",
		       "a#X |- (lam [a](app <[]X,a>)) -> []X"]
    |  example 20 (* 19 *) = L.map Rule.fromString
                     (* ucond-eta: rule obtained from eta reduction rule of the lambda calculus by dropping condition *)
			  [" |- (lam [a](app <[]X,a>)) -> []X"]
    |  example 21 (* 3 *) = L.map Rule.fromString
                      (* ucond-eta-exp: non-confluent system, Example 19 of 
                         [SKAT15, doi:10.4230/LIPIcs.RTA.2015.301] *)
			  [" |- (ucetaexp []X) -> (lam [a](app <[]X,a>))"]
    |  example 22 (* 10 *) = L.map Rule.fromString
                     (* lambda-mu-subst-bhf01: structural substitution for lambda-mu-terms *)
			 [" |- (ssub <[a]b,c,[]X>) -> b",
			  "b#X |- (ssub <[a](lam [b][]Y),c,[]X>) -> (lam [b](ssub <[a][]Y,c,[]X>))",
			  " |- (ssub <[a](app <[]Y,[]Z>),b,[]X>) -> (app <(ssub <[a][]Y,b,[]X>),(ssub <[a][]Z,b,[]X>)>)",
			  "b#X |- (ssub <[a](mu [b][]Y),c,[]X>) -> (mu [b](ssub <[a][]Y,c,[]X>))",
			  " |- (ssub <[a](ind <a,[]Y>),b,[]X>) -> (ind <b,(app <(ssub <[a][]Y,b,[]X>),[]X>)>)",
			  " |- (ssub <[a](ind <b,[]Y>),c,[]X>) -> (ind <b,(ssub <[a][]Y,c,[]X>)>)"]
    |  example 23  = L.map Rule.fromString
			  [" |- (f []X) -> (f [a][]X)"]
    |  example 24  = L.map Rule.fromString
			  [" |- (not (and  <[]X, []Y>))  -> (or <(not []X), (not []Y)>)",
			   "b#X |- (not (forall [a][]X)) -> (exists [b](not [(b a)]X))" ]
    |  example 25 (* 20 *) = L.map Rule.fromString
                     (* com-forall: commutativity of universal quantifiers, Example 33 of [SKAT 16] *)
			  ["|- (forall [a](forall [b][]X)) -> (forall [b](forall [a][]X))"]
    |  example 26 (* 23 *) = L.map Rule.fromString
                     (* pnf-conj: pnf for first-order formulas with conjunction and universal quantifiers *)
			  ["a#Y_1 |- (and <(forall [a][]X_1),[]Y_1>) -> (forall [a](and <[]X_1,[]Y_1>))",
			   "a#X_2 |- (and <[]X_2,(forall [a][]Y_2)>) -> (forall [a](and <[]X_2,[]Y_2>))" ]
    |  example 27 (* 21 *) = L.map Rule.fromString
                     (* pnf-conj-com: pnf for first-order formulas with conjunction and universal quantifiers
                                plus commutativity of universal quantifiers, Example 12 of [SKAT 16] *)
			  ["a#Y_1 |- (and <(forall [a][]X_1),[]Y_1>) -> (forall [a](and <[]X_1,[]Y_1>))",
			   "a#X_2 |- (and <[]X_2,(forall [a][]Y_2)>) -> (forall [a](and <[]X_2,[]Y_2>))",
			   "|- (forall [a](forall [b][]X)) -> (forall [b](forall [a][]X))"]
    |  example 28 (* 22 *) = L.map Rule.fromString
                      (* nnf-neg: rules to obtain NNF of first-order formulas with negation and quantifers,
                         Example 29 of [SKAT 16] *)
			  [" |- (not (exists [a][]Q)) -> (forall [a](not []Q))",
			   " |- (not (forall [a][]Q)) -> (exists [a](not []Q))",
			   " |- (not (not []X)) -> []X"]
    |  example 29 (* 5 *) = L.map Rule.fromString
                      (* nnf: rules to obtain NNF of first-order formulas *)
			 [" |- (not (exists [a][]Q)) -> (forall [a](not []Q))",
			  " |- (not (forall [a][]Q)) -> (exists [a](not []Q))",
			  " |- (not (and <[]P,[]Q>)) -> (or <(not []P),(not []Q)>)",
			  " |- (not (or <[]P,[]Q>)) -> (and <(not []P), (not []Q)>)",
			  " |- (not (not []P)) -> []P"]
    |  example 30 (* 2 *) = L.map Rule.fromString
                      (* nnf-wodne: rules to obtain NNF of first-order formulas without DNE *)
			 [" |- (not (exists [a][]Q)) -> (forall [a](not []Q))",
			  " |- (not (forall [a][]Q)) -> (exists [a](not []Q))",
			  " |- (not (and <[]P,[]Q>)) -> (or <(not []P), (not []Q)>)",
			  " |- (not (or <[]P,[]Q>)) -> (and <(not []P), (not []Q)>)"]
(***************
  | example 101 = L.map Rule.fromString
                      (* subst-orthogonal: explicit substitution calculus for lambda terms, orthogonal version  *)
			  [" |- (sub <[a] (app <[]X,[]Y>),[]Z>) -> (app <(sub <[a][]X,[]Z>), (sub <[a][]Y,[]Z>)>)",
			   " |- (sub <[a]a,[]X>) -> []X",
			   " |- (sub <[a]b,[]X>) -> b",
			   "b#Y |- (sub <[a](lam [b][]X),[]Y>) -> (lam [b](sub <[a][]X,[]Y>))"]
    |  example 102 = L.map Rule.fromString
                      (* nnf-wodne: rules to obtain NNF of first-order formulas without DNE *)
			 [" |- (not (exists [a][]Q)) -> (forall [a](not []Q))",
			  " |- (not (forall [a][]Q)) -> (exists [a](not []Q))",
			  " |- (not (and <[]P,[]Q>)) -> (or <(not []P), (not []Q)>)",
			  " |- (not (or <[]P,[]Q>)) -> (and <(not []P), (not []Q)>)"]
    |  example 103 = L.map Rule.fromString
                      (* ucond-eta-exp: non-confluent system, Example 19 of 
                         [SKAT15, doi:10.4230/LIPIcs.RTA.2015.301] *)
			  [" |- (ucetaexp []X) -> (lam [a](app <[]X,a>))"]
    |  example 104 = L.map Rule.fromString
                       (* pnf-fg07-44: rules to obtain PNF of first-order formulas
                         Example 44 of [FG07] *)
			  ["a#P |- (and <[]P,(forall [a][]Q)>) -> (forall [a](and <[]P,[]Q>))",
			   "a#P |- (and <(forall [a][]Q),[]P>) -> (forall [a](and <[]Q,[]P>))",
			   "a#P |- (or <[]P,(forall [a][]Q)>) -> (forall [a](or <[]P,[]Q>))",
			   "a#P |- (or <(forall [a][]Q),[]P>) -> (forall [a](or <[]Q,[]P>))",
			   "a#P |- (and <[]P,(exists [a][]Q)>) -> (exists [a](and <[]P,[]Q>))",
			   "a#P |- (and <(exists [a][]Q),[]P>) -> (exists [a](and <[]Q,[]P>))",
			   "a#P |- (or <[]P,(exists [a][]Q)>) -> (exists [a](or <[]P,[]Q>))",
			   "a#P |- (or <(exists [a][]Q),[]P>) -> (exists [a](or <[]Q,[]P>))",
			   " |- (not (exists [a][]Q)) -> (forall [a](not []Q))",
			   " |- (not (forall [a][]Q)) -> (exists [a](not []Q))"]
    |  example 105 = L.map Rule.fromString
                      (* nnf: rules to obtain NNF of first-order formulas *)
			 [" |- (not (exists [a][]Q)) -> (forall [a](not []Q))",
			  " |- (not (forall [a][]Q)) -> (exists [a](not []Q))",
			  " |- (not (and <[]P,[]Q>)) -> (or <(not []P),(not []Q)>)",
			  " |- (not (or <[]P,[]Q>)) -> (and <(not []P), (not []Q)>)",
			  " |- (not (not []P)) -> []P"]
    |  example 106 = L.map Rule.fromString
                      (* beta-gsubst: system of beta reduction for the lambda calculus via orthogonal explicit substitution rules *)
			 [" |- (app <(lam [a][]X),[]Y>) -> (sub <[a][]X,[]Y>)",
			  " |- (sub <[a](app <[]X,[]Y>),[]Z>) -> (app <(sub <[a][]X,[]Z>), (sub <[a][]Y,[]Z>)>)",
			  " |- (sub <[a]a,[]X>) -> []X",
			  " |- (sub <[a]b,[]Y>) -> b",
			  "b#Y |- (sub <[a](lam [b][]X),[]Y>) -> (lam [b](sub <[a][]X,[]Y>))"]
    |  example 107 = L.map Rule.fromString
                      (* beta-eta: system of beta eta reduction for the lambda calculus via orthogonal explicit substitution rules *)
		      [" |- (app <(lam [a][]X),[]Y>) -> (sub <[a][]X,[]Y>)",
		       " |- (sub <[a](app <[]X,[]Y>),[]Z>) -> (app <(sub <[a][]X,[]Z>),(sub <[a][]Y,[]Z>)>)",
		       " |- (sub <[a]a,[]X>) -> []X",
		       " |- (sub <[a]b,[]Y>) -> b",
		       "b#Y |- (sub <[a](lam [b][]X),[]Y>) -> (lam [b](sub <[a][]X,[]Y>))",
		       "a#X |- (lam [a](app <[]X,a>)) -> []X"]
    |  example 108 = L.map Rule.fromString
                     (* beta-fg07-43: system of beta reduction for the lambda calculus via explicit substitution *)
                     (* Example 43 of [FG07, doi:10.1016/j.ic.2006.12.002] *)
			 [" |- (app <(lam [a][]X),[]Y>) -> (sub <[a][]X,[]Y>)",
			  " |- (sub <[a](app <[]X,[]Y>),[]Z>) -> (app <(sub <[a][]X,[]Z>),(sub <[a][]Y,[]Z>)>)",
			  " |- (sub <[a]a,[]X>) -> []X",
			  "a#X |- (sub <[a][]X,[]Y>) -> []X",
			  "b#Y |- (sub <[a](lam [b][]X),[]Y>) -> (lam [b](sub <[a][]X,[]Y>))"]
    |  example 109 = L.map Rule.fromString
                     (* eta-exp: eta expansion rule of the lambda calculus *)
			 ["a#X |- []X -> (lam [a](app <[]X,a>))"]
    |  example 110 = L.map Rule.fromString
                     (* lambda-mu-subst-bhf01: structural substitution for lambda-mu-terms *)
			 [" |- (ssub <[a]b,c,[]X>) -> b",
			  "b#X |- (ssub <[a](lam [b][]Y),c,[]X>) -> (lam [b](ssub <[a][]Y,c,[]X>))",
			  " |- (ssub <[a](app <[]Y,[]Z>),b,[]X>) -> (app <(ssub <[a][]Y,b,[]X>),(ssub <[a][]Z,b,[]X>)>)",
			  "b#X |- (ssub <[a](mu [b][]Y),c,[]X>) -> (mu [b](ssub <[a][]Y,c,[]X>))",
			  " |- (ssub <[a](ind <a,[]Y>),b,[]X>) -> (ind <b,(app <(ssub <[a][]Y,b,[]X>),[]X>)>)",
			  " |- (ssub <[a](ind <b,[]Y>),c,[]X>) -> (ind <b,(ssub <[a][]Y,c,[]X>)>)"]
    |  example 111 = L.map Rule.fromString
                    (* ml-fg07-43: fragment of ML, Example 43 of [FG07, doi:10.1016/j.ic.2006.12.002] *)
			 [" |- (app <(lam [a][]X),[]Y>) -> (sub <[a][]X,[]Y>)",
			  " |- (sub <[a](app <[]X,[]Y>),[]Z>) -> (app <(sub <[a][]X,[]Z>),(sub <[a][]Y,[]Z>)>)",
			  " |- (sub <[a]a,[]X>) -> []X",
			  "a#X |- (sub <[a][]X,[]Y>) -> []X",
			  "b#Y |- (sub <[a](lam [b][]X),[]Y>) -> (lam [b](sub <[a][]X,[]Y>))",
			  " |- (let <[]X',[a][]X>) -> (sub <[a][]X,[]X'>)",
			  " |- (letrec [f]<[a][]X',[]X>) -> (sub <[f][]X,(lam [a](letrec [f]<[a][]X',[]X'>))>)",
			  "a#Y |- (sub <[b](let <[]X',[a][]X>),[]Y>) -> (let <(sub <[b][]X',[]Y>),[a](sub <[b][]X,[]Y>)>)",
			  "f#Y, a#Y |- (sub <[b](letrec [f]<[a][]X',[]X>),[]Y>) -> (letrec [f]<[a](sub <[b][]X',[]Y>),(sub <[b][]X,[]Y>)>)"]
    |  example 112 = L.map Rule.fromString
                    (* non-asp-fix: an abstract skeleton preserving rule *)
			  ["a#X |- (f []X) -> [a][]X"]
    |  example 113 = L.map Rule.fromString
                    (* non-asp: a rule satisfying variable conditions but is not abstract skeleton preserving *)
			  [" |- (f []X) -> [a][]X"]
    |  example 114 = L.map Rule.fromString
                    (* non-uniform-fg56-2: the second rule appearing in the proof of Lemma 56 of [FG07, doi:10.1016/j.ic.2006.12.002] *)
			  ["a#X |- []X -> [a][]X"]
    |  example 115 = L.map Rule.fromString
                    (* non-uniform-fg56: non-uniform system appearing in the proof of Lemma 56 of [FG07, doi:10.1016/j.ic.2006.12.002],
                       which has a trivial non-joinable critical pair *)
			  [" |- (f b) -> a", (* check !! **)
			   "a#X |- []X -> [a][]X"]
    |  example 116 = L.map Rule.fromString
                       (* pnf-fg07-44-addition: rules to obtain PNF of first-order formulas with additional rules, 
                                       Example 44 [FG07, doi:10.1016/j.ic.2006.12.002] *)
			  ["a#P |- (and <[]P,(forall [a][]Q)>) -> (forall [a](and <[]P,[]Q>))",
			   "a#P |- (and <(forall [a][]Q),[]P>) -> (forall [a](and <[]Q,[]P>))",
			   "a#P |- (or <[]P,(forall [a][]Q)>) -> (forall [a](or <[]P,[]Q>))",
			   "a#P |- (or <(forall [a][]Q),[]P>) -> (forall [a](or <[]Q,[]P>))",
			   "a#P |- (and <[]P,(exists [a][]Q)>) -> (exists [a](and <[]P,[]Q>))",
			   "a#P |- (and <(exists [a][]Q),[]P>) -> (exists [a](and <[]Q,[]P>))",
			   "a#P |- (or <[]P,(exists [a][]Q)>) -> (exists [a](or <[]P,[]Q>))",
			   "a#P |- (or <(exists [a][]Q),[]P>) -> (exists [a](or <[]Q,[]P>))",
			   " |- (not (exists [a][]Q)) -> (forall [a](not []Q))",
			   " |- (not (forall [a][]Q)) -> (exists [a](not []Q))",
			   "a#X |- (forall [a][]X) -> []X",
			   " |- (and <(forall [a][]X),(forall [a][]Y)>) -> (forall [a](and <[]X,[]Y>))",
			   "a#X |- (exists [a][]X) -> []X",
			   " |- (or <(exists [a][]X),(exists [a][]Y)>) -> (exists [a](or <[]X,[]Y>))"]
    |  example 117 = L.map Rule.fromString
                      (* subst-std: explicit substitution calculus for lambda terms, 
                         Example 43 of [FG07, doi:10.1016/j.ic.2006.12.002]*)
			  [" |- (sub <[a](app <[]X,[]Y>),[]Z>) -> (app <(sub <[a][]X,[]Z>),(sub <[a][]Y,[]Z>)>)",
			   " |- (sub <[a]a,[]X>) -> []X",
			   "a#X |- (sub <[a][]X,[]Y>) -> []X",
			   "b#Y |- (sub <[a](lam [b][]X),[]Y>) -> (lam [b](sub <[a][]X,[]Y>))"]
    |  example 118 = L.map Rule.fromString
                     (* eta: eta reduction rule of the lambda calculus *)
			 ["a#X |- (lam [a](app <[]X,a>)) -> []X"]
    |  example 119 = L.map Rule.fromString
                     (* ucond-eta: rule obtained from eta reduction rule of the lambda calculus by dropping condition *)
			  [" |- (lam [a](app <[]X,a>)) -> []X"]
    |  example 120 = L.map Rule.fromString
                     (* com-forall: commutativity of universal quantifiers, Example 33 of [SKAT 16] *)
			  ["|- (forall [a](forall [b][]X)) -> (forall [b](forall [a][]X))"]
    |  example 121 = L.map Rule.fromString
                     (* pnf-conj-com: pnf for first-order formulas with conjunction and universal quantifiers
                                plus commutativity of universal quantifiers, Example 12 of [SKAT 16] *)
			  ["a#Y_1 |- (and <(forall [a][]X_1),[]Y_1>) -> (forall [a](and <[]X_1,[]Y_1>))",
			   "a#X_2 |- (and <[]X_2,(forall [a][]Y_2)>) -> (forall [a](and <[]X_2,[]Y_2>))",
			   "|- (forall [a](forall [b][]X)) -> (forall [b](forall [a][]X))"]
    |  example 122 = L.map Rule.fromString
                      (* nnf-neg: rules to obtain NNF of first-order formulas with negation and quantifers,
                         Example 29 of [SKAT 16] *)
			  [" |- (not (exists [a][]Q)) -> (forall [a](not []Q))",
			   " |- (not (forall [a][]Q)) -> (exists [a](not []Q))",
			   " |- (not (not []X)) -> []X"]
    |  example 123 = L.map Rule.fromString
                     (* pnf-conj: pnf for first-order formulas with conjunction and universal quantifiers *)
			  ["a#Y_1 |- (and <(forall [a][]X_1),[]Y_1>) -> (forall [a](and <[]X_1,[]Y_1>))",
			   "a#X_2 |- (and <[]X_2,(forall [a][]Y_2)>) -> (forall [a](and <[]X_2,[]Y_2>))" ]
***************)
    |  example _ = []

fun toString rs = LU.toStringCommaLnCurly Rule.toString rs

fun breakVarCond rs  = if L.exists Rule.breakVarCond rs
		       then (PU.hintln "variable conditions not hold"; true) 
		       else false

fun isLeftLinear rs  = if L.all Rule.isLeftLinear rs
		       then (PU.hintln "left-linear"; true) else (PU.hintln "not left-linear"; false)

fun isLinear rs  = if L.all Rule.isLinear rs
		   then (PU.hintln "linear"; true) else (PU.hintln "not linear"; false)

fun isUniform rs = if L.all Rule.isUniform rs
		   then (PU.hintln "uniform"; true) else (PU.hintln "not uniform"; false)

fun isAbstractSkeletonPreserving rs = if L.all Rule.isAbstractSkeletonPreserving rs
				      then (PU.hintln "abstract skeleton preserving"; true) 
				      else (PU.hintln "not abstract skeleton preserving"; false)

fun test () =
    let val _  = PU.println "isUniform"
	val list = L.tabulate (23, fn x=>x+1)
	val _ = L.app (fn rs => 
			  (print (toString rs);
			   if isUniform rs 
			   then PU.println "uniform: yes"
			   else PU.println "uniform: no"))
		      (L.map example list)
	val _  = PU.println "isASP"
	val _ = L.app (fn rs => 
			  (print (toString rs);
			   if isAbstractSkeletonPreserving rs 
			   then PU.println "ASP: yes"
			   else PU.println "ASP: no"))
		      (L.map example list)
    in ()
    end

(* val _ = test () *)

end (* of local *)

end


