(* file: perm.sml *)
(* description: permutation for nominal terms *)
(* author: Takahito Aoto *)

signature PERM = 
sig 

type atom  = Atom.key
type swap = atom * atom
type perm = swap list
type key = perm

val toString: perm -> string
val fromString: string -> perm 

val id: perm
val act: perm -> atom -> atom
val atoms: perm -> atom list
val equal: perm * perm -> bool
val isIdentity: perm -> bool
val compose: perm -> perm -> perm
val inverse: perm -> perm
val single: atom * atom -> perm
val normalize: perm -> perm
val disagreements: perm * perm -> atom list
val fromMapping: (atom * atom) list -> perm
val apply: (atom * atom) list -> atom -> atom
end

structure Perm: PERM =
struct 
local 
   structure A = Atom
   structure L = List
   structure LU = ListUtil
   structure PU = PrintUtil
   structure TU = TestUtil
in

type atom = A.key
type swap = atom * atom
type perm = swap list
type key = perm

fun toString []  = "[]"
  | toString ((a,b)::pi) = "[" ^ "(" ^ (A.toString a) ^ " "  ^ (A.toString b) ^ ")" ^ (toStringSub pi)
and toStringSub []  = "]"
  | toStringSub ((a,b)::pi) = "(" ^ (A.toString a) ^ " "  ^ (A.toString b) ^ ")" ^ (toStringSub pi)
val _ = TU.test "toString"
	     [(toString [(A.fromString "a", A.fromString "b")] = "[(a b)]"),
	       (toString [(A.fromString "a", A.fromString "b"),(A.fromString "c", A.fromString "d")] = "[(a b)(c d)]"),
	       (toString [] = "[]") ]

local
    open ParseNominal
    fun mkAtom (x,y) = (Atom.fromString x, Atom.fromString y)
    infix 6 $--; infix 5 --;  infix 3 >>; infix 0 ||;
    fun perm toks =
        ( "[" $-- swaplist >> (fn x => x) ) toks
    and swaplist toks =
        ( swapseq -- $ "]" >> #1  || $ "]"  >> (fn _ => []) ) toks 
    and swapseq toks =
        ( swap -- swapseq >> (fn (x,ys) => x::ys) || swap >> (fn x => [x]) ) toks
    and swap toks  =
        ("(" $-- id -- id -- $ ")"  >> (mkAtom o #1) ) toks
in 
fun fromString str = reader perm str
end (* of local *)
val _ = TU.test "fromString"
	     [(fromString "[(a b)]"  = [(A.fromString "a", A.fromString "b")]),
	      (fromString "[(a b)(c d)]"  = [(A.fromString "a", A.fromString "b"),(A.fromString "c", A.fromString "d")]),
	      (fromString "[]"  = [])]

val id = []

fun act [] a = a
  | act ((a,b)::pi) c = Atom.swap (a,b) (act pi c)
val _ = TU.test "act" 
	     [(act (fromString "[(a b)(c d)]") (A.fromString "a") =  (A.fromString "b")),
	      (act (fromString "[(a b)(c d)]") (A.fromString "b") =  (A.fromString "a")),
	      (act (fromString "[(a b)(c d)]") (A.fromString "c") =  (A.fromString "d")),
	      (act (fromString "[(a b)(c d)]") (A.fromString "d") =  (A.fromString "c")),
	      (act (fromString "[(a b)(c a)]") (A.fromString "a") =  (A.fromString "c")),
	      (act (fromString "[(a b)(c a)]") (A.fromString "c") =  (A.fromString "b")),
	      (act (fromString "[(a b)(c a)]") (A.fromString "b") =  (A.fromString "a"))]

fun atoms pi = L.foldl (fn ((a,b),cs) => LU.add a (LU.add b cs)) [] pi
val _ = TU.test "atoms"
	     [(atoms (fromString "[(a b)(c a)]") = [A.fromString "c",A.fromString "a",A.fromString "b"])]

fun isIdentity pi  = L.all (fn a => a = (act pi a)) (atoms pi)
val _ = TU.test "isIdentity"
	      [(isIdentity (fromString "[(a b)(a b)]") = true),
	       (isIdentity (fromString "[(a b)(b a)]") = true),
	       (isIdentity (fromString "[(a b)(c b)]") = false),
	       (isIdentity (fromString "[(a a)]") = true),
	       (isIdentity (fromString "[(a b)(c d) (b a)]") = false),
	       (isIdentity (fromString "[(a b)(c c) (b a)]") = true),
	       (isIdentity (fromString "[]") = true)]

fun equal (pi1,pi2) = L.all (fn a => (act pi1 a = act pi2 a)) (LU.union (atoms pi1, atoms pi2))
val _ = TU.test "equal" 
	     [(equal (fromString "[(a b)]", fromString "[(b a)]") = true),
	      (equal (fromString "[(a b)]", fromString "[(a a)]") = false),
	      (equal (fromString "[(a b)(c d)]", fromString "[(c d)(a b)]") = true),
	      (equal (fromString "[(a b)(a c)]", fromString "[(a c)(a b)]") = false)]

fun inverse pi = L.rev pi

fun single (a,b) = [(a,b)]

fun normalize pi = 
    let val atms = atoms pi
	val sorted = ListMergeSort.sort (fn (a,b) => Atom.compare (a,b) = GREATER) atms
	fun main (pi',ans) [] = ans
          | main (pi',ans) (a::bs) = 
	    let val c = act pi' a
	    in if a = c then main (pi',ans) bs
	       else (main (pi' @ (single (a,c)), (a,c)::ans) (LU.delete c bs))
	    end
    in main (pi,[]) sorted
    end

val _ = TU.test "normalize" 
	    [ (equal (normalize (fromString "[(a b)(a b)]"), normalize (fromString "[]")) = true),
	      (equal (normalize (fromString "[(a b)]"), normalize (fromString "[(a b)]")) = true),
	      (equal (normalize (fromString "[(b a)]"), normalize (fromString "[(a b)]")) = true),
	      (equal (normalize (fromString "[(a b)(a c)]"), normalize (fromString "[(a c)(c b)]")) = true),
	      (equal (normalize (fromString "[(b c)(a b)(a c)]"), normalize (fromString "[(b a)]")) = true) ]

fun compose pi1 pi2 = normalize (pi1 @ pi2)

fun disagreements (pi1, pi2) = 
    let val atms = LU.union (atoms pi1, atoms pi2)
    in L.filter (fn a => (act pi1 a) <> (act pi2 a)) atms
    end
val _ = TU.test "disagreements" 
	    [ (disagreements (fromString "[(a b)]", fromString "[(a b) (b c)]") =   [A.fromString "c", A.fromString "b"]) ]


(* We assume input is a valid permunation mapping *)
(* e.g. { a|->b, b|->a } =>  [(a,b)] *)
(* e.g. { a|->b, b|->c, c|->a } =>  [(a,b),(b,c)]  *)
local
    val debug = false
    fun dprint mesg = if debug then PU.print mesg else ()
    fun dprintln mesg = if debug then PU.println mesg else ()
in
fun fromMapping xs = 
    let val _ = dprint ("map: " ^ (LU.toStringCommaCurly (fn (x,y)=> (Atom.toString x) ^ "|->" ^ (Atom.toString y)) xs))
	fun replace (a,b) xs = L.filter (fn (x,y)=> x<>y) 
					(L.map (fn (x,y) => let val y' = if y=a then b else if y = b then a else y
							    in (x,y')
							    end) xs)
					       
	fun mkPerm []  = []
	  | mkPerm ((a,b)::xs) = if a = b 
				 then mkPerm xs
				 else (a, b) :: mkPerm (replace (a,b) xs)
	val ans = mkPerm xs
	val _ = dprintln (" ==> perm: " ^ (toString ans))
    in ans
    end
end

fun apply [] a = a
  | apply ((a,b)::pimap) c = if c = a then b else (apply pimap c)



end (* of local *)
end (* of struct *)

