(* file: rewrite.sml *)
(* description: rewriting *)
(* author: Takahito Aoto *)

signature REWRITE = 
sig 
    type atom = Atom.key
    type var = Var.key
    type term = Term.key
    type fresh_env = FreshEnv.key
    type perm = Perm.key
    type subst = Subst.subst
    type rule = Rule.rule
    type rules = Rule.rules

    val rewriteRoot: rule -> perm -> fresh_env * term -> term option
    val eqvRewriteRoot: rule  -> fresh_env * term -> term list

    val rootStep: rules  -> fresh_env * term -> term option
    val rootReducts: rules  -> fresh_env * term -> term list
    val reducts: rules  -> fresh_env * term -> term list
    val parallelReducts: rules  -> fresh_env * term -> term list

    val liStep: rules  -> fresh_env * term -> term option
    val isNF: rules -> fresh_env * term -> bool

    val printStepsToNF: (fresh_env * term -> term option) -> fresh_env * term -> unit
    val liStepsToNF: rules -> fresh_env * term -> unit

    val linf: rules -> fresh_env * term -> term

    (* val isJoinableByLinf: Trs.trs -> Term.term * Term.term -> bool *)
    (* val joinSequences: Trs.trs -> Term.term * Term.term  *)
    (* 		       -> ((Term.term * Term.term) list * (Term.term * Term.term) list) *)

    val runDebug: bool ref
end

structure Rewrite: REWRITE =
struct 
local 
    structure L = List
    structure S = Subst
    structure T = Term
    structure LU = ListUtil
    structure PU = PrintUtil
    structure TU = TestUtil
in

type atom = Atom.key
type var = Var.key
type term = Term.key
type fresh_env = FreshEnv.key
type perm = Perm.key
type subst = Subst.subst
type rule = Rule.rule
type rules = Rule.rules

exception RewriteError
val runDebug = ref false
fun dprint mesg = if (!runDebug) then PU.print mesg else ()
fun dprintln mesg = if (!runDebug) then PU.println mesg else ()


(* whether we need renaming of rules *)
fun varCheck (cntxt,l,r) term = 
    let val ruleVars = (Term.vars l) @ (Term.vars r) @ (FreshEnv.vars cntxt)
	val termVars = Term.vars term
    in LU.isDisjoint (termVars, ruleVars)
    end

(* rewriting at root position by a specified rule and permutation *)
(* rename rules before using the function *)
local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in
fun rewriteRoot (cntxt,l,r) Pi (env,term) = 
    let exception NF
	val _ = if varCheck (cntxt,l,r) term then ()
		else (dprintln "some variables appear in rules and terms";raise RewriteError)

	val isInst = Subst.match (Term.actHat Pi l) term 
	val _ = if not (isSome isInst) 
		then (dprintln "not an instance of lhs"; raise NF)
		else ()

	val (env2,sigma) = valOf isInst
	val _ = dprintln ("env2: " ^ (FreshEnv.toStringCurly env2))
	val _ = dprintln ("sigma: " ^ (Subst.toString sigma))

	val isCntxt = FreshEnv.resolve (Subst.instantiateFreshEnv sigma (FreshEnv.act Pi cntxt))
	val _ = if not (isSome isCntxt) 
		then (dprintln "instantiated context inconsistent"; raise NF)
		else ()

	val env3 = valOf isCntxt
	val _ = dprintln ("env3: " ^ (FreshEnv.toStringCurly env))

	val isSubsum = FreshEnv.subsume env (env2 @ env3)
	val _ = if not isSubsum
		then (dprintln "instantiated context does not guranteed in current environment"; raise NF)
		else ()
    in SOME (Subst.apply sigma (Term.actHat Pi r))
    end
    handle NF => NONE
end

val _ = TU.test "rewriteRoot"
		[
                  (rewriteRoot 
		       (Rule.fromString "|- (app <(lam [a][]X), []Y>) -> (sub <[a][]X, []Y>)")
		       Perm.id
		       (FreshEnv.rdTermInEnv "|- (app <(lam [a](lam [b](app <b,a>))), b>)")
		   = SOME (Term.fromString "(sub <[a](lam [b](app <b,a>)),b>)"))
		  ,
		    (rewriteRoot 
		       (Rule.fromString "b#Y |- (sub <[a](lam [b][]X), []Y>) -> (lam [b](sub <[a][]X, []Y>))")
		       Perm.id
		       (FreshEnv.rdTermInEnv "|- (sub <[a](lam [b](app <b,a>)),b>)")
		     = NONE)
		  ,
		    (rewriteRoot 
		       (Rule.fromString "b#Y |- (sub <[a](lam [b][]X), []Y>) -> (lam [b](sub <[a][]X, []Y>))")
		       (Perm.fromString "[(b c)]")
		       (FreshEnv.rdTermInEnv "|- (sub <[a](lam [b](app <b,a>)),b>)")
		     = SOME (Term.fromString "(lam [c](sub <[a](app <c,a>),b>))"))
		]


(* rewriting at root position by a specified rule and some permutation *)
(* rename rules before using the function *)
local
    val debug = false
    fun dprint mesg = if (* (!runDebug) orelse *) debug then PU.print mesg else ()
    fun dprintln mesg = if (* (!runDebug) orelse *) debug then PU.println mesg else ()
in
fun eqvRewriteRoot rule (env,term) = 
    let exception NF
	(* val _ = if varCheck (cntxt,l,r) term then ()
		else (dprintln "some variables appear in rules and terms";raise RewriteError) *)

	val (cntxt,l,r) = if varCheck rule term 
			  then let val vars = LU.union (FreshEnv.vars env, Term.vars term)
			       in Rule.renameFrom rule vars
			       end
			  else rule

	val atoms = LU.union (Term.atoms l, Term.atoms r)
	val matchers = 	Equivariant.matchForRewrite (l,atoms) term

	fun checkReducts (sigma,env2,pimap) = 
	    let	val _ = dprintln ("pimap: " ^ (LU.toStringCommaCurly (fn (a,b) => (Atom.toString a) ^ "|->" ^ (Atom.toString b)) pimap))
		val _ = dprintln ("env2: " ^ (FreshEnv.toStringCurly env2))
		val _ = dprintln ("sigma: " ^ (Subst.toString sigma))

		val _  = dprintln ("cntxt': " ^ (FreshEnv.toStringCurly (FreshEnv.apply pimap cntxt)))
		val isCntxt = FreshEnv.resolve (Subst.instantiateFreshEnv sigma (env2 @ (FreshEnv.apply pimap cntxt)))
		val _ = if not (isSome isCntxt) 
			then (dprintln "instantiated context inconsistent"; raise NF)
			else ()
		val env3 = valOf isCntxt

		val isSubsum = FreshEnv.subsume env env3
		val _ = if not isSubsum
			then (dprintln "instantiated context does not guranteed in current environment"; raise NF)
			else ()
	    in SOME (Subst.apply sigma (Term.apply pimap r))
	    end
	    handle NF => NONE

	val reducts = L.mapPartial checkReducts matchers
	(* val _ = if null reducts then () *)
	(* 	else dprintln ("reducts: " ^ (LU.toStringCommaCurly Term.toString reducts)) *)

	val ans = LU.toSet' (Equiv.aeq env) reducts
	val _ = if null ans then ()
		else dprintln ("ans: " ^ (LU.toStringCommaCurly Term.toString ans))

    in ans
    end
end

val _ = TU.test "eqvRewriteRoot"
		[
                  (eqvRewriteRoot 
		       (Rule.fromString "b#Y |- (sub <[a](lam [b][]X), []Y>) -> (lam [b](sub <[a][]X, []Y>))")
		       (FreshEnv.rdTermInEnv "|- (sub <[a](lam [b](app <b,a>)),b>)")
		   = [ (Term.fromString "(lam [d](sub <[c](app <d,c>),b>))") ])
		  ]


(* 根位置での，ルール集合による書き換え *)
fun rootStep [] (env,term) = NONE
  | rootStep ((cntxt,l,r)::rs) (env,term) =
    case eqvRewriteRoot (cntxt,l,r) (env,term) of
	[] => rootStep rs (env,term)
      | (x::_) => SOME x

val _ = TU.test "rootStep"
		[ 
		  (rootStep (Nrs.example 18) 
			       (FreshEnv.rdTermInEnv "|- (sub <[a](lam [b](app <b,a>)),b>)")
		   = SOME (Term.fromString "(lam [d](sub <[c](app <d,c>),b>))")) 
		  ]

(* 根位置での，ルール集合による書き換え先 *)
fun rootReducts rs (env,term) = LU.mapUnion' (Equiv.aeq env) (fn rule => eqvRewriteRoot rule (env,term)) rs

val _ = TU.test "rootReducts"
		[  (* Example 17 of [SKAT RTA2015] *)
		  (rootReducts (Nrs.example 20)
                                  (FreshEnv.rdTermInEnv "|- (lam [a](app <a,a>))")
		   = [Term.fromString "a",  Term.fromString "b"])
		,  (* Example 19 of [SKAT RTA2015] *)
		  (rootReducts (Nrs.example 21)
                                  (FreshEnv.rdTermInEnv "|- (ucetaexp a)")
		   = [Term.fromString "(lam [a](app <a,a>))", Term.fromString "(lam [b](app <a,b>))"])
		, (* Example 9 of [SKAT SCSS2016] *)
		  (rootReducts (Nrs.example 26)
                                  (FreshEnv.rdTermInEnv "|- (and <(forall [a](p a)), (forall [a](q a))>)")
		   = [Term.fromString "(forall [b](and <(p b),(forall [a](q a))>))",
		      Term.fromString "(forall [b](and <(forall [a](p a)),(q b)>))"])
		, (* Example 12 of [SKAT SCSS2016] *)
		  (rootReducts (Nrs.example 27)
                           (FreshEnv.rdTermInEnv "|- (and <(forall [a][]X_1), (forall [a][]Y_2)>)")
		   =  [Term.fromString "(forall [a](and <[]X_1,(forall [a][]Y_2)>))",
		       Term.fromString "(forall [a](and <(forall [a][]X_1),[]Y_2>))"])
		]

(* the results of one step rewriting *)
fun reducts rs (env,term) = 
    let val vars = LU.union (FreshEnv.vars env, Term.vars term)
	val rs' = L.map (fn rule => Rule.renameFrom rule vars) rs
	fun main rs (env,t) = LU.toSet (rootReducts rs (env,t) @ argReducts rs (env,t))
	and argReducts rs (env, T.Atom a) = []
	  | argReducts rs (env, T.Susp (pi,X)) = []
	  | argReducts rs (env, T.Abst (a,t)) =  L.map (fn s => T.Abst (a,s)) (main rs (env,t))
	  | argReducts rs (env, T.Fun (f,t))  = L.map (fn s => T.Fun (f,s)) (main rs (env,t))
	  | argReducts rs (env, T.Tuple ts) = L.map (fn us=> T.Tuple us) (reducts_list rs (env,ts))
	and reducts_list rs (env, []) = []
	  | reducts_list rs (env, [t]) = L.map (fn u=>[u]) (main rs (env,t))
	  | reducts_list rs (env, (t::ts)) = L.map (fn u=>u::ts) (main rs (env,t))
					    @ L.map (fn us=>t::us) (reducts_list rs (env,ts))
    in LU.toSet' (Equiv.aeq env) (main rs' (env,term))
    end

val _ = TU.test "reducts"
		[ (* Example 9 of [SKAT SCSS2016] *)
		  (reducts (Nrs.example 26)
                           (FreshEnv.rdTermInEnv "|- (forall [b](and <(p b),(forall [a](q a))>))")
		   = [Term.fromString "(forall [b](forall [c](and <(p b),(q c)>)))"])
		, 
		  (reducts (Nrs.example 27)
                           (FreshEnv.rdTermInEnv "|- (and <(forall [a] (forall [b](p <a,b>))), q>)")
		   = [Term.fromString "(forall [c](and <(forall [b](p <c,b>)),q>))",
		      Term.fromString "(and <(forall [d](forall [c](p <c,d>))),q>)"])
		, 
		  (reducts (Nrs.example 28)
                           (FreshEnv.rdTermInEnv "|- (not (not (forall [a][]X)))")
		   = [Term.fromString "(forall [a][]X)",
		      Term.fromString "(not (exists [a](not []X)))"])
		]

(* the results of one step parallel rewriting *)
fun parallelReducts rs (env,term) = 
    let val vars = LU.union (FreshEnv.vars env, Term.vars term)
	val rs' = L.map (fn rule => Rule.renameFrom rule vars) rs
(*	fun main rs (env,t) = LU.toSet ([t] @ rootReducts rs (env,t) @ argReducts rs (env,t)) *)
	fun main rs (env,t) = rootReducts rs (env,t) @ (argReducts rs (env,t))
	and argReducts rs (env, T.Atom a) = [T.Atom a]
	  | argReducts rs (env, T.Susp (pi,X)) = [T.Susp (pi,X)]
	  | argReducts rs (env, T.Abst (a,t)) =  L.map (fn s => T.Abst (a,s)) (main rs (env,t))
	  | argReducts rs (env, T.Fun (f,t))  = L.map (fn s => T.Fun (f,s)) (main rs (env,t))
	  | argReducts rs (env, T.Tuple ts) = L.map (fn us=> T.Tuple us) (reducts_list rs (env,ts))
	and reducts_list rs (env, []) = [[]]
	  | reducts_list rs (env, (t::us)) = let val ts = [t] @ (main rs (env,t))
						 val uss = reducts_list rs (env,us)
					     in ListXProd.mapX (fn (t1,us1) => t1::us1) (ts,uss) end
    in LU.toSet' (Equiv.aeq env) (main rs' (env,term))
    end

val _ = TU.test "parallelReducts"
		[
		  (parallelReducts (Nrs.example 16)
                           (FreshEnv.rdTermInEnv "|- <(f a), (f b)>")
		   (* = map Term.fromString ["<(f a),(f b)>","<(f a),[c]b>","<[b]a,(f b)>","<[b]a,[c]b>"]) *)
		   = map Term.fromString ["<[b]a,[c]b>","<[b]a,(f b)>","<(f a),[c]b>","<(f a),(f b)>"])
		,
		  (parallelReducts []
                           (FreshEnv.rdTermInEnv "|- (f a)")
		   = [Term.fromString "(f a)"])
		]

(* liStep rs term : TRS rs による最左最内位置での書き換え *)
fun liStep rs (env,term) =
    let val vars = LU.union (FreshEnv.vars env, Term.vars term)
	val rs' = L.map (fn rule => Rule.renameFrom rule vars) rs
	fun rewrite (T.Susp _) = NONE
	  | rewrite (t0 as (T.Atom a)) = rootStep rs' (env,t0)
	  | rewrite (t0 as (T.Abst (a,t))) = 
	    (case rewrite t of NONE => rootStep rs' (env,t0) | SOME s => SOME (T.Abst (a,s)))
	  | rewrite (t0 as (T.Fun (f,t))) = 
	    (case rewrite t of NONE => rootStep rs' (env,t0) | SOME s => SOME (T.Fun (f,s)))
	  | rewrite (t0 as (T.Tuple ts)) = 
	    (case rewritelist ts of NONE => rootStep rs' (env,t0) | SOME us => SOME (T.Tuple us))
	and rewritelist [] = NONE
	  | rewritelist (t::ts) = 
	    (case rewrite t of
		 SOME u => SOME (u::ts)
	       | NONE => (case rewritelist ts of SOME us => SOME (t::us) | NONE => NONE))
    in rewrite term end

val _ = TU.test "liStep"
		[ 
		  (liStep (Nrs.example 6) 
			     (FreshEnv.rdTermInEnv "|- (app <(lam [a](lam [b](app <b,a>))), b>)")
		   = SOME (Term.fromString "(sub <[c](lam [b](app <b,c>)),b>)"))
		,
		  (liStep (Nrs.example 6) 
			     (FreshEnv.rdTermInEnv "|- (sub <[c](lam [b](app <b,c>)),b>)")
	(*	   = SOME (Term.fromString "(lam [a](sub <[d](app <a,d>),b>))")) *)
		   = SOME (Term.fromString "(lam [e](sub <[d](app <e,d>),b>))"))
		,
		  (liStep (Nrs.example 6) 
			     (FreshEnv.rdTermInEnv "|- (lam [d](sub <[a](app <d,a>),b>))")
		   = SOME (Term.fromString "(lam [d](app <(sub <[c]d,b>),(sub <[c]c,b>)>))"))
		,
		  (liStep (Nrs.example 6) 
			     (FreshEnv.rdTermInEnv "|- (lam [d](app <(sub <[c]d,b>),(sub <[c]c,b>)>))")
		   = SOME (Term.fromString "(lam [d](app <d,(sub <[c]c,b>)>))"))
		,
		  (liStep (Nrs.example 6) 
			     (FreshEnv.rdTermInEnv "|- (lam [d](app <d,(sub <[c]c,b>)>))")
		   = SOME (Term.fromString "(lam [d](app <d,b>))"))
		]

fun printStepsToNF stepfun (env,term) = 
    let val init = "    " ^ (FreshEnv.toString env) ^ "|-" ^ (T.toString term) ^ "\n"
        fun steps t =  case stepfun (env,t) of
                           SOME u => (PU.print ("->R " ^  (T.toString u) ^ "\n");
                                      steps u)
                         | NONE => ()
    in (PU.print init; steps term)
    end

fun liStepsToNF rs (env,term) = printStepsToNF (liStep rs) (env,term)

fun isNF rs (env,term) = not (isSome (liStep rs (env, term)))

val _ = TU.test "isNF"
		[ (* Example 12 of [SKAT SCSS2016] *)
		  (isNF (Nrs.example 21)
                        (FreshEnv.rdTermInEnv "|- (forall [a](and <[]X_1,(forall [a][]Y_2)>))")
		   = true)
		, (* Example 12 of [SKAT SCSS2016] *)
		  (isNF (Nrs.example 21)
                        (FreshEnv.rdTermInEnv "|- (forall [a](and <(forall [a][]X_1),[]Y_2>))")
		   = true)
		  ]


(* 最左最内(leftmost innermost)戦略で正規形(normal form)を求める *) 
fun linf rs (env,term) = 
    let val result = liStep rs (env,term)
    in case result of NONE => term
		    | SOME term' => linf rs (env,term')
    end

val _ = TU.test "linf"
		[ 
		  (linf (Nrs.example 6) 
			     (FreshEnv.rdTermInEnv "|- (app <(lam [a](lam [b](app <b,a>))), b>)")
		   (* = Term.fromString "(lam [a](app <a,b>))") *)
		   = Term.fromString "(lam [e](app <e,b>))")
		]

fun isJoinableByLinf rs (env,s,t) = 
    let val s1 = linf rs (env,s)
 	val t1 = linf rs (env,t)
    in s1 = t1 end 





end (* of local *)

end
