(* file: rule.sml *)
(* description: nominal rewriting rules *)
(* author: Takahito Aoto *)

signature RULE = 
sig 
    type atom = Atom.key
    type var = Var.key
    type term = Term.key
    type fresh_env = FreshEnv.key
    type rule = fresh_env * term * term
    type key = rule
    type rules = rule list

    val toString: rule -> string
    val fromString: string -> rule
    val prRules: rules -> string

    val renameFrom: rule -> var list -> rule
    val rename: rule * rule -> rule * rule
    val breakVarCond: rule -> bool
    val isLeftLinear: rule -> bool
    val isLinear: rule -> bool
    val isUniform: rule -> bool
    val isAbstractSkeletonPreserving: rule -> bool

    val areEquivariant: rule * rule -> bool

    structure IO : 
    sig 
      val rule: ParseNominal.token list -> rule * ParseNominal.token list
    end

    (* val NRSorthsubst: rules *)
end

structure Rule: RULE =
struct

local 
    structure L = List
    structure T = Term
    structure LP = ListPair
    structure LU = ListUtil
    structure PU = PrintUtil
    structure TU = TestUtil
in

type atom = Atom.key
type var = Var.key
type term = Term.key
type fresh_env = FreshEnv.key
type rule = fresh_env * term * term
type key = rule
type rules = rule list

fun toString (env,l,r) = (FreshEnv.toString env) ^ " |- " ^ (Term.toString l) ^ " -> " ^ (Term.toString r)
fun prRules rs = LU.toStringCommaLnSquare toString rs

structure IO = 
struct
    open ParseNominal
    infix 6 $--; infix 5 --;  infix 3 >>; infix 0 ||;
    open Term.IO
    open FreshEnv.IO
    fun rule toks =
        ( env -- term -- "->" $-- term >> (fn ((x,y),z) => (x,y,z)) ) toks
    fun fromString str = reader rule str
end (* of struct *)

val fromString = IO.fromString

val _ = TU.test "toString"
		[ (toString (fromString " |- (f b) -> a") = " |- (f b) -> a"),
                  (toString (fromString "a#X |- []X -> [a][]X") = "a#X |- []X -> [a][]X"),
                 (* spaces around "|-" and "->" are necessary *)
                  (toString (fromString "a#X, b#Y |- []X -> <[a][]X,[b][]X>") = "a#X,b#Y |- []X -> <[a][]X,[b][]X>") ]

fun renameFrom (env,l,r) vars = 
    let val idx = 1 + (List.foldr Int.max 0 (List.map Var.index vars))
    in (FreshEnv.incVarIndex idx env, Term.incVarIndex idx l, Term.incVarIndex idx r)
    end

val _ = TU.test "renameFrom"
		[ (renameFrom (fromString "a#X |- []X -> [a][]X") [(Var.fromString  "Y_1")] 
		   = (fromString "a#X_2 |- []X_2 -> [a][]X_2")),
		  (renameFrom (fromString "a#X_2 |- <[]X_3, []X_2> -> []X_3") [(Var.fromString  "Y_1"),(Var.fromString  "X_2")]
		   = (fromString "a#X_5 |- <[]X_6,[]X_5> -> []X_6")) ]

fun rename ((env1,l1,r1),(env2,l2,r2)) =
    let val vs1 = LU.union (LU.union (FreshEnv.vars env1, Term.vars l1), Term.vars r1)
	val idx = 1 + (List.foldr Int.max 0 (List.map Var.index vs1))
	val (env2b,l2b,r2b) = (FreshEnv.incVarIndex idx env2, Term.incVarIndex idx l2, Term.incVarIndex idx r2)
    in ((env1,l1,r1),(env2b,l2b,r2b))
    end

val _ = TU.test "rename"
		[ (rename (fromString "a#X |- []X -> [a][]X", fromString "a#X |- []X -> [a][]X")
		   = (fromString "a#X |- []X -> [a][]X", fromString "a#X_1 |- []X_1 -> [a][]X_1")),
		  (rename (fromString "a#X_5 |- <[]X_6,[]X_5> -> []X_6", fromString "a#Y |- []Y -> [a][]Y")
		   = (fromString "a#X_5 |- <[]X_6,[]X_5> -> []X_6", fromString "a#Y_7 |- []Y_7 -> [a][]Y_7")) ]

(* equal by ()^\pi and renaming of var?  for env, set equal is OK *)
fun areEquivariant ((env1,l1,r1), (env2,l2,r2)) =
    let val atms1 = LU.toSet ((FreshEnv.atoms env1) @ (Term.atoms l1) @ (Term.atoms r1))
	val atms2 = LU.toSet ((FreshEnv.atoms env2) @ (Term.atoms l2) @ (Term.atoms r2))
	val vrs1 = LU.toSet ((FreshEnv.vars env1) @ (Term.vars l1) @ (Term.vars r1))
	val vrs2 = LU.toSet ((FreshEnv.vars env2) @ (Term.vars l2) @ (Term.vars r2))
    in if length atms1 <> length atms2
	  orelse length vrs1 <> length vrs2
       then false
       else let val natms = Atom.newAtoms [] (length atms1)
		val aren1 = LP.zip (atms1,natms)
		val aren2 = LP.zip (atms2,natms)
		val nvrs = Var.newVars [] (length vrs1)
		val vren1 = LP.zip (vrs1,nvrs)
		val vren2 = LP.zip (vrs2,nvrs)
		val env1a = FreshEnv.rename vren1 (FreshEnv.apply aren1 env1)
		val env2a = FreshEnv.rename vren2 (FreshEnv.apply aren2 env2)
		val sub1 = LP.zip (vrs1,L.map (fn x => T.Susp ([],x)) nvrs)
		val sub2 = LP.zip (vrs2,L.map (fn x => T.Susp ([],x)) nvrs)
		val l1a = Subst.apply sub1 (Term.apply aren1 l1)
		val r1a = Subst.apply sub1 (Term.apply aren1 r1)
		val l2a = Subst.apply sub2 (Term.apply aren2 l2)
		val r2a = Subst.apply sub2 (Term.apply aren2 r2)
	    in (l1a,r1a) = (l2a,r2a) andalso LU.isEqset env1a env2a
	    end
    end

val _ = TU.test "areEquivariant"
		[ 
		  (areEquivariant (fromString "b#Y,a#Y |- (f <[b][]Y,[a][]Y>) -> [b][a][]Y",
				   fromString "a#Y,b#Y |- (f <[a][]Y,[b][]Y>) -> [a][b][]Y")
		   = true)
		,
		  (areEquivariant (fromString "b#Y,a#Y |- (f <[b][]Y,[a][]Y>) -> [b][a][]Y",
				   fromString "b#Y,c#Y |- (f <[c][]Y,[b][]Y>) -> [c][b][]Y")
		   = true)
		,
		  (areEquivariant (fromString "b#Y,b#Z,a#Y |- (f <[b][]Y,[a][]Z>) -> [b][a][]Y",
				   fromString "c#X,a#X,c#Z |- (f <[c][]X,[a][]Z>) -> [c][a][]X")
		   = true)
		  ]


fun breakVarCond (env,l,r) = (Term.isSusp l) orelse (not (LU.isSubset (Term.vars r) (Term.vars l)))

fun isLeftLinear (env,l,r) = LU.isDistinct (Term.varlist l)

fun isLinear (env,l,r) = LU.isDistinct (Term.varlist l) andalso LU.isDistinct (Term.varlist r)

val _ = TU.test "isLeftLinear"
		[ (isLeftLinear (fromString " |- []X -> [a][]X") = true),
		  (isLeftLinear (fromString " |- <[]X,[]X> -> [a][]X") = false),
		  (isLeftLinear (fromString " |- <a, a, []X> -> [a][]X") = true),
		  (isLeftLinear (fromString " |- (f <[]X,[]X>) -> [a][]X") = false),
		  (isLeftLinear (fromString " |- (f <[]X,[]X_1>) -> [a][]X") = true),
		  (isLeftLinear (fromString " |- (f <[]X,[]X_1>) -> <[]X, [a][]X>") = true) ]

local
    val debug = false
    fun dprint mesg = if debug then PU.print mesg else ()
    fun dprintln mesg = if debug then PU.println mesg else ()
in
fun isUniform (env,l,r) =
    let val atoms = LU.union (LU.union (FreshEnv.atoms env, Term.atoms l), Term.atoms r)
	val new = Atom.newAtom atoms
	val _  = dprintln (LU.toStringCommaCurly Atom.toString (new::atoms))
	fun resolve a (env,l,r) =
	    let val _ = dprintln ("check: " ^ (Atom.toString a) ^ " for " ^ (toString (env,l,r)))
		val isAfreshInL = FreshEnv.resolve [(a,l)]
	    in case isAfreshInL of
		   NONE => true
		 | SOME delta => let val isAfreshInR = FreshEnv.resolve [(a,r)]
				 in case isAfreshInR of
					NONE => false
				      | SOME rho => FreshEnv.subsume (env @ delta) rho
				 end
	    end
    in L.all (fn a => resolve a  (env,l,r)) (new::atoms) 
   end
end

val _ = TU.test "isUniform"
		[ (isUniform (fromString " |- (f b) -> a") = false),
		  (isUniform (fromString " |- [b]b -> b") = false),
		  (isUniform (fromString " a#X |- []X -> [a][]X") = true) ]

local
    val debug = false
    fun dprint mesg = if debug then PU.print mesg else ()
    fun dprintln mesg = if debug then PU.println mesg else ()
in
fun isAbstractSkeletonPreserving (env,l,r) =
    let fun isNullPermuations t = L.all (fn (pi,_) => Perm.isIdentity pi) (Term.susps t)
	val varSkelsInL = Term.varSkeletons l
	val varSkelsInR = Term.varSkeletons r
	fun checkASPprim env (rskel,lskel) =  
	    let val _ = dprintln ((Term.toString rskel) ^ " ... " ^ (Term.toString lskel))
		val latoms = Term.atoms lskel
		val lvars = Term.vars lskel
		(* val _ = dprintln (LU.toStringCommaCurly Var.toString lvars) *)
		(* val _ = dprintln (LU.toStringCommaCurly Atom.toString latoms) *)
		val ratoms = Term.atoms rskel
		val rvars = Term.vars rskel
		(* val _ = dprintln (LU.toStringCommaCurly Atom.toString ratoms) *)
		(* val _ = dprintln (LU.toStringCommaCurly Var.toString rvars) *)
		(* val _ = dprintln (" compare ") *)
		val lvar = hd (Term.vars lskel)
		val rvar = hd (Term.vars rskel)
	    in lvar = rvar
	       andalso L.all (fn a => FreshEnv.subsume env [(a,lvar)]) (LU.differenceByAll (latoms,ratoms))
	       andalso L.all (fn a => FreshEnv.subsume env [(a,lvar)]) (LU.differenceByAll (ratoms,latoms))
	    end
	fun checkASP env rskel lskels = L.exists (fn lskel => checkASPprim env (rskel,lskel)) lskels
    in isNullPermuations l
       andalso isNullPermuations r
       andalso LU.isSubset (Term.freeAtoms r) (Term.freeAtoms l) 
       andalso Term.nonDuplicatingSkeletons l
       andalso Term.nonDuplicatingSkeletons r
       andalso L.all (fn rskel => checkASP env rskel varSkelsInL) varSkelsInR
    end
end

val _ = let val isASP = isAbstractSkeletonPreserving
	in TU.test "isAbstractSkeletonPreserving"
		[ (isASP (fromString "b#Y |- (sub <[a](lam [b][]X),[]Y>) -> (lam [b](sub <[a][]X,[]Y>))") = true),
		  (isASP (fromString " |- (ucetaexp []X) -> (lam [a](app <[]X,a>))") = false),
		  (isASP (fromString "a#X |- (lam [a](app <[]X,a>)) -> []X") = true),
		  (isASP (fromString " |- (lam [a](app <[]X,a>)) -> []X") = false),
		  (isASP (fromString " |- [a][b][a][]X -> [a][b][]X") = false),
		  (isASP (fromString " |- [a][b][]X -> [b][a][(a b)]X") = false),
		  (isASP (fromString " |- [a][b][(a b)]X -> [b][a][]X") = false),
		  (isASP (fromString "c#X,d#X |- [a][b][c][]X -> [d][b][a][]X") = true),
		  (isASP (fromString "c#X |- <[a][b][]X,[b][c][]X> -> [b][]X") = true) ]
	end


end (* of local *)

end


