(* file: sn_solver.sml *)
(* description: termination by external snprover for TRSs *)
(* author: Takahito Aoto *)

signature SN_SOLVER = 
sig 
    type rules = Rule.rules

    val encodeNrsInTrs: rules -> string
    val snSolver: string -> string -> rules -> bool
    val isTerminatingByExternalSolver: rules -> bool

    val test: int -> bool
    val runDebug: bool ref
end

structure SnSolver: SN_SOLVER =
struct 
local 
    structure L = List
    structure AL = AssocList
    structure LP = ListPair
    structure LU = ListUtil
    structure T = Term
    structure PU = PrintUtil
    structure SU = StringUtil
    structure TU = TestUtil
in

type rules = Rule.rules
val runDebug = ref false
fun dprint mesg = if (!runDebug) then PU.print mesg else ()
fun dprintln mesg = if (!runDebug) then PU.println mesg else ()


datatype foterm = Var of string | Fun of string * (foterm list)

fun encodeTerm (lambda, diamond, pair) (T.Atom a) = Fun (diamond, [])
  | encodeTerm (lambda, diamond, pair) (T.Susp (pi,X)) = Var (Var.toString X)
  | encodeTerm (lambda, diamond, pair) (T.Abst (a,t)) = Fun (lambda, [encodeTerm (lambda, diamond, pair) t])
  | encodeTerm (lambda, diamond, pair) (T.Fun (f,t)) = Fun (Fun.toString f,  [encodeTerm (lambda, diamond, pair) t])
  | encodeTerm (lambda, diamond, pair) (T.Tuple ts) = Fun (pair (length ts),  L.map (encodeTerm (lambda, diamond, pair)) ts)

fun allnames term = (L.map Var.toString (Term.vars term)) @ (L.map Fun.toString (Term.funs term))

fun getFreshNames rs =
    let val names = LU.mapUnion (fn (env,l,r) => LU.union (allnames l, allnames r)) rs
	fun getLambdaSub lambda = if LU.member lambda names
				  then getLambdaSub "l" ^ lambda
				  else lambda
	val lambda = getLambdaSub "lambda"
	fun getDiamondSub diamond = if LU.member diamond names
				  then getDiamondSub "d" ^ diamond
				  else diamond
	val diamond = getLambdaSub "diamond"
	fun getPairSub prepair = if L.exists (String.isPrefix prepair) names
				 then getPairSub "p" ^ prepair
				 else prepair
	val prepair = getPairSub "pair"
	fun pair i = prepair ^ "_" ^ (Int.toString i)
    in (lambda,diamond,pair)
    end

fun encodeRules rs = 
    let val (lambda,diamond,pair) = getFreshNames rs
    in L.map (fn (env,l,r) => (encodeTerm (lambda, diamond, pair) l, encodeTerm (lambda, diamond, pair) r)) rs
    end

fun prFoTerm (Var name) = name
  | prFoTerm (Fun (f,ts)) = f ^ (prFoTerm_list ts)
and prFoTerm_list [] = ""
  | prFoTerm_list ts = LU.toStringCommaRound prFoTerm ts

fun prFoRules (vars,rs) = 
    "(VAR "  ^ (LU.toStringSpace (fn x=>x) vars)  ^ ")\n"
    ^ "(RULES\n" ^ (LU.toString (fn (l,r) => "  " ^ (prFoTerm l) ^ " -> " ^ (prFoTerm r) ^ "\n") rs)
    ^ ")\n"

fun encodeNrsInTrs rs = 
    let val vars = L.map Var.toString (LU.mapUnion (fn (env,l,r) => LU.union (Term.vars l, Term.vars r)) rs)
	val trs = encodeRules rs
    in prFoRules (vars,trs)
    end

local
    val debug = false
    fun dprint mesg = if (!runDebug) orelse debug then PU.print mesg else ()
    fun dprintln mesg = if (!runDebug) orelse debug then PU.println mesg else ()
in
fun snSolver proverPath tmpDir rs =
    let fun receiveReport ins str =
	    if TextIO.endOfStream ins then str
	    else receiveReport ins (str ^ (TextIO.inputN (ins,1)))

	val inTmpFile = let val tmpName = 
				((#file (OS.Path.splitDirFile (OS.FileSys.tmpName ()))) ^ ".trs")
			in OS.Path.joinDirFile {dir=tmpDir,file=tmpName}
			end

	val _ = dprintln "\nwriting to temporary file..."

	(* tmpファイルへ書き込み *)
	val outs = TextIO.openOut inTmpFile
	val _ = TextIO.output (outs, (encodeNrsInTrs rs))
	val _ = TextIO.closeOut outs

	(* snChecker の実行 *)
	val cmd = proverPath ^ " " ^ inTmpFile ^ "\n" (* 結果をエラーに出す場合あり *)
 	val _ = dprint cmd
 	val proverCmd = ("/bin/sh", ["-c",cmd])
	val proc = Unix.execute proverCmd
	val (ins,outs) = Unix.streamsOf proc

	val _ = TextIO.closeOut outs
	val ans = TextIO.inputLine ins
	val result =  (isSome ans) andalso (valOf ans = "YES\n")

	val _ = if (isSome ans) then dprint (valOf ans) else ()
	val report = receiveReport ins ""
	val _ = dprint report

	val _ = TextIO.closeIn ins
	val _ = Unix.reap proc  (* プロセス終了処理*)
	val _ = OS.FileSys.remove inTmpFile 

	val _ = if result then PU.hintln "terminating" else PU.hintln "termination proof failed"
    in result
    end
end

fun isTerminatingByExternalSolver rs = snSolver "./run_snprover.sh"  "." rs

fun test i = snSolver "./run_snprover.sh"  "./tmp" (Nrs.example i)


end (* of local *)
end (* of struct *)
