(* file: subst.sml *)
(* description: substutition *)
(* author: Takahito Aoto *)

signature SUBST = 
sig 
    type atom = Atom.key
    type var = Var.key
    type term = Term.key
    type fresh_env = FreshEnv.key
    type subst = (var * term) list
    type key = subst
    val toString: subst -> string
    val fromString: string -> subst
    val rdAssig: string -> var * term
    val identity: subst
    val singleton: var * term -> subst
    val apply: subst -> term -> term
    val dom: subst -> var list
    val compose: subst -> subst -> subst
    val instantiateFreshEnv: subst -> fresh_env -> (atom * term) list
    val matchTerm: term -> term -> ((atom * term) list * subst) option
    val unify: (atom * term) list * (term * term) list -> (fresh_env * subst) option 
    val match: term  -> term -> (fresh_env * subst) option 
end

structure Subst: SUBST =
struct 

local 
    structure L = List
    structure AL = AssocList
    structure LP = ListPair
    structure LU = ListUtil
    structure T = Term
    structure PU = PrintUtil
    structure SU = StringUtil
    structure TU = TestUtil
in

type atom = Atom.key
type var = Var.key
type term = Term.key
type fresh_env = FreshEnv.key
type subst = (var * term) list
type key = subst

local
    fun toStringPair (x,t) = (Var.toString x) ^ " := " ^ (Term.toString t)
in
fun toString []  = "{ }"
  | toString sigma = "{" ^ (toStringList sigma)
and toStringList [] = "" (* not reaching here *)
  | toStringList [xt] = (toStringPair xt) ^ "}"
  | toStringList (xt::sigma) = (toStringPair xt) ^ ", " ^ (toStringList sigma)
end

local
    open ParseNominal
    infix 6 $--; infix 5 --;  infix 3 >>; infix 0 ||;
    open Term.IO
    fun subst toks = 
	( "{" $-- substlist >> (fn x => x) ) toks
    and substlist toks =
        ( substseq -- $ "}" >> #1 || $ "}" >> (fn _ => []) ) toks
    and substseq toks =
        ( assig -- "," $-- substseq >> (fn (x,ys) => x::ys) || assig >> (fn x => [x]) ) toks
    and assig toks  =
        ( id -- ":=" $-- term >> (fn (X,t) => (Var.fromString X, t)) ) toks 
in
fun rdAssig str = reader assig str
fun fromString str = reader subst str 
end (* of local *)

val _ = TU.test "fromString"
		[ (fromString "{   }" =  []),
		  (fromString "{ X := (f []Z)  }" = [(Var.fromString "X", Term.fromString "(f []Z)")]),
		  (fromString "{X := [(a b)]Z}" = [(Var.fromString "X", Term.fromString "[(a b)]Z")]),
                 (* spaces around ":=" are necessary *)
		  (fromString "{ X := [(a b)]Z, Y := [a](f a) }" = 
		  [(Var.fromString "X", Term.fromString "[(a b)]Z"), (Var.fromString "Y", Term.fromString "[a](f a)")]) ]

val identity = []
fun singleton (X,t) = [(X,t)]

fun apply sigma (T.Atom a) = T.Atom a
  | apply sigma (T.Susp (pi,X)) = (case AL.find X sigma of SOME t => Term.actDot pi t | NONE => T.Susp (pi,X))
  | apply sigma (T.Abst (a,t)) = T.Abst (a, apply sigma t)
  | apply sigma (T.Fun (f,t)) = T.Fun (f, apply sigma t)
  | apply sigma (T.Tuple ts) = T.Tuple (L.map (apply sigma) ts)

val _ = TU.test "apply"
		[ (apply (fromString "{X := a}") (T.fromString "[(a b)]X") = (T.fromString "b")),
		  (apply (fromString "{X := a}") (T.fromString "<[(a b)]X, []Y>") = (T.fromString "<b, []Y>")),
		  (apply (fromString "{X := a, Y := (f []X)}") (T.fromString "<[(a b)]X, []Y>") = (T.fromString "<b, (f []X)>")),
		  (apply (fromString "{X := [(a b)]Y}") (T.fromString "[(b a)]X") <> (T.fromString "[(b a)(a b)]Y")),
		  (apply (fromString "{X := [(a b)]Y}") (T.fromString "[(b a)]X") = (T.fromString "[]Y"))
                 (* action normalizes permutation, while fromString does not *)
		]

fun dom [] = []
  | dom ((X,T.Susp (pi,Y))::sigma) = if Perm.isIdentity pi andalso X = Y
				  then dom sigma
				  else LU.add X (dom sigma)
  | dom ((X,_)::sigma) = LU.add X (dom sigma)

val _ = TU.test "dom"
		[ (dom (fromString "{X := a}")  = [(Var.fromString "X")]),
		  (dom (fromString "{X := a, Y := b}")  = [(Var.fromString "X"), (Var.fromString "Y")]) ]

(* rho o sigma *)
fun compose sigma rho = 
    let val sigma2 = L.map (fn (k,v) => (k, apply rho v)) sigma
	val domain = dom sigma
	val rho2 = L.filter (fn (k,v) => not (LU.member k domain)) rho
    in sigma2 @ rho2
    end

val _ = TU.test "compose"
		[ (compose (fromString "{X := a}") (fromString "{X := b}")  = fromString "{X := a}"),
		  (compose (fromString "{X := [(a b)]Y}") (fromString "{Y := b}")  = fromString "{X := a, Y := b}") ]


(* instantiate freshness env *)
fun instantiateFreshEnv sigma env = 
    L.map (fn (a,X) => (a, apply sigma (Term.Susp (Perm.id, X)))) env

val _ = TU.test "instantiateFreshEnv"
		[ (instantiateFreshEnv (fromString "{X := b}") (FreshEnv.fromString "a#X") 
		   = [(Atom.fromString "a",  Term.fromString "b")]) ]

(* cs: a set of problems of the form 'a # t' *)
(* eqs: a set of problems of the form 's \aeq t' *)
(* env, sigma: partial solution *)
fun unify (cs,eqs) = 
    let fun main [] cs sigma = (case FreshEnv.resolve cs of SOME env => SOME (env,sigma) | NONE => NONE)
	  | main (eq::eqs) cs sigma =
	    case eq of 
		(T.Atom a, T.Atom b) => if (a <> b) then NONE else main eqs cs sigma
	      | (T.Susp (pi,X), t) => if LU.member X (T.vars t)
				     then (case t of 
					       T.Susp (rho,Y) => if X = Y 
								then let val cs2 = L.map (fn a => (a,T.Susp (Perm.id,X)))
											  (Perm.disagreements (pi,rho))
								     in main eqs (cs2@cs) sigma 
								     end
								else NONE
					     | _ => NONE)
				     else let val rho = singleton (X, T.actDot (Perm.inverse pi) t)
					      val eqs2 = L.map (fn (u,v) => (apply rho u, apply rho v)) eqs
					      val cs2 = L.map (fn (a,v) => (a, apply rho v)) cs
					  in main eqs2 cs2 (compose sigma rho)
					  end
	      | (t, T.Susp (pi,X)) => main ((T.Susp (pi,X),t)::eqs) cs sigma
	      | (T.Abst (a,t), T.Abst (b,s)) => if a = b 
						then main ((t,s)::eqs) cs sigma
						(* else main ((T.actDot (Perm.single (a,b)) t, s)::eqs) ((b,t)::cs) sigma *)
                                                (* change of def of aeq reflected 2015/12/29 *)
						else main ((t, T.actDot (Perm.single (a,b)) s)::eqs) ((a,s)::cs) sigma
	      | (T.Fun (f,t), T.Fun (g,s)) => if (f <> g) then NONE else main ((t,s)::eqs) cs sigma
	      | (T.Tuple ts, T.Tuple ss) => main ((LP.zip (ts,ss)) @ eqs) cs sigma
	      | _ => NONE
    in main eqs cs identity
    end

val _ = TU.test "unify"
		[ (unify ([], [(Term.fromString "(f [a](f [b](app <[]X_1, (vr b)>)))", 
				Term.fromString "(f [b](f [a](app <(vr a), []X_1>)))")])
		   =  NONE),
		  (unify ([], [(Term.fromString "(f [a](f [b](app <[]X_2, (vr b)>)))", 
				Term.fromString "(f [b](f [a](app <(vr a), []X_3>)))")])
		   = SOME ([], fromString "{ X_2 := (vr b), X_3 := (vr a) }")),
		  (unify ([], [(Term.fromString "(f [a](f [b](app <(vr b), []X_4>)))", 
				Term.fromString "(f [b](f [a](app <(vr a), []X_5>)))")])
		   = SOME ([], fromString "{ X_4 := [(a b)]X_5 }")),
		  (unify ([], [(Term.fromString "(f [a](f [b](app <(vr b), []X_6>)))", 
				Term.fromString "(f [a](f [a](app <(vr a), []X_7>)))")])
		   = SOME (FreshEnv.fromString "b#X_7", fromString "{ X_6 := [(a b)]X_7 }")) ]

fun matchTerm pattern term  = 
    let exception UnMatched
        fun main [] (cntxt,sigma) = SOME (L.map (fn (a,t) => (a, apply sigma t)) cntxt,  sigma)
          | main (eq::rest) (cntxt,sigma) =
	    case eq of 
		(T.Atom a, T.Atom b) => if (a <> b) then NONE else main rest (cntxt,sigma)
               |(T.Susp (pi,X), t) => (case AL.add (X, Term.actDot (Perm.inverse pi) t) sigma of
					  SOME rho => main rest (cntxt,rho) | NONE => raise UnMatched)
	      | (T.Abst (a,t), T.Abst (b,s)) => if a = b 
						then main ((t,s)::rest) (cntxt,sigma)
						(* else main ((T.actDot (Perm.single (a,b)) t, s)::rest) ((b,t)::cntxt, sigma) *)
                                                (* change of def of aeq reflected 2015/12/29 *)
						else main ((t, T.actDot (Perm.single (a,b)) s)::rest) ((a,s)::cntxt, sigma)
	      | (T.Fun (f,t), T.Fun (g,s)) => if (f <> g) then NONE else main ((t,s)::rest) (cntxt,sigma)
	      | (T.Tuple ts, T.Tuple ss) => main ((LP.zip (ts,ss)) @ rest) (cntxt,sigma)
	      | _ => NONE
    in	main [(pattern,term)] ([],[])
	handle UnMatched => NONE
    end


fun match pattern term = case matchTerm pattern term of
			     NONE => NONE
			   | SOME (cntxt,sigma) => (case FreshEnv.resolve cntxt of 
							SOME env => SOME (env,sigma)
						      | NONE => NONE)

val _ = TU.test "match"
		[ (match (Term.fromString "(app <(lam [a][]X), []Y>)")
			 (Term.fromString "(app <(lam [a](lam [b](app <b,a>))), b>)")
		   = SOME ([], fromString "{ Y := b, X := (lam [b](app <b,a>)) }")) ]



end (* of local *)

end
