(* file: term.sml *)
(* description: nominal terms *)
(* author: Takahito Aoto *)

signature TERM = 
sig 
    type var = Var.key
    type atom = Atom.key
    type perm = Perm.key
    type fun_key = Fun.key
    datatype term = Atom of atom 
		  | Susp of perm * var 
		  | Abst of atom * term 
		  | Fun of fun_key * term 
		  | Tuple of term list
    type key = term

    val toString: term -> string
    val fromString: string -> term

    val vars: term -> var list
    val varlist: term -> var list
    val susps: term -> (perm * var) list
    val isSusp: term -> bool
    val atoms: term -> atom list
    val freeAtoms: term -> atom list
    val funs: term -> fun_key list

    val incVarIndex: int -> term -> term 
    val actDot: perm -> term -> term   (* pi \cdot t *)
    val actHat: perm -> term -> term   (* t^pi *)
    val apply: (atom * atom) list -> term -> term   (* update atoms by a function A -> A *)

    val nonDuplicatingSkeletons: term -> bool
    val varSkeletons: term -> term list

    structure IO : 
    sig 
      val term: ParseNominal.token list -> term * ParseNominal.token list
      val termlist: ParseNominal.token list -> term list * ParseNominal.token list
      val termseq: ParseNominal.token list -> term list * ParseNominal.token list
      val swaplist: ParseNominal.token list -> term * ParseNominal.token list
      val swapseq: ParseNominal.token list -> perm  * ParseNominal.token list
      val swap: ParseNominal.token list -> (atom * atom) * ParseNominal.token list
   end
end

structure Term : TERM =
struct
local 
    structure B = Bool
    structure L = List
    structure LP = ListPair
    structure LU = ListUtil
    structure PU = PrintUtil
    structure SU = StringUtil
    structure TU = TestUtil
in

type var = Var.key
type atom = Atom.key
type perm = Perm.key
type fun_key = Fun.key
datatype term = Atom of atom 
	      | Susp of perm * var 
	      | Abst of atom * term 
	      | Fun of fun_key * term 
	      | Tuple of term list
type key = term

structure IO = 
struct
    fun makeTuple ts = Tuple ts
    fun makeTuple0 _ = Tuple []
    fun makeFun ((str, t), _) = Fun (Fun.fromString str, t)
    fun makeAbst (((_,str),_), t) = Abst (Atom.fromString str, t)
    fun makeIdVar str = Susp (Perm.id, Var.fromString str)
    fun makeVar ((pi,_),str) = Susp (pi, Var.fromString str)
    fun makeList (t, ts) = t::ts
    fun makeList1 t = t::nil
    fun makeNull t = []
    fun takeFirst (a,b) = (print b; a)
    fun makeSwap ((a,b),_) = (Atom.fromString a, Atom.fromString b)
    fun makeAtom str = Atom (Atom.fromString str)

    open ParseNominal
    infix 6 $--
    infix 5 --
    infix 3 >>
    infix 0 ||

    fun term toks =
         ("<" $-- termlist >> makeTuple ||
          "(" $-- id -- term -- $ ")"  >> makeFun ||
          $ "[" -- id -- $ "]" -- term >> makeAbst || 
          "[" $-- swaplist >> (fn x => x) || 
	  atom) toks
    and termlist toks =
        ( termseq -- $ ">" >> #1 || $ ">" >> (fn _ => []) ) toks
    and termseq toks =
        ( term -- "," $-- termseq >> makeList || term >> makeList1 ) toks
    and swaplist toks =
        ( swapseq -- $ "]" -- id >> makeVar || "]" $-- id >> makeIdVar ) toks 
    and swapseq toks =
        ( swap -- swapseq >> makeList || swap >> makeList1 ) toks
    and swap toks  =
        ("(" $-- id -- id -- $ ")"  >> makeSwap) toks
    and atom toks  =
        (id  >> makeAtom) toks

    fun fromString str = reader term str

end

val fromString = IO.fromString

fun toString (Atom a) =  (Atom.toString a)
  | toString (Susp (pi,X))  = (Perm.toString (Perm.normalize pi)) ^ (Var.toString X)
  | toString (Abst (a,t))  = "[" ^ (Atom.toString a) ^ "]" ^ (toString t)
  | toString (Fun (f,t))  = "(" ^ (Fun.toString f) ^  " " ^ (toString t) ^ ")"
  | toString (Tuple ts) = LU.toStringCommaAngle toString ts

val _ = TU.test "toString"
		[ (toString (fromString "(sub <[a](app <[]X,[]Y>),[]Z>)") = "(sub <[a](app <[]X,[]Y>),[]Z>)"),
		  (toString (fromString "(sub <[a](app <[(a b)]X,[]Y>),[]Z>)") = "(sub <[a](app <[(a b)]X,[]Y>),[]Z>)"),
		  (toString (fromString "( sub < [a] b, [] X > )") = "(sub <[a]b,[]X>)")]


fun vars (Atom a) = []
  | vars (Susp (pi,X))  = [X]
  | vars (Abst (a,t))  = vars t
  | vars (Fun (f,t))  = vars t
  | vars (Tuple ts) = ListUtil.unionAll (L.map vars ts)

val _ = TU.test "vars"
 		[ (vars (fromString "(sub <[a](app <[]X,[]Y>),[]Z>)") = [Var.fromString "Z", Var.fromString "Y", Var.fromString "X"]),
		  (vars (fromString "(sub <[a](app <[]X,[]Y>),[]Y>)") = [Var.fromString "Y", Var.fromString "X"]) ]

fun varlist (Atom a) = []
  | varlist (Susp (pi,X))  = [X]
  | varlist (Abst (a,t))  = varlist t
  | varlist (Fun (f,t))  = varlist t
  | varlist (Tuple ts) = LU.mapAppend varlist ts

val _ = TU.test "varlist"
 		[ (varlist (fromString "(sub <[a](app <[]X,[]X>),[]Y>)") = [Var.fromString "X", Var.fromString "X", Var.fromString "Y"]) ]


fun susps (Atom a) = []
  | susps (Susp (pi,X))  = [(pi,X)]
  | susps (Abst (a,t))  = susps t
  | susps (Fun (f,t))  = susps t
  | susps (Tuple ts) = ListUtil.unionAll (L.map susps ts)

val _ = TU.test "susps"
 		[ (susps (fromString "(sub <[a](app <[]X,[]Y>),[]Z>)") = 
		   [(Perm.id, Var.fromString "Z"), (Perm.id, Var.fromString "Y"), (Perm.id, Var.fromString "X")]),
		  (susps (fromString "(sub <[a](app <[]X,[(a b)]Y>),[]Y>)") = 
		  [(Perm.id, Var.fromString "Y"), (Perm.fromString "[(a b)]", Var.fromString "Y"), (Perm.id, Var.fromString "X")]) ]

fun isSusp (Susp _) = true
  | isSusp _ = false


fun atoms (Atom a) = [a]
  | atoms (Susp (pi,X))  = Perm.atoms pi
  | atoms (Abst (a,t))  = LU.add a (atoms t)
  | atoms (Fun (f,t))  = atoms t
  | atoms (Tuple ts) = ListUtil.unionAll (L.map atoms ts)

val _ = TU.test "atoms"
 		[ (atoms (fromString "(sub <[a](app <[]X,[]Y>),[]Z>)") =  [Atom.fromString "a"]),
		  (atoms (fromString "(sub <[a](app <[]X,[(a b)]Y>),[]Y>)") = [Atom.fromString "a", Atom.fromString "b"]) ]

fun freeAtoms (Atom a)  = [a]
  | freeAtoms (Susp (pi,X))  = Perm.atoms pi  (* [] revised 2015, 11/16 *)
  | freeAtoms (Abst (a,t))  = ListUtil.delete a (freeAtoms t)
  | freeAtoms (Fun (f,t))  = freeAtoms t
  | freeAtoms (Tuple ts) = ListUtil.unionAll (L.map freeAtoms ts)

val _ = TU.test "freeAtoms"
 		[ (freeAtoms (fromString "(sub <[a](app <a,[]Y>),[]Z>)") =  []),
		  (freeAtoms (fromString "(sub <[a](app <[]X,[(a b)]Y>),[]Y>)") = [Atom.fromString "b"]),
		  (freeAtoms (fromString "(sub <[a](app <[]X,[(a b)]Y>),a>)") = [Atom.fromString "a",Atom.fromString "b"]) ]

fun funs (Atom a) = []
  | funs (Susp (pi,X))  = []
  | funs (Abst (a,t))  = funs t
  | funs (Fun (f,t))  = [f] @ (funs t)
  | funs (Tuple ts) = ListUtil.unionAll (L.map funs ts)

fun incVarIndex n (Atom a) = Atom a
  | incVarIndex n (Susp (pi,X)) = Susp (pi,Var.incIndex n X)
  | incVarIndex n (Abst (a,t)) = Abst (a, incVarIndex n t)
  | incVarIndex n (Fun (f,t)) = Fun (f, incVarIndex n t)
  | incVarIndex n (Tuple ts) = Tuple (L.map (incVarIndex n) ts)

val _ = TU.test "incVarIndex"
 		[ (incVarIndex 2 (fromString "(sub <[a](app <[]X_1,[]Y_2>),[]Z>)") =  
		   fromString "(sub <[a](app <[]X_3,[]Y_4>),[]Z_2>)") ]

(* pi \cdot t *)
fun actDot pi (Atom a) = Atom (Perm.act pi a)
  | actDot pi (Susp (pi2,X)) = Susp (Perm.compose pi pi2, X)
  | actDot pi (Abst (a,t)) = Abst (Perm.act pi a, actDot pi t)
  | actDot pi (Fun (f,t)) = Fun (f, actDot pi t)
  | actDot pi (Tuple ts) = Tuple (L.map (actDot pi) ts)

val _ = TU.test "actDot"
 		[ (actDot (Perm.fromString "[(a b)]") (fromString "(sub <[a](app <[(a b)]X,[]Y>),b>)") =  
		   (fromString "(sub <[b](app <[]X,[(a b)]Y>),a>)")) ]

(* t^pi *)
fun actHat pi (Atom a) = Atom (Perm.act pi a)
  | actHat pi (Susp (pi2,X)) = Susp ((Perm.compose pi (Perm.compose pi2 (Perm.inverse pi))), X)
  | actHat pi (Abst (a,t)) = Abst (Perm.act pi a, actHat pi t)
  | actHat pi (Fun (f,t)) = Fun (f, actHat pi  t)
  | actHat pi (Tuple ts) = Tuple (L.map (actHat pi ) ts)

val _ = TU.test "actHat"
 		[ (actHat (Perm.fromString "[(a b)]") (fromString "(sub <[a](app <[(a b)]X,[]Y>),b>)") =  
		   (fromString "(sub <[b](app <[(a b)]X,[]Y>),a>)")) ]

(* update atoms like t^pi, not by permutation pi but by a function A -> A *)
fun apply pimap (Atom a) = Atom (Perm.apply pimap a)
  | apply pimap (Susp (pi,X)) = Susp (L.map (fn (a,b) => (Perm.apply pimap a,Perm.apply pimap b)) pi, X)
  | apply pimap (Abst (a,t)) = Abst (Perm.apply pimap a, apply pimap t)
  | apply pimap (Fun (f,t)) = Fun (f, apply pimap  t)
  | apply pimap (Tuple ts) = Tuple (L.map (apply pimap) ts)

val _ = TU.test "actHat"
 		[ (actHat (Perm.fromString "[(a b)]") (fromString "(sub <[a](app <[(a b)]X,[]Y>),b>)") =  
		   (fromString "(sub <[b](app <[(a b)]X,[]Y>),a>)")) ]

fun nonDuplicatingSkeletons term = 
    let fun nds xs (Atom b) = true
	  | nds xs (Susp (pi,X)) = true
	  | nds xs (Abst (b,t)) = not (LU.member b xs) andalso nds (b::xs) t
	  | nds xs (Fun (f,t)) = nds xs t
	  | nds xs (Tuple ts) = L.all (nds xs) ts
    in nds [] term
    end

val _ = TU.test "nonDuplicatingSkeletons"
		[(nonDuplicatingSkeletons (fromString "(forall [a][(a b)]X)") = true),
		 (nonDuplicatingSkeletons (fromString "[a](forall [a][(a b)]X)") = false),
		 (nonDuplicatingSkeletons (fromString "[b](forall [a][(a b)]X)") = true)]

fun varSkeletons (Atom b) = []
  | varSkeletons (Susp (pi,X)) = [Susp (pi,X)]
  | varSkeletons (Abst (b,t)) = L.map (fn u => Abst (b,u)) (varSkeletons t)
  | varSkeletons (Fun (f,t)) = varSkeletons t
  | varSkeletons (Tuple ts) = LU.mapAppend varSkeletons ts

val _ = TU.test "varSkeletons"
		[(varSkeletons (fromString "(forall [a][(a b)]X)") = [fromString "[a][(a b)]X"]),
		 (varSkeletons (fromString "(app <(lam [a][]X), []Y>)")
		  = [(fromString "[a][]X"),  (fromString "[]Y") ]),
		 (varSkeletons (fromString "(app [b]<(lam [a][]X), []Y>)")
		  = [ (fromString "[b][a][]X"),  (fromString "[b][]Y") ]),
		 (varSkeletons (fromString "(app [c][b]<(lam [a][]X), []Y>)")
		  =  [ (fromString "[c][b][a][]X"),  (fromString "[c][b][]Y") ]),
		 (varSkeletons (fromString "[d](app [c][b]<(lam [a][]X), []Y>)")
		  =  [ (fromString "[d][c][b][a][]X"),  (fromString "[d][c][b][]Y") ])]

end (* of local *)		     

end
