(* file: tpdb.yacc *)
(* description: parser generator *)
(* author: Takahito Aoto *)

local 
    structure L = List
    structure LU = ListUtil
    structure PU = PrintUtil
    structure T = Term
in
datatype gterm = AVF of ((string * string) list)  * string * gterm
               | Tuple of (gterm list)
               | Abs of (string * gterm)

fun prGTerm (vars,atoms) (Tuple ts) = LU.toStringCommaAngle (prGTerm (vars,atoms)) ts
  | prGTerm (vars,atoms) (Abs (a,t)) = "[" ^ a ^ "]" ^ (prGTerm (vars,atoms) t)
  | prGTerm (vars,atoms) (AVF (xs,a,t)) = ((if null xs then ""
                              else (LU.toStringSpaceSquare (fn (x,y)=> "(" ^ x ^ " " ^ y ^ ")") xs))
			      ^ (if LU.member a atoms then "[" ^ a ^ "]"
				 else if LU.member a vars then a
				 else "(" ^ a ^ " " ^ (prGTerm (vars,atoms) t) ^ ")"  ))

fun prEnv xs = LU.toStringComma (fn (x,y)=> x ^ "#" ^ y) xs

datatype tpdb_decl = VAR of string list
		   | ATOMS of string list
                   | RULES of ((string * string) list * gterm * gterm) list
                   | REST of string

exception TpdbError of string

fun getCategory f xss = let val yss = L.mapPartial f xss in LU.mapAppend (fn x=>x) yss end
val getVAR = getCategory (fn top => case top of VAR xs => SOME xs | _ => NONE) 
val getATOMS = getCategory (fn top => case top of ATOMS xs => SOME xs | _ => NONE) 
val getRULES = getCategory (fn top => case top of RULES xs => SOME xs | _ => NONE)
val getREST = getCategory (fn top => case top of REST xs => SOME [xs] | _ => NONE)

exception RuleFormatError of (string list * string list * ((string * string) list * gterm * gterm) * string)
exception FormatError

fun checkRule vars atoms (env, lhs, rhs) = 
    let val checkenv = L.all (fn (x,y) => (LU.member x atoms) andalso (LU.member y vars)) env
	val checkdisj = L.all (fn x => not (LU.member x atoms)) vars
	val _ =  if checkenv andalso checkdisj then () 
                 else raise RuleFormatError (vars,atoms,(env, lhs, rhs), "undeclared atom or variable?\n")
	fun transTerm (AVF (perm, id, arg)) = 
	    if LU.member id vars 
	    then if (arg = Tuple []) 
		 then T.Susp (L.map (fn (x,y)=> (Atom.fromString x, Atom.fromString y)) perm, Var.fromString id) 
		 else raise RuleFormatError (vars,atoms,(env, lhs, rhs), "variable with arguments?\n")
	    else if LU.member id atoms 
	    then if (arg = Tuple []) andalso (null perm) 
		 then T.Atom (Atom.fromString id) 
		 else raise RuleFormatError (vars,atoms,(env, lhs, rhs), "atom with suspension or arguments?\n")
	    else if (null perm)  
	    then T.Fun (Fun.fromString id, transTerm arg)
	    else raise RuleFormatError (vars,atoms,(env, lhs, rhs), "non-recognizable function or atom or variable?\n")
	  | transTerm (Tuple ts) = T.Tuple (L.map transTerm ts)
	  | transTerm (Abs (id,arg)) = 
	    if LU.member id atoms  then T.Abst (Atom.fromString id, transTerm arg) 
            else raise RuleFormatError (vars,atoms,(env, lhs, rhs), "undeclared atom?\n")
	val env2 = L.map (fn (x,y) => (Atom.fromString x, Var.fromString y)) env
	val lhs2 = transTerm lhs
	val rhs2 = transTerm rhs
    in (env2, lhs2, rhs2)
    end

local
    val debug = false
    fun dprint mesg = if (!PU.runDebug) orelse debug then print mesg else ()
    fun dprintln mesg = if (!PU.runDebug) orelse debug then PU.println mesg else ()
in
fun checkDecl decls = 
    let val vars = getVAR decls
	(* val _ = print "(VAR" *)
	(* val _ = L.app (fn x => print (" " ^ x)) vars *)
	(* val _ = print ")\n"  *)
        val atoms = getATOMS decls
	(* val _ = print "(ATOMS"  *)
	(*  val _ = L.app (fn x => print (" " ^ x)) atoms  *)
	(*  val _ = print ")\n"  *)
	val rules = L.map (checkRule vars atoms) (getRULES decls)
	 (* val _ = print "(RULES\n"  *)
         (* val _ = Rule.prRules rules;  *)
	 (* val _ = print ")\n"  *)
	val rest = getREST decls 
        val _ = if null rest then () 
                else let val _ = dprintln "ignored inputs:"
	                 val _ = L.app (fn ys => dprintln ("(" ^ ys  ^ ")")) rest
                         val _ = dprintln ""
		       in () end 
    in rules
    end
     handle RuleFormatError (vars,atoms,(env, lhs, rhs),msg) => (print "Failed to read rule: "; 
                                                     print ((prEnv env) ^ " |- " 
							    ^ (prGTerm (vars,atoms) lhs) ^ " -> "  
							    ^ (prGTerm (vars,atoms) rhs)  ^ "\n");
                                                     print msg;
                                                     raise FormatError)




end

end
(* 桼 *)
%%
%name		Tpdb
%term           ID of string
              | STR of string
              | LPAREN
              | RPAREN
              | LSQBRA
              | RSQBRA
              | LBRA
              | RBRA
              | COMMA
              | SHARP
              | VDASH
              | ARROW
              | VAR 
              | ATOMS
              | RULES
              | EOF 

%nonterm	XstartX of Rule.rules
              | XspecX of tpdb_decl list
              | XdeclX of tpdb_decl
              | XidlistX of string list
              | XrulelistX of ((string * string) list * gterm * gterm) list
              | XruleX of (string * string) list * gterm * gterm 
              | XcntxtX of (string * string) list
              | XexprX of gterm
              | XtermX of gterm
              | XvarX of (string * string) list
              | XexprlistX of gterm list
              | XpermX of (string * string) list
              | XswaplistX of (string * string) list
              | XanylistX of string
              | XanyX of string
%pos	   	int
%eop		EOF
%noshift	EOF

%%
XstartX:      XspecX                       (checkDecl (XspecX))
XspecX:       LPAREN XdeclX RPAREN XspecX  (XdeclX::XspecX)
            |                              ([])
XdeclX:       VAR XidlistX                 (VAR XidlistX)
            | ATOMS XidlistX               (ATOMS XidlistX)
            | RULES XrulelistX             (RULES XrulelistX)
            | ID XanylistX                 (REST (ID ^ " " ^ XanylistX))
XidlistX:     ID XidlistX                  (ID::XidlistX)
            |                              ([])
XrulelistX:  XruleX COMMA XrulelistX      (XruleX::XrulelistX)
            | XruleX                       ([XruleX])
            |                              ([])
XruleX:      XcntxtX VDASH XexprX ARROW XexprX  ((XcntxtX,XexprX1,XexprX2))
XcntxtX:      ID SHARP ID COMMA XcntxtX   ((ID1,ID2)::XcntxtX)
            | ID SHARP ID                  ([(ID1,ID2)])
            |                              ([])
XexprX:       ID XexprX                    (AVF ([],ID,XexprX))
            | XtermX                       (XtermX)
XtermX:       ID                           (AVF ([],ID,Tuple []))
            | XpermX ID                    (AVF (XpermX,ID,Tuple []))
            | LBRA RBRA                    (Tuple [])
            | LPAREN XexprX RPAREN         (XexprX)
            | LBRA XexprlistX RBRA         (Tuple XexprlistX)
            | LSQBRA ID RSQBRA XexprX      (Abs (ID,XexprX))
XpermX:       LSQBRA XswaplistX RSQBRA     (XswaplistX)
XswaplistX:   LPAREN ID ID RPAREN          ([(ID1,ID2)])
            | LPAREN ID ID RPAREN XswaplistX ((ID1,ID2)::XswaplistX)
XexprlistX:   XexprX 			   ([XexprX])
            | XexprX COMMA XexprlistX      (XexprX::XexprlistX)
XanylistX:    XanyX XanylistX              (XanyX ^ " " ^ XanylistX)
            |                              ("")
XanyX:        ID                           (ID)
            | LPAREN XanylistX RPAREN      ("(" ^ XanylistX ^ ")")
            | COMMA                        (",")
            | SHARP                        ("#")
            | LSQBRA                       ("[")
            | RSQBRA                       ("]")
            | STR                          (STR)
